/**
   @file coord.h
   @brief 座標定義用ヘッダ

   二次元座標を表すクラスと専用の関数群からなるライブラリ

   @author tadah_fussy
   @date 2016/06/13 DrawingMethod.h から分離
   @date 2019/03/27 コメントの見直し
   @date 2021/03/13 名前空間の変更、Less, Convert, AbsDiff の追加
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef COORD_20210402_H
#define COORD_20210402_H

#include <cmath>
#include <limits>
#include <iostream>

/**
   @brief 作成者 tadah_fussy
**/
namespace tadah_fussy
{
  /**
     @brief 座標定義用構造体(2D用)

     二次元座標を表すクラス。テンプレート引数の T は座標値の型を表す。
     コピー・コンストラクタと代入演算子は利用可能。
  **/
  template< class T >
  struct Coord
  {
    T x; ///< x 座標
    T y; ///< y 座標

    /* コンストラクタ */

    /// @brief デフォルト・コンストラクタ
    ///
    /// ( x, y ) = ( 0, 0 ) で初期化される
    Coord() : x( T{} ), y ( T{} ) {}

    /// @brief x, y 座標を指定して構築
    ///
    /// @param x_ x 座標
    /// @param y_ y 座標
    Coord( T x_, T y_ ) : x( x_ ), y ( y_ ) {}

    /* 演算子の多重定義 */

    /* 四則演算(和と差) */

    /// @brief 座標を加算する
    ///
    /// @note 座標値によっては演算結果がオーバーフローする可能性があるが、チェックはしていない。
    ///
    /// @param c 加算対象の座標
    /// @return 自分自身への参照
    Coord& operator+=( const Coord& c )
    {
      x += c.x;
      y += c.y;

      return( *this );
    }
    /// @brief 座標を減算する
    ///
    /// @note 座標値によっては演算結果がオーバーフローする可能性があるが、チェックはしていない。
    ///
    /// @param c 減算対象の座標
    /// @return 自分自身への参照
    Coord& operator-=( const Coord& c )
    {
      x -= c.x;
      y -= c.y;

      return( *this );
    }

    /// @brief 座標どうしの和を求める
    ///
    /// @note 座標値によっては演算結果がオーバーフローする可能性があるが、チェックはしていない。
    ///
    /// @param dest 加算対象の座標
    /// @return 求めた和
    Coord operator+( const Coord& dest ) const
    {
      Coord< T > c = *this;
      c += dest;

      return( c );
    }

    /// @brief 座標どうしの差を求める
    ///
    /// @note 座標値によっては演算結果がオーバーフローする可能性があるが、チェックはしていない。
    ///
    /// @param dest 減算対象の座標
    /// @return 求めた差
    Coord operator-( const Coord& dest ) const
    {
      Coord< T > c = *this;
      c -= dest;

      return( c );
    }

    /* 比較用関数(等号・不等号) */

    /// @brief 座標どうしが一致するかを返す
    ///
    /// @param c 比較対象の座標
    /// @return 一致すれば true を返す
    bool operator==( const Coord& c ) const
    {
      return( x == c.x && y == c.y );
    }

    /// @brief 座標どうしが一致しないかを返す
    ///
    /// @param c 比較対象の座標
    /// @return 一致しなければ true を返す
    bool operator!=( const Coord& c ) const
    {
      return( x != c.x || y != c.y );
    }
  };

  /// @brief ストリーム出力
  ///
  /// @param os 対象の出力ストリームへの参照
  /// @param c 出力対象の座標への参照
  /// @return 出力したストリーム(os)への参照
  template< class T >
  std::ostream& operator<<( std::ostream& os, const Coord< T >& c )
  {
    os << c.x << "," << c.y;

    return( os );
  }

  /**
     @brief 座標計算関数用名前空間
  **/
  namespace coord_lib
  {
    /**
       @brief 座標どうしの大小比較 ( Less ) 用関数オブジェクト

       @note std::map などのキーとして CoordLib::Coord を利用する場合を想定して用意されている。
    **/
    struct Less
    {
      /// @brief 座標どうしの大小比較 ( より小さい )
      ///
      /// まずは x 座標で比較し、x 座標が等しければ y 座標で比較する
      ///
      /// @param c1,c2 比較対象の座標
      /// @return c1 の方がより小さいと判定されたら true を返す
      template< class T >
      bool operator()( const Coord< T >& c1, const Coord< T >& c2 ) const
      {
        if ( c1.x == c2.x )
          return( c1.y < c2.y );
        else
          return( c1.x < c2.x );
      }
    };

    /*
      Coord 用ヘルパ関数 ( 座標を使った処理関数群 )
    */

    /// @brief 異なる型の座標へ変換する
    ///
    /// @param c コピー元の座標
    template< class U, class T >
    Coord< U > Convert( const Coord< T >& c )
    {
      return( Coord< U >( static_cast< U >( c.x ), static_cast< U >( c.y ) ) );
    }

    /// @brief 二数の差の絶対値を返す
    ///
    /// @param t1,t2 ニ数
    /// @return 差の絶対値
    template< class T >
    T AbsDiff( const T& t1, const T& t2 )
    {
      return( ( t2 > t1 ) ? t2 - t1 : t1 - t2 );
    }

    /// @brief 距離(ノルム)の計算
    ///
    /// @param c1,c2 演算対象の座標
    /// @return c1, c2 間の距離
    template< class D, class T >
    D Distance( Coord< T > c1, Coord< T > c2 )
    {
      return( ( std::sqrt( std::pow( static_cast< D >( AbsDiff( c1.x, c2.x ) ), D{2} ) +
                           std::pow( static_cast< D >( AbsDiff( c1.y, c2.y ) ), D{2} )
                           ) ) );
    }

    /// @brief 二座標間の差分(絶対値)を求める
    ///
    /// x, y それぞれに対して差の絶対値を求め、変数 diff に代入する
    ///
    /// @note c1 と c2 の値によっては、差分がオーバーフローする可能性がある
    /// @param c1,c2 差分を計算する二点
    /// @param diff 求めた差分を代入する変数へのポインタ
    template< class T >
    void Diff( Coord< T > c1, Coord< T > c2, Coord< T >* diff )
    {
      if ( diff != nullptr ) {
        diff->x = AbsDiff( c1.x, c2.x );
        diff->y = AbsDiff( c1.y, c2.y );
      }
    }

    /// @brief 二座標 s, e 間の直線の向きを求める
    ///
    /// x, y それぞれ e >= s ならば 1 e < s ならば -1 とする
    ///
    /// @param s 始点
    /// @param e 終点
    /// @param sign 求めた直線の向きを代入する変数へのポインタ
    template< class T >
    void Sign( Coord< T > s, Coord< T > e, Coord< int8_t >* sign )
    {
      if ( sign != nullptr ) {
        sign->x = ( e.x >= s.x ) ? 1 : -1;
        sign->y = ( e.y >= s.y ) ? 1 : -1;
      }
    }

    /// @brief 傾きを返す
    ///
    /// @param c0,c1 直線を通る二点
    /// @param rotate falseで dy/dx, trueで dx/dyを返す
    /// @return 傾き(dy/dx or dx/dy)
    template< class D, class T >
    D Gradient( Coord< T > c0, Coord< T > c1, bool rotate = false )
    {
      if ( c0.x == c1.x && c0.y == c1.y ) return( std::numeric_limits< D >::quiet_NaN() );
      if ( ( ! rotate ) && ( c0.x == c1.x ) ) return( std::numeric_limits< D >::infinity() );
      if ( (   rotate ) && ( c0.y == c1.y ) ) return( std::numeric_limits< D >::infinity() );

      Coord< D > d0 = Convert< D >( c0 );
      Coord< D > d1 = Convert< D >( c1 );

      if ( rotate )
        return( ( d1.x - d0.x ) / ( d1.y - d0.y ) );
      else
        return( ( d1.y - d0.y ) / ( d1.x - d0.x ) );
    }

    /// @brief 切片を返す
    ///
    /// @param c0,c1 直線を通る二点
    /// @param rotate false で y 軸, true で x 軸との交点を返す
    /// @param axis : 軸の位置 ( x 軸または y 軸がもう一方の軸と交差する位置 )
    /// @return 切片
    template< class D, class T >
    D Intercept( Coord< T > c0, Coord< T > c1, bool rotate = false, T axis = 0 )
    {
      if ( c0.x == c1.x && c0.y == c1.y ) return( std::numeric_limits< D >::quiet_NaN() );
      if ( ( ! rotate ) && ( c0.x == c1.x ) ) return( std::numeric_limits< D >::infinity() );
      if ( (   rotate ) && ( c0.y == c1.y ) ) return( std::numeric_limits< D >::infinity() );

      Coord< D > d0 = Convert< D >( c0 );
      Coord< D > d1 = Convert< D >( c1 );

      if ( rotate )
        return( ( d1.x - d0.x ) * ( axis - d0.y ) / ( d1.y - d0.y ) + d0.x );
      else
        return( ( d1.y - d0.y ) * ( axis - d0.x ) / ( d1.x - d0.x ) + d0.y );
    }
  } // namespace coord_lib

} // namespace tadah_fussy

#endif
