/**
   @file equation.h
   @brief 連立方程式定義用ヘッダ

   @author tadah_fussy
   @date 2015/11/15 コメント追加
   @date 2019/03/27 コメントの見直し
   @date 2019/04/04 コメントの見直し
   @date 2019/04/07 Matrix::clone の廃止に伴う変更
   @date 2019/04/16 error_library.h の見直しに伴う変更。LinearEquationSystem の廃止。
   @date 2019/04/27 全ての解法について関数オブジェクトの形式に変更
   @date 2021/03/14 名前空間の変更
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef EQUATION_20210402_H
#define EQUATION_20210402_H

#include <string>
#include <vector>
#include <iostream>
#include <exception>

#include <common/math_library.h>

#include <cmath>

/**
   @brief 作成者 tadah_fussy
**/
namespace tadah_fussy
{
  namespace math_lib
  {
    /**
       @brief 連立方程式・最適化問題の解法用名前空間
    **/
    namespace solver
    {
      /**
	 @brief ガウスの消去法による連立方程式の解法
      **/
      struct GaussianElimination
      {
	/// @brief ガウスの消去法により連立方程式の解を求める
	///
	/// - 処理途中で係数行列の対角成分にゼロが見つかった場合、ゼロ除算が発生するため例外 std::runtime_error を投げる
	///
	/// @note 係数行列と右辺(解ベクトル)の大きさが等しいかチェックしていないことに注意
	/// @note 解が不定または不能の場合、理論上は処理途中で係数行列の対角成分にゼロが発生する。
	///       しかし、誤差により完全にゼロにはならず、処理が継続される場合もある。
	///       そのとき、解が不定ならば無数にある解の一つが得られ、不能の場合は正しくない結果となる。
	///       得られた結果が正しいかどうかは、 MathLib::VerifyLinearEquationSystem() を使って確認することができる。
	///
	/// @param a 対象の連立方程式の係数行列
	/// @param b 対象の連立方程式の解ベクトルの先頭位置(計算結果の未知数がそのまま返される)
	/// @return なし
	template< typename T, class Ran >
	void operator()( SquareMatrix< T > a, Ran b );

	/// @brief ガウスの消去法により連立方程式の解を一度に求める
	///
	/// - b が未定義の場合、assert を呼び出す
	/// - 処理途中で係数行列の対角成分にゼロが見つかった場合、ゼロ除算が発生するため例外 std::runtime_error を投げる
	/// - 係数行列の列数と右辺(解行列)の行数が等しくない場合は例外 ExceptionNotEqualLength< SquareMatrix< T >::size_type > を投げる
	///
	/// @note 解が不定または不能の場合、理論上は処理途中で係数行列の対角成分にゼロが発生する。
	///       しかし、誤差により完全にゼロにはならず、処理が継続される場合もある。
	///       そのとき、解が不定ならば無数にある解の一つが得られ、不能の場合は正しくない結果となる。
	///       得られた結果が正しいかどうかは、 MathLib::VerifyLinearEquationSystem_Multi() を使って確認することができる。
	///
	/// @param a 対象の連立方程式の係数行列
	/// @param b 対象の連立方程式の解行列へのポインタ(計算結果の未知数がそのまま返される)
	/// @return なし
	template< typename T, class MATRIX >
	void multiple( SquareMatrix< T > a, MATRIX* b );
      };

      /**
	 @brief ガウス・ジョルダン法による連立方程式の解法
      **/
      struct GaussJordanElimination
      {
	/// @brief ガウス・ジョルダン法により連立方程式の解を求める
	///
	/// - 処理途中で係数行列の対角成分にゼロが見つかった場合、ゼロ除算が発生するため例外 std::runtime_error を投げる
	///
	/// @note 係数行列と右辺(解ベクトル)の大きさが等しいかチェックしていないことに注意
	/// @note 解が不定または不能の場合、理論上は処理途中で係数行列の対角成分にゼロが発生する。
	///       しかし、誤差により完全にゼロにはならず、処理が継続される場合もある。
	///       そのとき、解が不定ならば無数にある解の一つが得られ、不能の場合は正しくない結果となる。
	///       得られた結果が正しいかどうかは、 MathLib::VerifyLinearEquationSystem() を使って確認することができる。
	///
	/// @param a 対象の連立方程式の係数行列
	/// @param b 対象の連立方程式の解ベクトルの先頭位置(計算結果の未知数がそのまま返される)
	/// @return なし
	template< typename T, class Ran >
	void operator()( SquareMatrix< T > a, Ran b );

	/// @brief ガウス・ジョルダン法により連立方程式の解を一度に求める
	///
	/// - b が未定義の場合、assert を呼び出す
	/// - 処理途中で係数行列の対角成分にゼロが見つかった場合、ゼロ除算が発生するため例外 std::runtime_error を投げる
	/// - 係数行列の列数と右辺(解行列)の行数が等しくない場合は例外 ExceptionNotEqualLength< SquareMatrix< T >::size_type > を投げる
	///
	/// @note 解が不定または不能の場合、理論上は処理途中で係数行列の対角成分にゼロが発生する。
	///       しかし、誤差により完全にゼロにはならず、処理が継続される場合もある。
	///       そのとき、解が不定ならば無数にある解の一つが得られ、不能の場合は正しくない結果となる。
	///       得られた結果が正しいかどうかは、 MathLib::VerifyLinearEquationSystem_Multi() を使って確認することができる。
	///
	/// @param a 対象の連立方程式の係数行列
	/// @param b 対象の連立方程式の解行列へのポインタ(計算結果の未知数がそのまま返される)
	/// @return なし
	template< typename T, class MATRIX >
	void multiple( SquareMatrix< T > a, MATRIX* b );
      };

      /**
	 @brief LU分解(ドゥーリトル法)による連立方程式の解法
      **/
      struct LUdecomposition_Doolittle
      {
	/// @brief LU分解(ドゥーリトル法)により連立方程式の解を求める
	///
	/// - 処理途中で係数行列の対角成分にゼロが見つかった場合、ゼロ除算が発生するため例外 std::runtime_error を投げる
	///
	/// @note 係数行列と右辺(解ベクトル)の大きさが等しいかチェックしていないことに注意
	/// @note 解が不定または不能の場合、理論上は処理途中で係数行列の対角成分にゼロが発生する。
	///       しかし、誤差により完全にゼロにはならず、処理が継続される場合もある。
	///       そのとき、解が不定ならば無数にある解の一つが得られ、不能の場合は正しくない結果となる。
	///       得られた結果が正しいかどうかは、 MathLib::VerifyLinearEquationSystem() を使って確認することができる。
	///
	/// @param a 対象の連立方程式の係数行列
	/// @param b 対象の連立方程式の解ベクトルの先頭位置(計算結果の未知数がそのまま返される)
	/// @return なし
	template< typename T, class Ran >
	void operator()( const SquareMatrix< T >& a, Ran b );

	/// @brief LU分解(ドゥーリトル法)により連立方程式の解を一度に求める
	///
	/// - b が未定義の場合、assert を呼び出す
	/// - 処理途中で係数行列の対角成分にゼロが見つかった場合、ゼロ除算が発生するため例外 std::runtime_error を投げる
	/// - 係数行列の列数と右辺(解行列)の行数が等しくない場合は例外 ExceptionNotEqualLength< SquareMatrix< T >::size_type > を投げる
	///
	/// @note 解が不定または不能の場合、理論上は処理途中で係数行列の対角成分にゼロが発生する。
	///       しかし、誤差により完全にゼロにはならず、処理が継続される場合もある。
	///       そのとき、解が不定ならば無数にある解の一つが得られ、不能の場合は正しくない結果となる。
	///       得られた結果が正しいかどうかは、 MathLib::VerifyLinearEquationSystem_Multi() を使って確認することができる。
	///
	/// @param a 対象の連立方程式の係数行列
	/// @param b 対象の連立方程式の解行列へのポインタ(計算結果の未知数がそのまま返される)
	/// @return なし
	template< typename T, class MATRIX >
	void multiple( const SquareMatrix< T >& a, MATRIX* b );
      };

      /**
	 @brief LU分解(クラウト法)による連立方程式の解法
      **/
      struct LUdecomposition_Crout
      {
	/// @brief LU分解(クラウト法)により連立方程式の解を求める
	///
	/// - 処理途中で係数行列の対角成分にゼロが見つかった場合、ゼロ除算が発生するため例外 std::runtime_error を投げる
	///
	/// @note 係数行列と右辺(解ベクトル)の大きさが等しいかチェックしていないことに注意
	/// @note 解が不定または不能の場合、理論上は処理途中で係数行列の対角成分にゼロが発生する。
	///       しかし、誤差により完全にゼロにはならず、処理が継続される場合もある。
	///       そのとき、解が不定ならば無数にある解の一つが得られ、不能の場合は正しくない結果となる。
	///       得られた結果が正しいかどうかは、 MathLib::VerifyLinearEquationSystem() を使って確認することができる。
	///
	/// @param a 対象の連立方程式の係数行列
	/// @param b 対象の連立方程式の解ベクトルの先頭位置(計算結果の未知数がそのまま返される)
	/// @return なし
	template< typename T, class Ran >
	void operator()( const SquareMatrix< T >& a, Ran b );

	/// @brief LU分解(クラウト法)により連立方程式の解を一度に求める
	///
	/// - b が未定義の場合、assert を呼び出す
	/// - 処理途中で係数行列の対角成分にゼロが見つかった場合、ゼロ除算が発生するため例外 std::runtime_error を投げる
	/// - 係数行列の列数と右辺(解行列)の行数が等しくない場合は例外 ExceptionNotEqualLength< SquareMatrix< T >::size_type > を投げる
	///
	/// @note 解が不定または不能の場合、理論上は処理途中で係数行列の対角成分にゼロが発生する。
	///       しかし、誤差により完全にゼロにはならず、処理が継続される場合もある。
	///       そのとき、解が不定ならば無数にある解の一つが得られ、不能の場合は正しくない結果となる。
	///       得られた結果が正しいかどうかは、 MathLib::VerifyLinearEquationSystem_Multi() を使って確認することができる。
	///
	/// @param a 対象の連立方程式の係数行列
	/// @param b 対象の連立方程式の解行列へのポインタ(計算結果の未知数がそのまま返される)
	/// @return なし
	template< typename T, class MATRIX >
	void multiple( const SquareMatrix< T >& a, MATRIX* b );
      };

    } // namespace solver

    namespace convergence_test
    {
      /**
	 @brief 差分の和と全体の和の比率による収束判定

	 反復処理において、前の値 p と今回の値 c の差分の絶対値 | c - p | と c のそれぞれに対して和をを求める。
	 一回の処理で得られる和 Δs = Σ| c - p | と S = Σc の比率 Δs / S を求め、
	 その値が指定したしきい値以下かどうかを判定する。但し、S = 0 であれば判定結果は常に true となる。

	 下記のように利用する。

	 @code
	 BySum< double > test( e ); // e はしきい値
	 std::vector< double > preVal( n ); // 前の値
	 : // preVal に初期値をセット
	 std::vector< double > curVal( n ); // 現在の値 

	 for (;;) {
	 : // curVal に値をセット
	 test.init();
	 for ( std::vector< double >::size_type i = 0 ; i < preVal.size() ; ++i )
	 test( preVal[i], curVal[i] );
	 if ( test.isConvergent() )
	 break;
	 }
	 @endcode

	 一回の反復処理で扱う値がただ一つであれば、ConvergenceTest_BySum と ConvergenceTest_ByMax の結果は一致する。
      **/
      template< class T >
      class BySum
      {
	T e_;    // 判定のためのしきい値
	T diff_; // 差分の和
	T sum_;  // 全体の和

      public:

	/// @brief しきい値を指定して構築
	///
	/// @param e 判定時のしきい値
	BySum( T e ) { init( e ); }

	/// @brief 求めた総和を初期化(ゼロクリア)する
	///
	/// @return なし
	void init()
	{ diff_ = sum_ = T(); }

	/// @brief しきい値を再設定し、求めた総和を初期化(ゼロクリア)する
	///
	/// @param e 新たなしきい値
	/// @return なし
	void init( T e )
	{
	  const std::string FUNC_NAME = "ConvergenceTest_BySum::init";

	  e_ = error_lib::PNum( e, FUNC_NAME + " : e " + error_message::PNUM );
	  init();
	}

	/// @brief 前後の値から差分と総和を求める
	///
	/// @param preVal 前の値
	/// @param newVal 現在の値
	/// @return なし
	void operator()( T preVal, T newVal );

	/// @brief 収束しているかを判定する
	///
	/// @return 差分の和と総和の比率が充分に小さければ(しきい値より小さければ) true を返す
	bool isConvergent()
	{ return( sum_ == 0 || ( diff_ / sum_ <= e_ ) ); }

	/// @brief 設定されているしきい値を返す
	///
	/// @return しきい値
	T e() const { return( e_ ); }
      };

      /**
	 @brief 差分と値の比率の最大値による収束判定

	 反復処理において、前の値 p と今回の値 c の差分 c - p と c との比率の絶対値 | ( c - p ) / c |
	 を求める ( 但し c = 0 ならば | c - p | をそのまま用いる )。
	 一回の処理で得られる全ての値の中からその最大値を抽出し、その値が指定したしきい値以下かどうかを判定する。

	 下記のように利用する。

	 @code
	 ByMax< double > test( e ); // e はしきい値
	 std::vector< double > preVal( n ); // 前の値
	 : // preVal に初期値をセット
	 std::vector< double > curVal( n ); // 現在の値 

	 for (;;) {
	 : // curVal に値をセット
	 test.init();
	 for ( std::vector< double >::size_type i = 0 ; i < preVal.size() ; ++i )
	 test( preVal[i], curVal[i] );
	 if ( test.isConvergent() )
	 break;
	 }
	 @endcode

	 一回の反復処理で扱う値がただ一つであれば、ConvergenceTest_BySum と ConvergenceTest_ByMax の結果は一致する。
      **/
      template< class T >
      class ByMax
      {
	T e_;    // 判定のためのしきい値
	T max_;  // 差分

      public:

	/// @brief しきい値を指定して構築
	///
	/// @param e 判定時のしきい値
	ByMax( T e ) { init( e ); }

	/// @brief 求めた最大値を初期化(ゼロクリア)する
	///
	/// @return なし
	void init()
	{ max_ = T(); }

	/// @brief しきい値を再設定し、求めた総和を初期化(ゼロクリア)する
	///
	/// @param e 新たなしきい値
	/// @return なし
	void init( T e )
	{
	  const std::string FUNC_NAME = "ConvergenceTest_ByMax::init";

	  e_ = error_lib::PNum( e, FUNC_NAME + " : e " + error_message::PNUM );
	  init();
	}

	/// @brief 前後の値から差分との比率の絶対値を求め、前回の最大値と比較してより大きければ入れ替える
	///
	/// @param preVal 前の値
	/// @param newVal 現在の値
	/// @return なし
	void operator()( T preVal, T newVal );

	/// @brief 収束しているかを判定する
	///
	/// @return 差分と値の比率の絶対値の最大値が充分に小さければ(しきい値より小さければ) true を返す
	bool isConvergent()
	{ return( max_ <= e_ ); }

	/// @brief 設定されているしきい値を返す
	///
	/// @return しきい値
	T e() const { return( e_ ); }
      };

    } // namespace convergence_test

    namespace solver
    {
      /**
	 @brief ヤコビ法による連立方程式の計算

	 テンプレート引数の T は係数や解の型を、Test は収束判定用関数オブジェクト
	 (BySum か ByMax の利用を想定。デフォルトは BySum)
	 をそれぞれ表す。
      **/
      template< class T, template< class > class Test = convergence_test::BySum >
      class Jacobi
      {
	Test< T > test_; // 収束判定
	uint32_t maxCnt_; // 収束しなかった場合の最大処理回数

      public:

	/// @brief 収束判定のしきい値と最大処理回数を指定して構築
	///
	/// @param e 収束判定のしきい値
	/// @param maxCnt 収束しなかった場合の最大処理回数
	Jacobi( T e, uint32_t maxCnt )
	  : test_( e ), maxCnt_( maxCnt ) {}

	/// @brief 連立方程式を解く
	///
	/// - 計算回数が最大処理回数 maxCnt を超えた場合は例外 ExceptionExcessLimit を投げる。
	///   但し、それまでに得られた近似解は連立方程式の解ベクトル b に返される。
	/// - 係数行列の対角成分にゼロが含まれる場合、std::runtime_error を投げる。
	///
	/// @note 係数行列と右辺(解ベクトル)の大きさが等しいかチェックしていないことに注意
	///
	/// @todo 対角成分にゼロが含まれる場合、std::runtime_error を投げるが、実際には式の入れ替えで
	///       解ける場合がある。対角成分が大きくなるほど速く収束するため、そのための処理を実装したほうがよい。
	///
	/// @param a 連立方程式の係数行列への参照
	/// @param b 対象の連立方程式の解ベクトルの先頭位置(計算結果の未知数がそのまま返される)
	/// @return なし
	template< class Ran >
	void operator()( const SquareMatrix< T >& a, Ran b );
      };

      /**
	 @brief ガウス=ザイデル法による連立方程式の計算

	 テンプレート引数の T は係数や解の型を、Test は収束判定用関数オブジェクト
	 (BySum か ByMax の利用を想定。デフォルトは BySum)
	 をそれぞれ表す。
      **/
      template< class T, template< class > class Test = convergence_test::BySum >
      class GaussSeidel
      {
	Test< T > test_; // 収束判定
	uint32_t maxCnt_; // 収束しなかった場合の最大処理回数

      public:

	/// @brief 収束判定のしきい値と最大処理回数を指定して構築
	///
	/// @param e 収束判定のしきい値
	/// @param maxCnt 収束しなかった場合の最大処理回数
	GaussSeidel( T e, uint32_t maxCnt )
	  : test_( e ), maxCnt_( maxCnt ) {}

	/// @brief 連立方程式を解く
	///
	/// - 計算回数が最大処理回数 maxCnt を超えた場合は例外 ExceptionExcessLimit を投げる。
	///   但し、それまでに得られた近似解は連立方程式の解ベクトル b に返される。
	/// - 係数行列の対角成分にゼロが含まれる場合、std::runtime_error を投げる。
	///
	/// @note 係数行列と右辺(解ベクトル)の大きさが等しいかチェックしていないことに注意
	///
	/// @todo 対角成分にゼロが含まれる場合、std::runtime_error を投げるが、実際には式の入れ替えで
	///       解ける場合がある。対角成分が大きくなるほど速く収束するため、そのための処理を実装したほうがよい。
	///
	/// @param a 連立方程式の係数行列への参照
	/// @param b 対象の連立方程式の解ベクトルの先頭位置(計算結果の未知数がそのまま返される)
	/// @return なし
	template< class Ran >
	void operator()( const SquareMatrix< T >& a, Ran b );
      };

      /**
	 @brief ヤコビ法に SOR 法を適用した連立方程式の計算

	 緩和係数により収束するスピードを早くすることができる。1 のときは通常のヤコビ法と等しい。

	 テンプレート引数の T は係数や解の型を、Test は収束判定用関数オブジェクト
	 (BySum か ByMax の利用を想定。デフォルトは BySum)
	 をそれぞれ表す。
      **/
      template< class T, template< class > class Test = convergence_test::BySum >
      class Jacobi_SOR
      {
	T w_;             // 緩和係数
	Test< T > test_;  // 収束判定
	uint32_t maxCnt_; // 収束しなかった場合の最大処理回数

      public:

	/// @brief 緩和係数・収束判定のしきい値・最大処理回数を指定して構築
	///
	/// @param w 緩和係数
	/// @param e 収束判定のしきい値
	/// @param maxCnt 収束しなかった場合の最大処理回数
	Jacobi_SOR( T w, T e, uint32_t maxCnt )
	  : w_( w ), test_( e ), maxCnt_( maxCnt ) {}

	/// @brief 連立方程式を解く
	///
	/// - 計算回数が最大処理回数 maxCnt を超えた場合は例外 ExceptionExcessLimit を投げる。
	///   但し、それまでに得られた近似解は連立方程式の解ベクトル b に返される。
	/// - 係数行列の対角成分にゼロが含まれる場合、std::runtime_error を投げる。
	///
	/// @note 係数行列と右辺(解ベクトル)の大きさが等しいかチェックしていないことに注意
	///
	/// @todo 対角成分にゼロが含まれる場合、std::runtime_error を投げるが、実際には式の入れ替えで
	///       解ける場合がある。対角成分が大きくなるほど速く収束するため、そのための処理を実装したほうがよい。
	///
	/// @param a 連立方程式の係数行列への参照
	/// @param b 対象の連立方程式の解ベクトルの先頭位置(計算結果の未知数がそのまま返される)
	/// @return なし
	template< class Ran >
	void operator()( const SquareMatrix< T >& a, Ran b );
      };

      /**
	 @brief ガウス=ザイデル法に SOR 法を適用した連立方程式の計算

	 緩和係数により収束するスピードを早くすることができる。1 のときは通常のガウス=ザイデル法と等しい。

	 テンプレート引数の T は係数や解の型を、Test は収束判定用関数オブジェクト
	 (BySum か ByMax の利用を想定。デフォルトは BySum)
	 をそれぞれ表す。
      **/
      template< class T, template< class > class Test = convergence_test::BySum >
      class GaussSeidel_SOR
      {
	T w_;             // 緩和係数
	Test< T > test_;  // 収束判定
	uint32_t maxCnt_; // 収束しなかった場合の最大処理回数

      public:

	/// @brief 緩和係数・収束判定のしきい値・最大処理回数を指定して構築
	///
	/// @param w 緩和係数
	/// @param e 収束判定のしきい値
	/// @param maxCnt 収束しなかった場合の最大処理回数
	GaussSeidel_SOR( T w, T e, uint32_t maxCnt )
	  : w_( w ), test_( e ), maxCnt_( maxCnt ) {}

	/// @brief 連立方程式を解く
	///
	/// - 計算回数が最大処理回数 maxCnt を超えた場合は例外 ExceptionExcessLimit を投げる。
	///   但し、それまでに得られた近似解は連立方程式の解ベクトル b に返される。
	/// - 係数行列の対角成分にゼロが含まれる場合、std::runtime_error を投げる。
	///
	/// @note 係数行列と右辺(解ベクトル)の大きさが等しいかチェックしていないことに注意
	///
	/// @todo 対角成分にゼロが含まれる場合、std::runtime_error を投げるが、実際には式の入れ替えで
	///       解ける場合がある。対角成分が大きくなるほど速く収束するため、そのための処理を実装したほうがよい。
	///
	/// @param a 連立方程式の係数行列への参照
	/// @param b 対象の連立方程式の解ベクトルの先頭位置(計算結果の未知数がそのまま返される)
	/// @return なし
	template< class Ran >
	void operator()( const SquareMatrix< T >& a, Ran b );
      };

      /**
	 @brief 共役勾配法 ( Conjugate Gradient Method ) による連立方程式の計算
      **/
      template< class T >
      class CGM_LinearEquation
      {
	T e_;      // 収束判定のしきい値

      public:

	/// @brief 収束判定のしきい値を指定して構築
	///
	/// @param e 収束判定のしきい値(前回求めた解との差の絶対値が全てこの値以下になったら処理を終了する)
	CGM_LinearEquation( T e )
	  : e_( e ) {}

	/// @brief 連立方程式を解く
	///
	/// @note 係数行列と右辺(解ベクトル)の大きさが等しいかチェックしていないことに注意
	/// @note 解 y が不定の場合、得られる解は最小ノルム解 ( |y| が最小の解 ) になる。
	///       また、解 y が不能の場合、得られる解は最小二乗解 ( |ay - x| が最小の解 ) になる。
	///
	/// @param a 連立方程式の係数行列への参照
	/// @param x 対象の連立方程式の解ベクトルの先頭位置(計算結果の未知数がそのまま返される)
	///          反復処理の開始時には元の x の値が初期値として利用されることに注意
	/// @return なし
	template< class Ran >
	void operator()( const SquareMatrix< T >& a, Ran x );
      };

      /**
	 @brief 連立方程式を解いた結果が正しいかを検証する
      **/
      template< class T >
      class Verify
      {
	T threshold_; // 一致判定のしきい値

      public:

	/// @brief 一致判定のためのしきい値を指定して構築
	///
	/// @param threshold 一致判定のためのしきい値
	explicit Verify( T threshold )
	  : threshold_( threshold ) {}

	/// @brief 連立方程式を解いた結果が正しいかを検証する
	///
	/// 求めた結果を連立方程式に代入し、左辺と右辺が一致するかを確認する。
	/// 全方程式について一致すれば true を返し、一つでも一致しない式があれば false を返す。
	///
	/// @note 係数行列 a と右辺(解ベクトル) b、及び解いた結果 ans の大きさが全て等しいかチェックしていないことに注意
	///
	/// @param a 対象の連立方程式の係数行列
	/// @param b 対象の連立方程式の解ベクトルの先頭位置
	/// @param ans 連立方程式を解いた結果
	/// @return 不一致と判定された場合は false を返す
	template< class In1, class In2 >
	bool operator()( const SquareMatrix< T >& a, In1 b, In2 ans );

	/// @brief 複数の連立方程式に対して解いた結果が正しいかを一度に検証する
	///
	/// 全ての連立方程式について一致すれば true を返し、一つでも一致しない連立方程式があれば false を返す。
	///
	/// - 係数行列 a の列数と右辺(解行列) b または解いた結果 ans の行数が等しくない場合は例外 ExceptionNotEqualLength< SquareMatrix< T >::size_type > を投げる
	/// - 右辺(解行列)の列数と解いた結果の列数が等しくない場合は例外 ExceptionNotEqualLength< MATRIX1::size_type > を投げる
	///
	/// @param a 対象の連立方程式の係数行列
	/// @param b 対象の連立方程式の解行列
	/// @param ans 連立方程式を解いた結果(bと同じ大きさの行列)
	/// @return 不一致と判定された場合は false を返す
	template< class MATRIX1, class MATRIX2 >
	bool multiple( const SquareMatrix< T >& a, const MATRIX1& b, const MATRIX2& ans );
      };

    } // namespace solver

  } // namespace math_lib

} // namespace tadah_fussy

#include "equation_inline.h"

#endif
