/**
   @file equation_inline.h
   @brief 連立方程式用インライン・テンプレート関数

   @author tadah_fussy
   @date 2015/11/15 コメント追加
   @date 2017/04/09 連立方程式計算関数の引数変更
   @date 2019/03/27 コメントの見直し
   @date 2019/04/04 VerifyLinearEquationSystem, VerifyLinearEquationSystem_Multi の追加と若干の修正、コメントの見直し
   @date 2019/04/07 Matrix::clone の廃止に伴う変更
   @date 2019/04/16 error_library.h の見直しに伴う変更。LinearEquationSystem の廃止。
   @date 2019/04/18 エラーメッセージの定数化。ConjugateGradientMethod の計算最適化と収束条件の見直し。
   @date 2019/04/24 解行列を解く・チェックする関数のエラー処理の修正
   @date 2019/04/27 全ての解法について関数オブジェクトの形式に変更
   @date 2021/03/14 名前空間の変更
   @date 2021/04/02 文字コードを UTF-8 に変更
**/

/**
   @brief 作成者 tadah_fussy
**/
namespace tadah_fussy
{
  /***************************************************************
    連立方程式の計算
  ***************************************************************/

  namespace error_message
  {
    /// @brief 連立方程式が解けなかった場合 ( 不定・不能 ) のエラー・メッセージ
    const std::string UNSOLVED_EQUATION = "unsolved equation";
  }

  namespace math_lib
  {
    namespace solver
    {
      /*
	VectorOperator : 連立方程式計算用関数群 ( 解ベクトル専用 )
      */
      template< typename T, class Ran >
      struct VectorOperator
      {
	/*
	  swap : 方程式の入れ替え
	*/
	void swap( SquareMatrix< T >* a, Ran b,
		   typename SquareMatrix< T >::size_type e1, typename SquareMatrix< T >::size_type e2 )
	{
	  matrix_lib::SwapRow( a, e1, e2 );
	  std::swap( b[e1], b[e2] );
	}

	/*
	  sub : src 番目の方程式を val 倍して dest 番目の方程式から引く
	*/
	void sub( SquareMatrix< T >* a, Ran b,
		  typename SquareMatrix< T >::size_type src, T val, typename SquareMatrix< T >::size_type dest )
	{
	  auto sRow = (*a)[src];
	  auto dRow = (*a)[dest];

	  while ( sRow != sRow.end() )
	    *dRow++ -= *sRow++ * val;

	  b[dest] -= b[src] * val;
	}

	/*
	  div : r 番目の方程式を val で割る
	*/
	void div( SquareMatrix< T >* a, Ran b,
		  typename SquareMatrix< T >::size_type r, typename SquareMatrix< T >::value_type val )
	{
	  auto row = (*a)[r];
	  std::transform( row, row.end(), row, std::bind( std::divides< T >(), std::placeholders::_1, val ) );
	  b[r] /= val;
	}

	/*
	  forwardSub : 上側から解を求める ( 下三角行列用 )
	*/
	bool forwardSub( const SquareMatrix< T >& a, Ran b )
	{
	  using size_type = typename SquareMatrix< T >::size_type;

	  for ( size_type i = 0 ; i < a.size() ; ++i ) {
	    if ( a[i][i] == 0 )
	      return( false );
	    for ( size_type j = 0 ; j < i ; ++j )
	      b[i] -= a[i][j] * b[j];
	    b[i] /= a[i][i];
	  }

	  return( true );
	}

	/*
	  backSub : 下側から解を求める ( 上三角行列用 )
	*/
	bool backSub( const SquareMatrix< T >& a, Ran b )
	{
	  using size_type = typename SquareMatrix< T >::size_type;

	  for ( size_type i = a.size() ; i > 0 ; --i ) {
	    if ( a[i - 1][i - 1] == 0 )
	      return( false );
	    for ( size_type j = i ; j < a.size() ; ++j )
	      b[i - 1] -= b[j] * a[i - 1][j];
	    b[i - 1] /= a[i - 1][i - 1];
	  }

	  return( true );
	}

	/*
	  rearrange : 解を並べ直す ( Doolittle 専用 )
	*/
	void rearrange( Ran b, const std::vector< std::pair< size_t, size_t > >& index )
	{
	  using size_type = std::vector< std::pair< size_t, size_t > >::size_type;

	  for ( size_type i = index.size() ; i > 0 ; --i )
	    std::swap( b[index[i - 1].first], b[index[i - 1].second] );
	}
      };

      /*
	MatrixOperator : 連立方程式計算用関数群 ( 解行列専用 )
      */
      template< typename T, class MATRIX >
      struct MatrixOperator
      {
	/*
	  swap : 方程式の入れ替え
	*/
	void swap( SquareMatrix< T >* a, MATRIX* b,
		   typename SquareMatrix< T >::size_type e1, typename SquareMatrix< T >::size_type e2 )
	{
	  matrix_lib::SwapRow( a, e1, e2 );
	  matrix_lib::SwapRow( b, e1, e2 );
	}

	/*
	  sub : src 番目の方程式を val 倍して dest 番目の方程式から引く
	*/
	void sub( SquareMatrix< T >* a, MATRIX* b,
		  typename SquareMatrix< T >::size_type src, T val, typename SquareMatrix< T >::size_type dest )
	{
	  auto sRow = (*a)[src];
	  auto dRow = (*a)[dest];

	  while ( sRow != sRow.end() )
	    *dRow++ -= *sRow++ * val;

	  auto sRow2 = (*b)[src];
	  auto dRow2 = (*b)[dest];

	  while ( sRow2 != sRow2.end() )
	    *dRow2++ -= *sRow2++ * val;
	}

	/*
	  div : r 番目の方程式を val で割る
	*/
	void div( SquareMatrix< T >* a, MATRIX* b,
		  typename SquareMatrix< T >::size_type r, typename SquareMatrix< T >::value_type val )
	{
	  using value_type = typename MATRIX::value_type;

	  auto row = (*a)[r];
	  std::transform( row, row.end(), row,
			  std::bind( std::divides< T >(), std::placeholders::_1, val ) );

	  auto row2 = (*b)[r];
	  std::transform( row2, row2.end(), row2,
			  std::bind( std::divides< value_type >(), std::placeholders::_1, val ) );
	}

	/*
	  forwardSub : 上側から解を求める ( 下三角行列用 )
	*/
	bool forwardSub( const SquareMatrix< T >& a, MATRIX* b )
	{
	  using size_type = typename SquareMatrix< T >::size_type;

	  for ( size_type i = 0 ; i < a.size() ; ++i ) {
	    if ( a[i][i] == 0 )
	      return( false );
	    for ( size_type c = 0 ; c < b->cols() ; ++c ) {
	      for ( size_type j = 0 ; j < i ; ++j )
		(*b)[i][c] -= a[i][j] * (*b)[j][c];
	      (*b)[i][c] /= a[i][i];
	    }
	  }

	  return( true );
	}

	/*
	  backSub : 下側から解を求める ( 上三角行列用 )
	*/
	bool backSub( const SquareMatrix< T >& a, MATRIX* b )
	{
	  using size_type = typename SquareMatrix< T >::size_type;

	  for ( size_type i = a.size() ; i > 0 ; --i ) {
	    if ( a[i - 1][i - 1] == 0 )
	      return( false );
	    for ( size_type c = 0 ; c < b->cols() ; ++c ) {
	      for ( size_type j = i ; j < a.size() ; ++j )
		(*b)[i - 1][c] -= (*b)[j][c] * a[i - 1][j];
	      (*b)[i - 1][c] /= a[i - 1][i - 1];
	    }
	  }

	  return( true );
	}

	/*
	  rearrange : 解を並べ直す ( Doolittle 専用 )
	*/
	void rearrange( MATRIX* b, const std::vector< std::pair< size_t, size_t > >& index )
	{
	  using size_type = std::vector< std::pair< size_t, size_t > >::size_type;

	  for ( size_type i = index.size() ; i > 0 ; --i )
	    matrix_lib::SwapRow( b, index[i - 1].first, index[i - 1].second );
	}
      };

      /*
	Normalize : 正規化した係数を返す

	正規化する ( r, c ) の位置の係数から行方向に末尾までたどって最大の値(スケール因子;scaled factor)を探し、
	( r, c ) の位置の係数をその値で除算した結果を絶対値で返す。
      */
      template< typename T >
      T Normalize( const SquareMatrix< T >& a, typename SquareMatrix< T >::size_type r, typename SquareMatrix< T >::size_type c )
      {
	using size_type = typename SquareMatrix< T >::size_type;

	T coef = std::abs( a[r][c] ); // 正規化対象の係数

	T max = coef; // 現在の最大値を正規化対象の係数で初期化
	for ( size_type i = c + 1 ; i < a.size() ; ++i ) {
	  T t = std::abs( a[r][i] );
	  if ( max < t ) max = t;
	}

	return( ( max > 0 ) ? coef / max : T() );
      }

      /*
	Pivot : 部分ピボットを行う(正規化処理付き)

	i 行目の対角成分より列方向下側の係数の中で最大値を求め、最大値を持つ行と入れ替える。

	戻り値 : i 行目と入れ替えした行の番号
      */
      template< typename T >
      typename SquareMatrix< T >::size_type
      Pivot( const SquareMatrix< T >& a, typename SquareMatrix< T >::size_type i )
      {
	using size_type = typename SquareMatrix< T >::size_type;

	size_type maxIdx = i; // 最大値を持つ行をピボット対象の行番号で初期化
	T max = Normalize( a, maxIdx, i ); // 現在の最大値

	for ( size_type r = i + 1 ; r < a.size() ; ++r ) {
	  T t = Normalize( a, r, i );
	  if ( max < t ) {
	    maxIdx = r;
	    max = t;
	  }
	}

	return( maxIdx );
      }

      /*
	GaussianElimination_ : ガウスの消去法により連立方程式の解を求める

	解が不定または不能の場合は false を返す

	a : 対象の連立方程式の係数行列
	b : 対象の連立方程式の解ベクトル・行列の先頭位置(計算結果の未知数がそのまま返される)
	les : 連立方程式計算用関数群
      */
      template< typename T, class Ran, class LES >
      bool GaussianElimination_( SquareMatrix< T > a, Ran b, LES les )
      {
	using size_type = typename SquareMatrix< T >::size_type;

	// 前進消去
	for ( size_type i = 0 ; i < a.size() ; ++i ) {
	  size_type swapped = Pivot( a, i );
	  if ( swapped != i )
	    les.swap( &a, b, swapped, i );
	  T k = a[i][i];
	  if ( k == 0 ) // 最大係数が 0 の場合は不定または不能となるため false を返す
	    return( false );
	  les.div( &a, b, i, k );
	  for ( size_type r = i + 1 ; r < a.size() ; ++r )
	    les.sub( &a, b, i, a[r][i], r );
	}

	// 後退代入
	return( les.backSub( a, b ) );
      }

      /*
	GaussianElimination::operator() : ガウスの消去法により連立方程式 ax = b の解を求める
      */
      template< typename T, class Ran >
      void GaussianElimination::operator()( SquareMatrix< T > a, Ran b )
      {
	const std::string FUNC_NAME = "GaussianElimination::operator()";

	if ( ! GaussianElimination_( a, b, VectorOperator< T, Ran >() ) )
	  throw std::runtime_error( FUNC_NAME + " : " + error_message::UNSOLVED_EQUATION );
      }

      /*
	GaussianElimination::multiple : ガウスの消去法により連立方程式 ax = b の解を一度に求める
      */
      template< typename T, class MATRIX >
      void GaussianElimination::multiple( SquareMatrix< T > a, MATRIX* b )
      {
	const std::string FUNC_NAME = "GaussianElimination::multiple";

	assert( b != nullptr );

	error_lib::CheckMultMatrix( a, *b, FUNC_NAME + " : a.cols() " + error_message::NOT_EQUAL_LENGTH + " b->rows()" );
	if ( ! GaussianElimination_( a, b, MatrixOperator< T, MATRIX >() ) )
	  throw std::runtime_error( FUNC_NAME + " : " + error_message::UNSOLVED_EQUATION );
      }

      /*
	GaussJordanElimination_ : ガウス・ジョルダン法により連立方程式の解を求める

	解が不定または不能の場合は false を返す

	a : 対象の連立方程式の係数行列
	b : 対象の連立方程式の解ベクトル・行列の先頭位置(計算結果の未知数がそのまま返される)
	les : 連立方程式計算用関数群
      */
      template< typename T, class Ran, class LES >
      bool GaussJordanElimination_( SquareMatrix< T > a, Ran b, LES les )
      {
	using size_type = typename SquareMatrix< T >::size_type;

	for ( size_type i = 0 ; i < a.size() ; ++i ) {
	  size_type swapped = Pivot( a, i );
	  if ( swapped != i )
	    les.swap( &a, b, swapped, i );
	  T k = a[i][i];
	  if ( k == 0 ) // 最大係数が 0の場合は不定または不能となるため false を返す
	    return( false );
	  les.div( &a, b, i, k );
	  for ( size_type r = 0 ; r < a.size() ; ++r ) {
	    if ( r == i ) continue;
	    les.sub( &a, b, i, a[r][i], r );
	  }
	}

	return( true );
      }

      /*
	GaussJordanElimination::operator() : ガウス・ジョルダン法により連立方程式 ax = b の解を求める
      */
      template< typename T, class Ran >
      void GaussJordanElimination::operator()( SquareMatrix< T > a, Ran b )
      {
	const std::string FUNC_NAME = "GaussJordanElimination::operator()";

	if ( ! GaussJordanElimination_( a, b, VectorOperator< T, Ran >() ) )
	  throw std::runtime_error( FUNC_NAME + " : " + error_message::UNSOLVED_EQUATION );
      }

      /*
	GaussJordanElimination::multiple : ガウス・ジョルダン法により連立方程式の解を一度に求める
      */
      template< typename T, class MATRIX >
      void GaussJordanElimination::multiple( SquareMatrix< T > a, MATRIX* b )
      {
	const std::string FUNC_NAME = "GaussJordanElimination::multiple";

	assert( b != nullptr );

	error_lib::CheckMultMatrix( a, *b, FUNC_NAME + " : a.cols() " + error_message::NOT_EQUAL_LENGTH + " b->rows()." );
	if ( ! GaussJordanElimination_( a, b, MatrixOperator< T, MATRIX >() ) )
	  throw std::runtime_error( FUNC_NAME + " : " + error_message::UNSOLVED_EQUATION );
      }

      /*
	PivotRow : 行 i に対して部分ピボットを行う

	戻り値 : i 行目と入れ替えした行の番号
      */
      template< typename T >
      typename Matrix< T >::size_type
      PivotRow( SquareMatrix< T >* a, typename SquareMatrix< T >::size_type i )
      {
	using size_type = typename SquareMatrix< T >::size_type;

	size_type maxIdx = i;
	T max = std::abs( (*a)[maxIdx][i] );

	for ( size_type r = i + 1 ; r < a->size() ; ++r ) {
	  T t = std::abs( (*a)[r][i] );
	  if ( max < t ) {
	    maxIdx = r;
	    max = t;
	  }
	}
	if ( maxIdx != i ) matrix_lib::SwapRow( a, maxIdx, i );

	return( maxIdx );
      }

      /*
	PivotCol : 列 i に対して部分ピボットを行う

	戻り値 : i 列目と入れ替えした列の番号
      */
      template< typename T >
      typename Matrix< T >::size_type
      PivotCol( SquareMatrix< T >* a, typename SquareMatrix< T >::size_type i )
      {
	using size_type = typename SquareMatrix< T >::size_type;

	size_type maxIdx = i;
	T max = std::abs( (*a)[i][maxIdx] );

	for ( size_type c = i + 1 ; c < a->size() ; ++c ) {
	  T t = std::abs( (*a)[i][c] );
	  if ( max < t ) {
	    maxIdx = c;
	    max = t;
	  }
	}
	if ( maxIdx != i ) matrix_lib::SwapCol( a, maxIdx, i );

	return( maxIdx );
      }

      /*
	Doolittle : LU分解 ドゥーリトル法

	行列aがLU分解できない場合(処理途中で対角成分にゼロが見つかった場合)は例外 std::runtime_error を投げる
	精度向上のため未知数の入れ替え(ピボット選択)を行なっている。よって、行列のLU分解の用途には利用できない。

	a : LU分解する対象の行列
	l : 分解した下三角行列 L
	u : 分解した上三角行列 U
	index 処理前に対して処理後の未知数を入れ替えた番号
      */
      template< typename T >
      void Doolittle( SquareMatrix< T > a,
		      SquareMatrix< T >* l, SquareMatrix< T >* u,
		      std::vector< std::pair< size_t, size_t > >* index )
      {
	using size_type = typename SquareMatrix< T >::size_type;

	const std::string FUNC_NAME = "Doolittle";

	size_type sz = a.size();
	l->assign( sz );
	u->assign( sz );

	for ( size_type i = 0 ; i < sz ; ++i ) {
	  (*l)[i][i] = 1;         // Lの対角成分
	  (*u)[0][i] = a[0][i];   // Uの一行目
	}

	for ( size_type i = 1 ; i < sz ; ++i ) {
	  size_type swapped = PivotCol( u, i - 1 );
	  if ( swapped != i - 1 ) {
	    matrix_lib::SwapCol( &a, swapped, i - 1 );
	    index->push_back( std::pair< size_t, size_t >( i - 1, swapped ) );
	  }
	  if ( (*u)[i - 1][i - 1] == 0 )
	    throw std::runtime_error( FUNC_NAME + " : " + error_message::UNSOLVED_EQUATION );
	  for ( size_type j = i ; j < sz ; ++j ) {
	    // L成分の算出
	    (*l)[j][i - 1] = a[j][i - 1];
	    for ( size_type k = 0 ; k < i - 1 ; ++k )
	      (*l)[j][i - 1] -= (*l)[j][k] * (*u)[k][i - 1];
	    (*l)[j][i - 1] /= (*u)[i - 1][i - 1];

	    // U成分の算出
	    (*u)[i][j] = a[i][j];
	    for ( size_type k = 0 ; k < i ; ++k )
	      (*u)[i][j] -= (*l)[i][k] * (*u)[k][j];
	  }
	}
      }

      /*
	Crout : LU分解 クラウト法

	行列aがLU分解できない場合(処理途中で対角成分にゼロが見つかった場合)は例外 std::runtime_error を投げる
	精度向上のため式の入れ替え(ピボット選択)を行なっている。よって、行列のLU分解の用途には利用できない。

	a : LU分解する対象の行列
	b : 解ベクトル(右辺の値)
	l : 分解した下三角行列 L
	u : 分解した上三角行列 U
	les : 連立方程式計算用関数群
      */
      template< typename T, class Ran, class LES >
      void Crout( SquareMatrix< T > a, Ran b,
		  SquareMatrix< T >* l, SquareMatrix< T >* u, LES les )
      {
	using size_type = typename SquareMatrix< T >::size_type;

	const std::string FUNC_NAME = "Crout";

	size_type sz = a.size();
	l->assign( sz );
	u->assign( sz );

	for ( size_type i = 0 ; i < sz ; ++i ) {
	  (*u)[i][i] = 1;         // Uの対角成分
	  (*l)[i][0] = a[i][0];   // Lの一列目
	}

	for ( size_type i = 1 ; i < sz ; ++i ) {
	  size_type swapped = PivotRow( l, i - 1 );
	  if ( swapped != i - 1 )
	    les.swap( &a, b, swapped, i - 1 );

	  if ( (*l)[i - 1][i - 1] == 0 )
	    throw std::runtime_error( FUNC_NAME + " : " + error_message::UNSOLVED_EQUATION );
	  for ( size_type j = i ; j < sz ; ++j ) {
	    // U成分の算出
	    (*u)[i - 1][j] = a[i - 1][j];
	    for ( size_type k = 0 ; k < i - 1 ; ++k )
	      (*u)[i - 1][j] -= (*l)[i - 1][k] * (*u)[k][j];
	    (*u)[i - 1][j] /= (*l)[i - 1][i - 1];

	    // L成分の算出
	    (*l)[j][i] = a[j][i];
	    for ( size_type k = 0 ; k < i ; ++k )
	      (*l)[j][i] -= (*l)[j][k] * (*u)[k][i];
	  }
	}
      }

      /*
	LUdecomposition_Doolittle_ : LU分解(ドゥーリトル法)を利用した連立方程式の計算

	行列aがLU分解できない場合(処理途中で対角成分にゼロが見つかった場合)は false を返す

	a : 対象の連立方程式の係数行列
	b : 対象の連立方程式の解ベクトル・行列の先頭位置(計算結果の未知数がそのまま返される)
	les : 連立方程式計算用関数群
      */
      template< class T, class Ran, class LES >
      bool LUdecomposition_Doolittle_( const SquareMatrix< T >& a, Ran b, LES les )
      {
	using size_type = typename SquareMatrix< T >::size_type;

	SquareMatrix< T > l, u;
	std::vector< std::pair< size_t, size_t > > index;
	Doolittle( a, &l, &u, &index );
	if ( ! les.forwardSub( l, b ) )
	  return( false );
	if ( ! les.backSub( u, b ) )
	  return( false );
	les.rearrange( b, index );

	return( true );
      }

      /*
	LUdecomposition_Doolittle::operator() : LU分解(ドゥーリトル法)により連立方程式 ax = b の解を求める
      */
      template< class T, class Ran >
      void LUdecomposition_Doolittle::operator()( const SquareMatrix< T >& a, Ran b )
      {
	const std::string FUNC_NAME = "LUdecomposition_Doolittle::operator()";

	if ( ! LUdecomposition_Doolittle_( a, b, VectorOperator< T, Ran >() ) )
	  throw std::runtime_error( FUNC_NAME + " : " + error_message::UNSOLVED_EQUATION );
      }

      /*
	LUdecomposition_Doolittle::multiple : LU分解(ドゥーリトル法)により連立方程式 ax = b の解を一度に求める
      */
      template< class T, class MATRIX >
      void LUdecomposition_Doolittle::multiple( const SquareMatrix< T >& a, MATRIX* b )
      {
	const std::string FUNC_NAME = "LUdecomposition_Doolittle::multiple";

	assert( b != nullptr );

	error_lib::CheckMultMatrix( a, *b, FUNC_NAME + " : a.cols() " + error_message::NOT_EQUAL_LENGTH + " b->rows()." );
	if ( ! LUdecomposition_Doolittle_( a, b, MatrixOperator< T, MATRIX >() ) )
	  throw std::runtime_error( FUNC_NAME + " : " + error_message::UNSOLVED_EQUATION );
      }

      /*
	LUdecomposition_Crout_ : LU分解(クラウト法)を利用した連立方程式の計算

	行列aがLU分解できない場合(処理途中で対角成分にゼロが見つかった場合)は false を返す

	a : 対象の連立方程式の係数行列
	b : 対象の連立方程式の解ベクトル・行列の先頭位置(計算結果の未知数がそのまま返される)
	les : 連立方程式計算用関数群
      */
      template< class T, class Ran, class LES >
      bool LUdecomposition_Crout_( const SquareMatrix< T >& a, Ran b, LES les )
      {
	using size_type = typename SquareMatrix< T >::size_type;

	SquareMatrix< T > l, u;
	Crout( a, b, &l, &u, les );
	if ( ! les.forwardSub( l, b ) )
	  return( false );
	return( les.backSub( u, b ) );
      }

      /*
	LUdecomposition_Crout::operator() : LU分解(クラウト法)により連立方程式 ax = b の解を求める
      */
      template< class T, class Ran >
      void LUdecomposition_Crout::operator()( const SquareMatrix< T >& a, Ran b )
      {
	const std::string FUNC_NAME = "LUdecomposition_Crout::operator()";

	if ( ! LUdecomposition_Crout_( a, b, VectorOperator< T, Ran >() ) )
	  throw std::runtime_error( FUNC_NAME + " : " + error_message::UNSOLVED_EQUATION );
      }

      /*
	LUdecomposition_Crout::multiple : LU分解(クラウト法)により連立方程式の解を一度に求める
      */
      template< class T, class MATRIX >
      void LUdecomposition_Crout::multiple( const SquareMatrix< T >& a, MATRIX* b )
      {
	const std::string FUNC_NAME = "LUdecomposition_Crout::multiple";

	assert( b != nullptr );

	error_lib::CheckMultMatrix( a, *b, FUNC_NAME + " : a.cols() " + error_message::NOT_EQUAL_LENGTH + " b->rows()." );
	if ( ! LUdecomposition_Crout_( a, b, MatrixOperator< T, MATRIX >() ) )
	  throw std::runtime_error( FUNC_NAME + " : " + error_message::UNSOLVED_EQUATION );
      }

    } // namespace solver

  } // namespace math_lib

  namespace matrix_lib
  {
    /// @brief 逆行列を求める
    ///
    /// 内部で連立方程式を解いて逆行列を求めている。連立方程式の解法は選択が可能。
    /// 例えば、GaussianElimination を利用する場合は次のように記述する。
    ///
    /// @code
    /// math_lib::matrix::Inverse( mat, &inv, GaussianElimination() );
    /// @endcode
    ///
    /// - inv が未定義の場合、assert を呼び出す
    /// - 連立方程式の計算処理の過程で例外 std::runtime_error が投げられる場合がある
    ///
    /// @note mat が正則行列でない場合、理論上は処理途中で 例外 std::runtime_error が投げられる。
    ///       しかし、誤差により処理が継続される場合もあり、正しい結果が得られない可能性があることに注意。
    ///
    /// @param mat 対象の行列
    /// @param inv 求めた逆行列を書き込む正方行列へのポインタ
    /// @param solver 連立方程式計算用ソルバ
    /// @return なし
    template< class MATRIX, class Solver >
    void Inverse( const MATRIX& mat, SquareMatrix< typename MATRIX::value_type >* inv, Solver solver )
    {
      using size_type = typename MATRIX::size_type;
      using value_type = typename MATRIX::value_type;

      assert( inv != nullptr );

      size_type sz = mat.size();
      inv->assign( sz, value_type() );
      for ( size_type r = 0 ; r < sz ; ++r )
	(*inv)[r][r] = 1;

      SquareMatrix< value_type > a( mat );

      solver.multiple( a, inv );
    }

  } // namespace matrix_lib

  namespace math_lib
  {
    namespace convergence_test
    {
      /*
	BySum< T >::operator() :
	前の値 preVal と新たな値 newVal に対し、その差分の絶対値と newVal の絶対値を積算していく
      */
      template< class T >
      void BySum< T >::operator()( T preVal, T newVal )
      {
	diff_ += std::abs( preVal - newVal );
	sum_ += std::abs( newVal );
      }

      /*
	ByMax< T >::operator() :
	前の値 preVal と新たな値 newVal に対し、その差分と newVal の絶対値の比率を求め、
	前回の結果と比較する。より大きい場合は入れ替える。
      */
      template< class T >
      void ByMax< T >::operator()( T preVal, T newVal )
      {
	// 新たな値による比率
	T ratio = std::abs( preVal - newVal );
	if ( newVal != 0 ) ratio /= std::abs( newVal );

	if ( max_ < ratio )
	  max_ = ratio;
      }

    } // namespace convergence_test

    namespace solver
    {
      /*
	Jacobi< T, Test >::operator() : Jacobi法による連立方程式の計算

	計算回数が maxCnt_ を超えた場合は例外 exception::ExcessLimit を投げる。
      */
      template< class T, template< class > class Test >
      template< class Ran >
      void Jacobi< T, Test >::operator()( const SquareMatrix< T >& a, Ran b )
      {
	using sm_size = typename SquareMatrix< T >::size_type;
	using vec_size = typename std::vector< T >::size_type;

	const std::string FUNC_NAME = "Jacobi< T, Test >::operator()";

	std::vector< T > ans( a.size(), T() ); // 仮の解
	uint32_t cnt = 1;                      // 処理回数

	// 対角成分が 0 なら処理失敗
	for ( sm_size i = 0 ; i < a.size() ; ++i )
	  if ( a[i][i] == 0 ) throw std::runtime_error( FUNC_NAME + " : " + error_message::UNSOLVED_EQUATION );

	while ( cnt - 1 < maxCnt_ ) {
	  std::vector< T > buff( a.size() ); // 次に計算する近似解

	  // 近似解の計算
	  for ( sm_size i = 0 ; i < a.size() ; ++i ) {
	    buff[i] = b[i];
	    for ( sm_size j = 0 ; j < a.size() ; ++j ) {
	      if ( i == j ) continue;
	      buff[i] -= a[i][j] * ans[j];
	    }
	    buff[i] /= a[i][i];
	  }

	  // 収束判定
	  test_.init();
	  for ( vec_size i = 0 ; i < ans.size() ; ++i ) {
	    test_( ans[i], buff[i] );
	    ans[i] = buff[i];
	  }
	  if ( test_.isConvergent() ) break;
	  ++cnt;
	}

	// 求めた解を代入
	std::copy( ans.begin(), ans.end(), b );

#ifdef DEBUG
	std::cout << FUNC_NAME + " : cnt = " << cnt << std::endl;
#endif

	if ( cnt - 1 >= maxCnt_ )
	  throw exception::ExcessLimit( FUNC_NAME + " : " + error_message::EXCESS_LIMIT, maxCnt_ );
      }

      /*
	GaussSeidel< T, Test >::operator() : ガウス=ザイデル法による連立方程式の計算

	計算回数が maxCnt_ を超えた場合は例外 exception::ExcessLimit を投げる。
      */
      template< class T, template< class > class Test >
      template< class Ran >
      void GaussSeidel< T, Test >::operator()( const SquareMatrix< T >& a, Ran b )
      {
	using size_type = typename SquareMatrix< T >::size_type;

	const std::string FUNC_NAME = "GaussSeidel< T, Test >::operator()";

	std::vector< T > ans( a.size(), T() ); // 仮の解
	uint32_t cnt = 1;                      // 処理回数

	// 対角成分が 0 なら処理失敗
	for ( size_type i = 0 ; i < a.size() ; ++i )
	  if ( a[i][i] == 0 ) throw std::runtime_error( FUNC_NAME + " : " + error_message::UNSOLVED_EQUATION );

	while ( cnt - 1 < maxCnt_ ) {
	  test_.init();
	  for ( size_type i = 0 ; i < a.size() ; ++i ) {
	    T preAns = ans[i]; // 前回の解を保持しておく

	    // 近似解の計算
	    ans[i] = b[i];
	    for ( size_type j = 0 ; j < a.size() ; ++j ) {
	      if ( i == j ) continue;
	      ans[i] -= a[i][j] * ans[j];
	    }
	    ans[i] /= a[i][i];

	    // 前回の解との差を求める
	    test_( preAns, ans[i] );
	  }
	  if ( test_.isConvergent() )
	    break;
	  ++cnt;
	}

	// 求めた解を代入
	std::copy( ans.begin(), ans.end(), b );

#ifdef DEBUG
	std::cout << FUNC_NAME + " : cnt = " << cnt << std::endl;
#endif

	if ( cnt - 1 >= maxCnt_ )
	  throw exception::ExcessLimit( FUNC_NAME + " : " + error_message::EXCESS_LIMIT, maxCnt_ );
      }

      /*
	Jacobi_SOR< T, Test >::operator() : Jacobi法にSOR法を適用して連立方程式を解く

	計算回数が maxCnt_ を超えた場合は例外 exception::ExcessLimit を投げる。
      */
      template< class T, template< class > class Test >
      template< class Ran >
      void Jacobi_SOR< T, Test >::operator()( const SquareMatrix< T >& a, Ran b )
      {
	using sm_size = typename SquareMatrix< T >::size_type;
	using vec_size = typename std::vector< T >::size_type;

	const std::string FUNC_NAME = "Jacobi_SOR< T, Test >::operator()";

	std::vector< T > ans( a.size(), T() ); // 仮の解
	uint32_t cnt = 1;                      // 処理回数

	// 対角成分が 0 なら処理失敗
	for ( sm_size i = 0 ; i < a.size() ; ++i )
	  if ( a[i][i] == 0 ) throw std::runtime_error( FUNC_NAME + " : " + error_message::UNSOLVED_EQUATION );

	while ( cnt - 1 < maxCnt_ ) {
	  std::vector< T > buff( a.size() ); // 次に計算する近似解

	  // 近似解の計算
	  for ( sm_size i = 0 ; i < a.size() ; ++i ) {
	    T t = b[i];
	    for ( sm_size j = 0 ; j < a.size() ; ++j ) {
	      if ( i == j ) continue;
	      t -= a[i][j] * ans[j];
	    }
	    t = t * w_ / a[i][i];
	    buff[i] = ans[i] * ( 1.0 - w_ ) + t;
	  }

	  // 収束判定
	  test_.init();
	  for ( vec_size i = 0 ; i < buff.size() ; ++i ) {
	    test_( ans[i], buff[i] );
	    ans[i] = buff[i];
	  }
	  if ( test_.isConvergent() )
	    break;
	  ++cnt;
	}

	// 求めた解を代入
	std::copy( ans.begin(), ans.end(), b );

#ifdef DEBUG
	std::cout << FUNC_NAME + " : w = " << w_ << ", cnt = " << cnt << std::endl;
#endif

	if ( cnt - 1 >= maxCnt_ )
	  throw exception::ExcessLimit( FUNC_NAME + " : " + error_message::EXCESS_LIMIT, maxCnt_ );
      }

      /*
	GaussSeidel_SOR< T, Test >::operator() : ガウス=ザイデル法にSOR法を適用して連立方程式を解く

	計算回数が maxCnt_ を超えた場合は例外 exception::ExcessLimit を投げる。
      */
      template< class T, template< class > class Test >
      template< class Ran >
      void GaussSeidel_SOR< T, Test >::operator()( const SquareMatrix< T >& a, Ran b )
      {
	using size_type =  typename SquareMatrix< T >::size_type;

	const std::string FUNC_NAME = "GaussSeidel_SOR< T, Test >::operator()";

	std::vector< T > ans( a.size(), T() ); // 仮の解
	uint32_t cnt = 1;                      // 処理回数

	// 対角成分が 0 なら処理失敗
	for ( size_type i = 0 ; i < a.size() ; ++i )
	  if ( a[i][i] == 0 ) throw std::runtime_error( FUNC_NAME + " : " + error_message::UNSOLVED_EQUATION );

	while ( cnt - 1 < maxCnt_ ) {
	  test_.init();
	  for ( size_type i = 0 ; i < a.size() ; ++i ) {
	    T preAns = ans[i]; // 前回の解を保持しておく

	    // 近似解の計算
	    T t = b[i];
	    for ( size_type j = 0 ; j < a.size() ; ++j ) {
	      if ( i == j ) continue;
	      t -= a[i][j] * ans[j];
	    }
	    t = t * w_ / a[i][i];
	    ans[i] = ans[i] * ( 1.0 - w_ ) + t;

	    // 前回の解との差を求める
	    test_( preAns, ans[i] );
	  }
	  if ( test_.isConvergent() )
	    break;
	  ++cnt;
	}

	// 求めた解を代入
	std::copy( ans.begin(), ans.end(), b );

#ifdef DEBUG
	std::cout << FUNC_NAME + " : w = " << w_ << ", cnt = " << cnt << std::endl;
#endif

	if ( cnt - 1 >= maxCnt_ )
	  throw exception::ExcessLimit( FUNC_NAME + " : " + error_message::EXCESS_LIMIT, maxCnt_ );
      }

      /*
	CGM_CalcAlpha : 共役勾配法による連立方程式計算 ( αを求める )

	α(k) = -( m(k-1), aTaaT・ax_b(k) ) / ( m(k-1), aTa・m(k-1) )
      */
      template< class T >
      T CGM_CalcAlpha( const std::vector< T >& m, const Matrix< T >& aTa, const std::vector< T >& aTax_b )
      {
	// 分子の計算
	std::vector< T > aTaaTax_b( m.size() );
	matrix_lib::multMV( aTa, aTax_b.begin(), aTaaTax_b.begin() );
	T n = std::inner_product( m.begin(), m.end(), aTaaTax_b.begin(), T() );

	// 分母の計算
	std::vector< T > aTam( m.size() );
	matrix_lib::multMV( aTa, m.begin(), aTam.begin() );
	T d = std::inner_product( m.begin(), m.end(), aTam.begin(), T() );

	return( ( d == 0 ) ? 0 : -n / d );
      }

      /*
	CGM_CalcM : 共役勾配法による連立方程式計算 ( mを求める )

	m(k) = aT・ax_b(k) + alpha(k)・m(k-1)
      */
      template< class T >
      void CGM_CalcM( const std::vector< T >& aTax_b, T alpha, std::vector< T >* m )
      {
	// alpha・m(k-1) の計算
	std::transform( m->begin(), m->end(), m->begin(), std::bind2nd( std::multiplies< T >(), alpha ) );

	// 両者の和を求める
	std::transform( aTax_b.begin(), aTax_b.end(), m->begin(), m->begin(), std::plus< T >() );
      }

      /*
	CGM_CalcT : 共役勾配法による連立方程式計算 ( tを求める )

	t(k) = -( m(k), aT・ax_b(k) ) / ( m(k), aTa・m(k) )
      */
      template< class T >
      T CGM_CalcT( const std::vector< T >& m, const std::vector< T >& aTax_b, const Matrix< T >& aTa )
      {
	// 分子の計算
	T n = std::inner_product( m.begin(), m.end(), aTax_b.begin(), T() );

	// 分母の計算
	std::vector< T > aTam( m.size() );
	matrix_lib::multMV( aTa, m.begin(), aTam.begin() );
	T d = std::inner_product( m.begin(), m.end(), aTam.begin(), T() );

	return( ( d == 0 ) ? 0 : -n / d );
      }

      /*
	CGM_CalcX : 共役勾配法による連立方程式計算 ( xを求める )

	x が収束したかを同時に判定する

	x(k+1) = x(k) + t(k)・ m(k)
      */
      template< class T, class Ran >
      bool CGM_CalcX( T t, const std::vector< T >& m, Ran x, T e )
      {
	// t(k)・m(k) の計算
	std::vector< T > tm( m );
	std::transform( tm.begin(), tm.end(), tm.begin(), std::bind2nd( std::multiplies< T >(), t ) );

	// x(k+1) = x(k) + t(k)・ m(k) の計算
	std::transform( tm.begin(), tm.end(), x, x, std::plus< T >() );

	// 収束したかを t(k)・m(k) の値で判断
	for ( typename std::vector< T >::size_type i = 0 ; i < tm.size() ; ++i )
	  if ( std::abs( tm[i] / x[i] ) > e ) return( false );

	return( true );
      }

      /*
	共役勾配法による連立方程式計算

	a 係数行列への参照
	x 対象の連立方程式の解ベクトルの先頭位置(計算結果の未知数がそのまま返される)
      */
      template< class T >
      template< class Ran >
      void CGM_LinearEquation< T >::operator()( const SquareMatrix< T >& a, Ran x )
      {
	std::vector< T > b( x, x + a.size() ); // b
	T alpha = 0;
	SquareMatrix< T > aT( a ); // At
	aT.transpose();
	Matrix< T > aTa; // At・A
	matrix_lib::mult( aT, a, &aTa );

	std::vector< T > aTax_b( b.size() ); // At( Ax - b )

	std::vector< T > m( b.size() );    // m
	std::fill( x, x + b.size(), T() ); // x = 0 から開始

	for ( ; ; ) {
	  // At( Ax - b ) の計算
	  matrix_lib::multMV( a, x, aTax_b.begin() ); // Ax
	  std::transform( aTax_b.begin(), aTax_b.end(), b.begin(), aTax_b.begin(), std::minus< T >() ); // Ax - b
	  matrix_lib::multMV( aT, aTax_b.begin(), aTax_b.begin() ); // At( Ax - b )

	  alpha = CGM_CalcAlpha( m, aTa, aTax_b ); // α(k) = -( m(k-1), AtAAt( Ax(k) - b ) ) / ( m(k-1), AtA・m(k-1) )
	  CGM_CalcM( aTax_b, alpha, &m );          // m(k) = At( Ax(k) - b ) + α(k)m(k-1)
	  T t = CGM_CalcT( m, aTax_b, aTa );       // t(k) = -( m(k), At( Ax(k) - b ) ) / ( m(k), AtA・m(k) )
	  if ( CGM_CalcX( t, m, x, e_ ) ) break;   // x(k+1) = x(k) + t(k)・m(k)
	}
      }

      /*
	Verify< T >::operator() : 連立方程式を解いた結果が正しいかを検証する

	ans が正解であるとして、b に正しい解が得られているかを確認する
      */
      template< typename T >
      template< class In1, class In2 >
      bool Verify< T >::operator()( const SquareMatrix< T >& a, In1 b, In2 ans )
      {
	for ( typename Matrix< T >::size_type r = 0 ; r < a.size() ; ++r ) {
	  In1 in = ans;
	  T t = std::inner_product( a[r], a[r].end(), in, T() );
	  if ( std::abs( t - *b++ ) >= threshold_ )
	    return( false );
	}

	return( true );
      }

      /*
	Verify< T >::multiple : 複数の連立方程式に対して解いた結果が正しいかを一度に検証する

	全ての連立方程式について一致すれば true を返し、一つでも一致しない連立方程式があれば false を返す。
      */
      template< class T >
      template< class MATRIX1, class MATRIX2 >
      bool Verify< T >::multiple( const SquareMatrix< T >& a, const MATRIX1& b, const MATRIX2& ans )
      {
	const std::string FUNC_NAME = "Verify< T >::multiple";

	error_lib::CheckMultMatrix( a, b, FUNC_NAME + " : a.cols() " + error_message::NOT_EQUAL_LENGTH + " b.rows()." );
	error_lib::CheckMultMatrix( a, ans, FUNC_NAME + " : a.cols() " + error_message::NOT_EQUAL_LENGTH + " ans.rows()." );

	if ( b.cols() != ans.cols() )
	  throw exception::NotEqualLength< typename MATRIX1::size_type >
	    ( FUNC_NAME + " : b.cols() " + error_message::NOT_EQUAL_LENGTH + " ans.cols()",
	      b.cols(), ans.cols() );

	for ( typename Matrix< T >::size_type c = 0 ; c < ans.cols() ; ++c ) {
	  if ( ! (*this)( a, b.cbegin( c ), ans.cbegin( c ) ) )
	    return( false );
	}

	return( true );
      }

    } // namespace solver

  } // namespace math_lib

} // namespace tadah_fussy
