/**
   @file error_library.h
   @brief エラー・例外処理定義用ヘッダ

   @author tadah_fussy
   @date 2015/12/19 新規作成
   @date 2019/03/27 コメントの見直し
   @date 2019/04/16 例外が出力するメッセージを外部から指定できるように仕様変更
   @date 2019/04/18 デフォルトのエラー・メッセージを追加
   @date 2019/04/19 ポリシー・クラスの廃止
   @date 2019/04/24 例外情報を除いたエラーチェック関数を追加
   @date 2021/03/13 名前空間の変更、bind2nd の廃止(bindへ変更)
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef ERROR_LIBRARY_20210402_H
#define ERROR_LIBRARY_20210402_H

#include <functional>
#include <stdexcept>
#include <iostream>

/**
   @brief 作成者 tadah_fussy
**/
namespace tadah_fussy
{
  /**
     @brief 例外用名前空間

     例外をまとめた名前空間
  **/
  namespace exception
  {
    /* --------------------------------------
       コンテナ(反復子)専用例外の定義
       -------------------------------------- */

    /**
       @brief コンテナクラスの長さに関する例外クラス

       コンテナの先頭・末尾位置が適正か、二つのコンテナの長さが等しいかなどをチェックし、
       そうでなかった場合に投げられる例外クラスの基底となる。
       std::length_error からの派生クラスなので std::length_error として捕捉することができる。
    **/
    struct Length : public std::length_error
    {
      /// @brief 例外に関する情報を指定して構築
      ///
      /// @param info 例外に関する情報
      explicit Length( const std::string& info )
	: std::length_error( info ) {}

      /// @brief 不正値や指定範囲などの情報を出力する(純粋仮想関数)
      ///
      /// @param os 出力ストリームへの参照
      /// @return なし
      virtual void printInfo( std::ostream& os ) const = 0;
    };

    /**
       @brief 長さが不正だった場合に対する例外クラス

       不正な長さ(負値であるなど)の場合に投げられる例外クラス。内部に不正と判断された長さを保持する。

       テンプレート引数の T は通常、対象イテレータやコンテナクラスの difference_type となる。
    **/
    template< class T >
    class InvalidLength : public Length
    {
      T invalidLength_; // 不正な長さ

    public:

      /// @brief 不正と判断された長さを指定して構築
      ///
      /// @param info 例外に関する情報
      /// @param invalidLength 不正な長さ
      InvalidLength( const std::string& info, T invalidLength )
	: Length( info ), invalidLength_( invalidLength ) {}

      /// @brief 不正な長さを返す
      ///
      /// @return 不正な長さ
      T invalidLength() const
      { return( invalidLength_ ); }

      /// @brief 不正値の出力
      ///
      /// @param os 出力ストリームへの参照
      /// @return なし
      void printInfo( std::ostream& os ) const
      { os << "invalid length = " << invalidLength(); }
    };

    /**
       @brief 長さが負値だった場合に対する例外クラス

       長さが負値だった(ゼロより小さかった)場合に投げられる例外クラス。内部に不正と判断された長さを保持する。
       長さゼロならば問題ない場合はこの例外を用いる。

       テンプレート引数の T は通常、対象イテレータやコンテナクラスの difference_type となる。
    **/
    template< class T >
    class NegativeLength : public InvalidLength< T >
    {
    public:

      /// @brief 不正と判断された値を指定して構築
      ///
      /// @param info 例外に関する情報
      /// @param invalidLength 不正な長さ
      explicit NegativeLength( const std::string& info, T invalidLength )
	: InvalidLength< T >( info, invalidLength ) {}
    };

    /**
       @brief 長さがゼロ以下だった場合に対する例外クラス

       長さがゼロ以下だった場合に投げられる例外クラス。内部に不正と判断された長さを保持する。
       長さゼロの場合は不正としたい場合はこの例外を用いる。

       テンプレート引数の T は通常、対象イテレータやコンテナクラスの difference_type となる。
    **/
    template< class T >
    class NotPositiveLength : public InvalidLength< T >
    {
    public:

      /// @brief 不正と判断された値を指定して構築
      ///
      /// @param info 例外に関する情報
      /// @param invalidLength 不正な長さ
      explicit NotPositiveLength( const std::string& info, T invalidLength )
	: InvalidLength< T >( info, invalidLength ) {}
    };

    /**
       @brief 長さを比較して等しくなかった場合に投げられる例外クラス

       内部に両者の長さを保持する。

       テンプレート引数の T は通常、対象イテレータやコンテナクラスの difference_type となる。
    **/
    template< class T >
    class NotEqualLength : public Length
    {
      T diff1_; // 長さ 1
      T diff2_; // 長さ 2

    public:

      /// @brief 二つの長さを指定して構築
      ///
      /// @param info 例外に関する情報
      /// @param diff1 長さ 1
      /// @param diff2 長さ 2
      NotEqualLength( const std::string& info, T diff1, T diff2 )
	: Length( info ), diff1_( diff1 ), diff2_( diff2 ) {}

      /// @brief 長さ 1 を返す
      ///
      /// @return 長さ 1
      T diff1() const
      { return( diff1_ ); }

      /// @brief 長さ 2 を返す
      ///
      /// @return 長さ 2
      T diff2() const
      { return( diff2_ ); }

      /// @brief 二つの長さの出力
      ///
      /// @param os 出力ストリームへの参照
      /// @return なし
      void printInfo( std::ostream& os ) const
      { os << "length[1] = " << diff1() << " length[2] = " << diff2(); }
    };

    /* --------------------------------------
       反復処理の処理回数超過例外
       -------------------------------------- */

    /**
       @brief 反復処理において処理回数が最大回数を超過した場合の例外クラス

       std::runtime_error からの派生クラスなので std::runtime_error として捕捉することができる。
    **/
    class ExcessLimit : public std::runtime_error
    {
      size_t limit_; // 処理回数の最大限度

    public:

      /// @brief 処理回数の最大限度を指定して構築
      ///
      /// @param info 例外に関する情報
      /// @param limit 処理回数の最大限度
      explicit ExcessLimit( const std::string& info, size_t limit )
	: std::runtime_error( info ), limit_( limit ) {}

      /// @brief 処理回数の最大限度を返す
      ///
      /// @return 処理回数の最大限度
      size_t limit() const
      { return( limit_ ); }

      /// @brief 設定されていた処理回数の最大限度を出力する
      ///
      /// @param os 出力ストリームへの参照
      /// @return なし
      void printInfo( std::ostream& os ) const
      { os << "limit = " << limit(); }
    };

    /* --------------------------------------
       数値演算専用例外
       -------------------------------------- */

    /**
       @brief 指定範囲外だった場合に投げられる例外クラス

       渡された値がある範囲内にあるか、正値か、負値ではないかなどをチェックし、
       そうでなかった場合に投げられる例外クラスの基底となる。
       std::domain_error からの派生クラスなので std::domain_error として捕捉することができる。
    **/
    struct Domain : public std::domain_error
    {
      /// @brief 例外に関する情報を指定して構築
      ///
      /// @param info 例外に関する情報
      explicit Domain( const std::string& info )
	: std::domain_error( info ) {}

      /// @brief 不正値や指定範囲などの情報を出力する(純粋仮想関数)
      ///
      /// @param os 対象の出力ストリーム
      virtual void printInfo( std::ostream& os ) const = 0;
    };

    /**
       @brief 指定した値より小さい場合に投げられる例外クラス

       内部に範囲の最小値と不正と判断された値を保持する。

       テンプレート引数の T は値の型を表す。不正値と範囲の最小値は同じ型であることを前提とする。

       @note 最小値と等しい場合を不正とするかどうかはこのクラスには含めていない。
    **/
    template< class T >
    class TooLittleNumber : public Domain
    {
      T invalidValue_; // 不正値
      T min_;          // 適正範囲の境界

    public:

      /// @brief 範囲の最小値と不正と判断された値を指定して構築
      ///
      /// @param info 例外に関する情報
      /// @param invalidValue 不正値
      /// @param min 範囲の最小値
      TooLittleNumber( const std::string& info, T invalidValue, T min )
	: Domain( info ), invalidValue_( invalidValue ), min_( min ) {}

      /// @brief 範囲の最小値を返す
      ///
      /// @return 範囲の最小値
      T min() const
      { return( min_ ); }

      /// @brief 不正値を返す
      ///
      /// @return 不正値
      T invalidValue() const
      { return( invalidValue_ ); }

      /// @brief 不正値・指定範囲の出力
      ///
      /// @param os 出力ストリームへの参照
      /// @return なし
      void printInfo( std::ostream& os ) const
      { os << "min = " << min() << " invalid value = " << invalidValue(); }
    };

    /**
       @brief 指定した値より大きい場合に投げられる例外クラス

       内部に範囲の最大値と不正と判断された値を保持する。

       テンプレート引数の T は値の型を表す。不正値と範囲の最大値は同じ型であることを前提とする。

       @note 最大値と等しい場合を不正とするかどうかはこのクラスには含めていない。
    **/
    template< class T >
    class TooGreatNumber : public Domain
    {
      T invalidValue_; // 不正値
      T max_; // 適正範囲の境界

    public:

      /// @brief 範囲の最大値と不正と判断された値を指定して構築
      ///
      /// @param info 例外に関する情報
      /// @param invalidValue 不正値
      /// @param max 範囲の最大値
      TooGreatNumber( const std::string& info, T invalidValue, T max )
	: Domain( info ), invalidValue_( invalidValue ), max_( max ) {}

      /// @brief 範囲の最大値を返す
      ///
      /// @return 範囲の最大値
      T max() const
      { return( max_ ); }

      /// @brief 不正値を返す
      ///
      /// @return 不正値
      T invalidValue() const
      { return( invalidValue_ ); }

      /// @brief 不正値・指定範囲の出力
      ///
      /// @param os 出力ストリームへの参照
      /// @return なし
      void printInfo( std::ostream& os ) const
      { os << "max = " << max() << " invalid value = " << invalidValue(); }
    };

    /**
       @brief 指定した値と等しくない場合に投げられる例外クラス

       内部に指定値と不正と判断された値を保持する。

       テンプレート引数の T は値の型を表す。不正値と指定値は同じ型であることを前提とする。
    **/
    template< class T >
    class NotEqualNumber : public Domain
    {
      T invalidValue_; // 不正値
      T correct_; // 指定値

    public:

      /// @brief 指定値と不正と判断された値を指定して構築
      ///
      /// @param info 例外に関する情報
      /// @param invalidValue 不正値
      /// @param correct 指定値
      NotEqualNumber( const std::string& info, T invalidValue, T correct )
	: Domain( info ), invalidValue_( invalidValue ), correct_( correct ) {}

      /// @brief 指定値(正しい値)を返す
      ///
      /// @return 指定値
      T correct() const
      { return( correct_ ); }

      /// @brief 不正値を返す
      ///
      /// @return 不正値
      T invalidValue() const
      { return( invalidValue_ ); }

      /// @brief 不正値と指定値の出力
      ///
      /// @param os 出力ストリームへの参照
      /// @return なし
      void printInfo( std::ostream& os ) const
      { os << "correct value = " << correct() << " invalid value = " << invalidValue(); }
    };

    /**
       @brief ゼロを境界とする範囲チェック用例外クラス

       「ゼロ以上」や「ゼロ未満」など、ゼロを境界とする範囲チェックに利用する。

       テンプレート引数の T は値の型を表す。
    **/
    template< class T >
    class ZeroLimits : public Domain
    {
      T invalidValue_; // 不正値

    public:

      /// @brief 不正と判断された値を指定して構築
      ///
      /// @param info 例外に対する情報
      /// @param invalidValue 不正値
      ZeroLimits( const std::string& info, T invalidValue )
	: Domain( info ), invalidValue_( invalidValue ) {}

      /// @brief 不正値を返す
      ///
      /// @return 不正値
      T invalidValue() const
      { return( invalidValue_ ); }

      /// @brief 不正値・指定範囲の出力
      ///
      /// @param os 出力ストリームへの参照
      /// @return なし
      void printInfo( std::ostream& os ) const
      { os << "invalid value = " << invalidValue(); }
    };

    /**
       @brief 値が負値だった場合に対する例外クラス

       値が負値だった(ゼロより小さかった)場合に投げられる例外クラス。内部に不正と判断された値を保持する。
       ゼロ以上でなければならない値のチェック時に利用する。

       テンプレート引数の T は値の型を表す。
    **/
    template< class T >
    class NegativeNumber : public ZeroLimits< T >
    {
    public:

      /// @brief 不正と判断された値を指定して構築
      ///
      /// @param info 例外に関する情報
      /// @param invalidValue 不正値
      explicit NegativeNumber( const std::string& info, T invalidValue )
	: ZeroLimits< T >( info, invalidValue ) {}
    };

    /**
       @brief 値がゼロ以下だった場合に対する例外クラス

       値がゼロ以下だった場合に投げられる例外クラス。内部に不正と判断された値を保持する。
       正値(ゼロより大きい数)でなければならない値のチェック時に利用する。

       テンプレート引数の T は値の型を表す。
    **/
    template< class T >
    class NotPositiveNumber : public ZeroLimits< T >
    {
    public:

      /// @brief 不正と判断された値を指定して構築
      ///
      /// @param info 例外に関する情報
      /// @param invalidValue 不正値
      explicit NotPositiveNumber( const std::string& info, T invalidValue )
	: ZeroLimits< T >( info, invalidValue ) {}
    };

  } // namespace exception

  /**
     @brief エラーチェック・処理ルーチン用名前空間

     エラーチェック・処理用の関数をまとめた名前空間
  **/
  namespace error_lib
  {
    /* --------------------------------------
       エラーチェック・処理用関数
       -------------------------------------- */

    /// @brief 判定関数と、判定結果に対する処理関数の組み合わせにより値を処理するための関数
    ///
    /// 渡された数を判定関数(叙述関数)で判定し、その結果に対して単項関数を使って処理を行う。
    /// 値の判定に使うことを想定している。例えば、
    ///
    /// @code
    /// CheckTool( std::bind( std::equal_to< int32_t >(), _1, 0 ),
    ///            tadah_fussy::exception::NotEqualNumber< int32_t >( "NotEqualNumber", i, 0 ),
    ///            i );
    /// @endcode
    ///
    /// とすれば、引数 i が 0 と等しくない場合は例外 tadah_fussy::exception::NotEqualNumber< int32_t > を返す。
    ///
    /// テンプレート引数の T は判定対象の値の型、Pred は判定用の単項関数オブジェクト、
    /// Exception は例外クラスをそれぞれ表す。
    ///
    /// @param pred 判定用単項叙述関数オブジェクト
    /// @param exception 判定結果が false の場合に投げられる例外
    /// @param t 判定する値
    /// @return 判定結果が true なら t をそのまま返す
    template< class T, class Pred, class Exception >
    T CheckTool( Pred pred, const Exception& exception, T t )
    {
      if ( ! pred( t ) )
        throw exception;

      return( t );
    }

    /// @brief データ列の長さを返す
    ///
    /// データ列の長さ( s に対する e の相対位置)を取得した後に値が負数になっていないか
    /// (開始と終了が逆になっていないか)をチェックし、テストにパスしなかった場合は
    /// 例外 tadah_fussy::exception::NegativeLength を投げる。
    ///
    /// テンプレート引数の In はデータ列の反復子の型を表す。
    /// このとき、tadah_fussy::exception::NegativeLength のテンプレート引数は std::iterator_traits< In >::difference_type となる。
    ///
    /// @param s データ列の開始位置
    /// @param e データ列の末尾の次の位置
    /// @param info 例外に関する情報
    /// @return データ列の長さ
    template< class In >
    typename std::iterator_traits< In >::difference_type
    ArrayLength( In s, In e, const std::string& info )
    {
      using D = typename std::iterator_traits< In >::difference_type;

      D diff = std::distance( s, e );

      return
        ( CheckTool
          ( std::bind( std::greater_equal< D >(), std::placeholders::_1, D() ),
            exception::NegativeLength< D >( info, diff ),
            diff
            )
          );
    }

    /// @brief データ列の長さを返す
    ///
    /// データ列の長さ( s に対する e の相対位置)を取得した後に値が負数になっていないか
    /// (開始と終了が逆になっていないか)をチェックし、テストにパスしなかった場合は
    /// 例外 ExceptionNegativeLength を投げる。
    ///
    /// テンプレート引数の In はデータ列の反復子の型を表す。
    /// このとき、ExceptionNegativeLength のテンプレート引数は std::iterator_traits< In >::difference_type となる。
    ///
    /// @param s データ列の開始位置
    /// @param e データ列の末尾の次の位置
    /// @return データ列の長さ
    template< class In >
    typename std::iterator_traits< In >::difference_type
    ArrayLength( In s, In e )
    {
      const std::string INFO = "tadah_fussy::exception::NegativeLength at ArrayLength";
      return( ArrayLength( s, e, INFO ) );
    }

    /// @brief データ列の長さをチェックする
    ///
    /// データ列の長さ( s に対する e の相対位置)が負数になっていないか(開始と終了が逆になっていないか)
    /// をチェックし、テストにパスしなかった場合は例外 tadah_fussy::exception::NegativeLength を投げる。
    ///
    /// テンプレート引数の In はデータ列の反復子の型を表す。
    /// このとき、tadah_fussy::exception::NegativeLength のテンプレート引数は std::iterator_traits< In >::difference_type となる。
    ///
    /// 内部では tadah_fussy::err_lib::ArrayLength を呼び出しているだけである。
    ///
    /// @param s データ列の開始位置
    /// @param e データ列の末尾の次の位置
    /// @param info 例外に関する情報
    /// @return なし
    template< class In >
    void CheckArrayLength( In s, In e, const std::string& info )
    { ArrayLength( s, e, info ); }

    /// @brief データ列の長さをチェックする
    ///
    /// データ列の長さ( s に対する e の相対位置)が負数になっていないか(開始と終了が逆になっていないか)
    /// をチェックし、テストにパスしなかった場合は例外 tadah_fussy::exception::NegativeLength を投げる。
    ///
    /// テンプレート引数の In はデータ列の反復子の型を表す。
    /// このとき、tadah_fussy::exception::NegativeLength のテンプレート引数は std::iterator_traits< In >::difference_type となる。
    ///
    /// 内部では tadah_fussy::err_lib::ArrayLength を呼び出しているだけである。
    ///
    /// @param s データ列の開始位置
    /// @param e データ列の末尾の次の位置
    /// @return なし
    template< class In >
    void CheckArrayLength( In s, In e )
    {
      const std::string INFO = "tadah_fussy::exception::NegativeLength at CheckArrayLength";
      ArrayLength( s, e, INFO );
    }

    /// @brief データ列の長さをチェックする
    ///
    /// データ列の範囲が len と等しいかをチェックし、テストにパスしなかった場合は例外を投げる。
    ///
    /// データ列の長さ( s に対する e の相対位置)が負数なら例外 tadah_fussy::exception::NegativeLength が投げられ、
    /// チェックする長さ len と一致しない場合は例外 tadah_fussy::exception::NotEqualLength が投げられる。
    ///
    /// テンプレート引数の In はデータ列の反復子の型を表す。
    /// このとき、tadah_fussy::::exception::NegativeLength・tadah_fussy::exception::NotEqualLength のテンプレート引数は
    /// std::iterator_traits< In >::difference_type となる。
    ///
    /// @param s データ列の開始位置
    /// @param e データ列の末尾の次の位置
    /// @param len チェックする長さ
    /// @param infoNegativeLength tadah_fussy::exception::NegativeLength 例外に関する情報
    /// @param infoNotEqualLength tadah_fussy::exception::NotEqualLength 例外に関する情報
    /// @return なし
    template< class In >
    void CheckArrayLength( In s, In e,
                           typename std::iterator_traits< In >::difference_type len,
                           const std::string& infoNegativeLength,
                           const std::string& infoNotEqualLength )
    {
      using D = typename std::iterator_traits< In >::difference_type;

      D diff = ArrayLength( s, e, infoNegativeLength );

      CheckTool
        ( std::bind( std::equal_to< D >(), std::placeholders::_1, len ),
          exception::NotEqualLength< D >( infoNotEqualLength, diff, len ),
          diff
          );
    }

    /// @brief データ列の長さをチェックする
    ///
    /// データ列の範囲が len と等しいかをチェックし、テストにパスしなかった場合は例外を投げる。
    ///
    /// データ列の長さ( s に対する e の相対位置)が負数なら例外 tadah_fussy::exception::NegativeLength が投げられ、
    /// チェックする長さ len と一致しない場合は例外 tadah_fussy::::exception::NotEqualLength が投げられる。
    ///
    /// テンプレート引数の In はデータ列の反復子の型を表す。
    /// このとき、tadah_fussy::exception::NegativeLength・tadah_fussy::exception::NotEqualLength のテンプレート引数は std::iterator_traits< In >::difference_type となる。
    ///
    /// @param s データ列の開始位置
    /// @param e データ列の末尾の次の位置
    /// @param len チェックする長さ
    /// @return なし
    template< class In >
    void CheckArrayLength( In s, In e,
                           typename std::iterator_traits< In >::difference_type len )
    {
      const std::string NEGATIVE_LENGTH_INFO = "tadah_fussy::exception::NegativeLength at CheckArrayLength";
      const std::string NOT_EQUAL_LENGTH_INFO = "tadah_fussy::exception::NotEqualLength at CheckArrayLength";

      CheckArrayLength( s, e, len, NEGATIVE_LENGTH_INFO, NOT_EQUAL_LENGTH_INFO );
    }

    /// @brief 負値でないかをチェックする (Not Negative Number)
    ///
    /// 値が負値だった場合は例外 tadah_fussy::exception::NegativeNumber< T > を投げ、そうでなければ値をそのまま返す。
    /// 値のフィルタリングに利用する。
    ///
    /// テンプレート引数の T は値の型を表す。
    ///
    /// @param t チェック対象の値
    /// @param info 例外に関する情報
    /// @return フィルタ結果(対象の値 t がゼロ以上ならそのまま返す)
    template< class T >
    T NNNum( T t, const std::string& info )
    {
      return( CheckTool( std::bind( std::greater_equal< T >(), std::placeholders::_1, T() ),
                         exception::NegativeNumber< T >( info, t ), t ) );
    }

    /// @brief 負値でないかをチェックする (Not Negative Number)
    ///
    /// 値が負値だった場合は例外 ExceptionNegativeNumber< T > を投げ、そうでなければ値をそのまま返す。
    /// 値のフィルタリングに利用する。
    ///
    /// テンプレート引数の T は値の型を表す。
    ///
    /// @param t チェック対象の値
    /// @return フィルタ結果(対象の値 t がゼロ以上ならそのまま返す)
    template< class T >
    T NNNum( T t )
    {
      return( NNNum( t, "tadah_fussy::exception::NegativeNumber at NNNum" ) );
    }

    /// @brief 正値(ゼロより大)であるかをチェックする (Positive Number)
    ///
    /// 値がゼロ以下だった場合は例外 tadah_fussy::exception::NotPositiveNumber< T > を投げ、そうでなければ値をそのまま返す。
    /// 値のフィルタリングに利用する。
    ///
    /// テンプレート引数の T は値の型を表す。
    ///
    /// @param t チェック対象の値
    /// @param info 例外に関する情報
    /// @return フィルタ結果(対象の値 t が正値ならそのまま返す)
    template< class T >
    T PNum( T t, const std::string& info )
    {
      return( CheckTool( std::bind( std::greater< T >(), std::placeholders::_1, T() ),
                         exception::NotPositiveNumber< T >( info, t ), t ) );
    }

    /// @brief 正値(ゼロより大)であるかをチェックする (Positive Number)
    ///
    /// 値がゼロ以下だった場合は例外 tadah_fussy::exception::NotPositiveNumber< T > を投げ、そうでなければ値をそのまま返す。
    /// 値のフィルタリングに利用する。
    ///
    /// テンプレート引数の T は値の型を表す。
    ///
    /// @param t チェック対象の値
    /// @return フィルタ結果(対象の値 t が正値ならそのまま返す)
    template< class T >
    T PNum( T t )
    {
      return( PNum( t, "tadah_fussy::exception::NotPositiveNumber at PNum" ) );
    }

    /// @brief 指定値以下であるかをチェックする
    ///
    /// 値が指定値より大きい場合は例外 tadah_fussy::exception::TooGreatNumber< T > を投げ、そうでなければ値をそのまま返す。
    /// 値のフィルタリングに利用する。
    ///
    /// テンプレート引数の T は値の型を表す。
    ///
    /// @param t チェック対象の値
    /// @param max 指定値(しきい値)
    /// @param info 例外に関する情報
    /// @return フィルタ結果(値 t が max 以下ならそのまま返す)
    template< class T >
    T LE( T t, T max, const std::string& info )
    {
      return( CheckTool( std::bind( std::less_equal< T >(), std::placeholders::_1, max ),
                         exception::TooGreatNumber< T >( info, t, max ), t ) );
    }

    /// @brief 指定値以下であるかをチェックする
    ///
    /// 値が指定値より大きい場合は例外 tadah_fussy::exception::TooGreatNumber< T > を投げ、そうでなければ値をそのまま返す。
    /// 値のフィルタリングに利用する。
    ///
    /// テンプレート引数の T は値の型を表す。
    ///
    /// @param t チェック対象の値
    /// @param max 指定値(しきい値)
    /// @return フィルタ結果(値 t が max 以下ならそのまま返す)
    template< class T >
    T LE( T t, T max )
    {
      return( LE( t, max, "tadah_fussy::exception::TooGreatNumber at LE" ) );
    }

    /// @brief 指定値より小さいかをチェックする
    ///
    /// 値が指定値以上だった場合は例外 tadah_fussy::exception::TooGreatNumber< T > を投げ、そうでなければ値をそのまま返す。
    /// 値のフィルタリングに利用する。
    ///
    /// テンプレート引数の T は値の型を表す。
    ///
    /// @param t チェック対象の値
    /// @param max 指定値(しきい値)
    /// @param info 例外に関する情報
    /// @return フィルタ結果(値 t が max より小さければそのまま返す)
    template< class T >
    T LT( T t, T max, const std::string& info )
    {
      return( CheckTool( std::bind( std::less< T >(), std::placeholders::_1, max ),
                         exception::TooGreatNumber< T >( info, t, max ), t ) );
    }

    /// @brief 指定値より小さいかをチェックする
    ///
    /// 値が指定値以上だった場合は例外 tadah_fussy::exception::TooGreatNumber< T > を投げ、そうでなければ値をそのまま返す。
    /// 値のフィルタリングに利用する。
    ///
    /// テンプレート引数の T は値の型を表す。
    ///
    /// @param t チェック対象の値
    /// @param max 指定値(しきい値)
    /// @return フィルタ結果(値 t が max より小さければそのまま返す)
    template< class T >
    T LT( T t, T max )
    {
      return( LT( t, max, "tadah_fussy::exception::TooGreatNumber at LT" ) );
    }

    /// @brief 指定値以上であるかをチェックする
    ///
    /// 値が指定値より小さい場合は例外 tadah_fussy::exception::TooLittleNumber< T > を投げ、そうでなければ値をそのまま返す。
    /// 値のフィルタリングに利用する。
    ///
    /// テンプレート引数の T は値の型を表す。
    ///
    /// @param t チェック対象の値
    /// @param min 指定値(しきい値)
    /// @param info 例外に関する情報
    /// @return フィルタ結果(値 t が max 以上ならそのまま返す)
    template< class T >
    T GE( T t, T min, const std::string& info )
    {
      return( CheckTool( std::bind( std::greater_equal< T >(), std::placeholders::_1, min ),
                         exception::TooLittleNumber< T >( info, t, min ), t ) );
    }

    /// @brief 指定値以上であるかをチェックする
    ///
    /// 値が指定値より小さい場合は例外 ExceptionTooLittleNumber< T > を投げ、そうでなければ値をそのまま返す。
    /// 値のフィルタリングに利用する。
    ///
    /// テンプレート引数の T は値の型を表す。
    ///
    /// @param t チェック対象の値
    /// @param min 指定値(しきい値)
    /// @return フィルタ結果(値 t が max 以上ならそのまま返す)
    template< class T >
    T GE( T t, T min )
    {
      return( GE( t, min, "tadah_fussy::exception::TooLittleNumber at GE" ) );
    }

    /// @brief 指定値より大きいかをチェックする
    ///
    /// 値が指定値以下だった場合は例外 tadah_fussy::exception::TooLittleNumber< T > を投げ、そうでなければ値をそのまま返す。
    /// 値のフィルタリングに利用する。
    ///
    /// テンプレート引数の T は値の型を表す。
    ///
    /// @param t チェック対象の値
    /// @param min 指定値(しきい値)
    /// @param info 例外に関する情報
    /// @return フィルタ結果(値 t が min 以下ならそのまま返す)
    template< class T >
    T GT( T t, T min, const std::string& info )
    {
      return( CheckTool( std::bind( std::greater< T >(), std::placeholders::_1, min ),
                         exception::TooLittleNumber< T >( info, t, min ), t ) );
    }

    /// @brief 指定値より大きいかをチェックする
    ///
    /// 値が指定値以下だった場合は例外 tadah_fussy::exception::TooLittleNumber< T > を投げ、そうでなければ値をそのまま返す。
    /// 値のフィルタリングに利用する。
    ///
    /// テンプレート引数の T は値の型を表す。
    ///
    /// @param t チェック対象の値
    /// @param min 指定値(しきい値)
    /// @return フィルタ結果(値 t が min 以下ならそのまま返す)
    template< class T >
    T GT( T t, T min )
    {
      return( GT( t, min, "tadah_fussy::exception::TooLittleNumber at GT" ) );
    }

    /// @brief 指定値と等しいかをチェックする
    ///
    /// 値が指定値と一致しない場合は例外 tadah_fussy::exception::NotEqualNumber< T > を投げ、そうでなければ値をそのまま返す。
    /// 値のフィルタリングに利用する。
    ///
    /// テンプレート引数の T は値の型を表す。
    ///
    /// @param t チェック対象の値
    /// @param correct 指定値(正しい値)
    /// @param info 例外に関する情報
    /// @return フィルタ結果(値 t が correct と等しいならそのまま返す)
    template< class T >
    T EQ( T t, T correct, const std::string& info )
    {
      return( CheckTool( std::bind( std::equal_to< T >(), std::placeholders::_1, correct ),
                         exception::NotEqualNumber< T >( info, t, correct ), t ) );
    }

    /// @brief 指定値と等しいかをチェックする
    ///
    /// 値が指定値と一致しない場合は例外 tadah_fussy::exception::NotEqualNumber< T > を投げ、そうでなければ値をそのまま返す。
    /// 値のフィルタリングに利用する。
    ///
    /// テンプレート引数の T は値の型を表す。
    ///
    /// @param t チェック対象の値
    /// @param correct 指定値(正しい値)
    /// @return フィルタ結果(値 t が correct と等しいならそのまま返す)
    template< class T >
    T EQ( T t, T correct )
    {
      return( EQ( t, correct, "tadah_fussy::exception::NotEqualNumber at EQ" ) );
    }

    /// @brief 二つのデータ列が有効かをチェックする
    ///
    /// 二つのデータ列(独立変数と従属変数)に対して以下のテストを行い、テストにパスしなかった場合は例外を投げる。
    ///
    /// -# 独立変数、従属変数の要素の範囲が適切か(開始と終了が逆になっていないか)
    /// -# 独立変数と従属変数の要素数が一致しているか
    ///
    /// 1 のテストにパスしなかった場合は例外 tadah_fussy::exception::NegativeLength が、2 のテストに
    /// パスしなかった場合は例外 tadah_fussy::exception::NotEqualLength が投げられる。
    ///
    /// テンプレート引数の In1 と In2 はそれぞれ二つのデータ列の反復子の型を表す。
    /// tadah_fussy::exception::NegativeLength のテンプレート引数は、テストにパスしなかったデータ列が [ s1, e1 ) か [ s2, e2 ) かによって変化し、
    /// [ s1, e1 ) なら std::iterator_traits< In1 >::difference_type、[ s2, e2 ) なら std::iterator_traits< In2 >::difference_type となる。
    /// tadah_fussy::exception::NotEqualLength のテンプレート引数は、std::iterator_traits< In1 >::difference_type 側になることに注意。
    ///
    /// 戻り値は、独立変数の反復子に対する距離の型で返される。
    ///
    /// @param s1 独立変数の要素の開始位置
    /// @param e1 独立変数の要素の末尾の次の位置
    /// @param s2 従属変数の要素の開始位置
    /// @param e2 従属変数の要素の末尾の次の位置
    /// @param infoNegativeLength1 [ s1, e1 ) の tadah_fussy::exception::NegativeLength 例外に関する情報
    /// @param infoNegativeLength2 [ s2, e2 ) の tadah_fussy::exception::NegativeLength 例外に関する情報
    /// @param infoNotEqualLength tadah_fussy::exception::NotEqualLength 例外に関する情報
    /// @return データ列の長さ
    template< class In1, class In2 >
    typename std::iterator_traits< In1 >::difference_type
    CheckBivariable( In1 s1, In1 e1, In2 s2, In2 e2,
                     const std::string& infoNegativeLength1, const std::string& infoNegativeLength2,
                     const std::string& infoNotEqualLength )
    {
      using D1 = typename std::iterator_traits< In1 >::difference_type;
      using D2 = typename std::iterator_traits< In2 >::difference_type;

      D1 diff1 = ArrayLength( s1, e1, infoNegativeLength1 );
      D2 diff2 = ArrayLength( s2, e2, infoNegativeLength2 );

      return
        ( CheckTool
          ( std::bind( std::equal_to< D2 >(), std::placeholders::_1, diff2 ),
            exception::NotEqualLength< D1 >( infoNotEqualLength, diff1, diff2 ),
            diff1
            )
          );
    }

    /// @brief 二つのデータ列が有効かをチェックする
    ///
    /// 二つのデータ列(独立変数と従属変数)に対して以下のテストを行い、テストにパスしなかった場合は例外を投げる。
    ///
    /// -# 独立変数、従属変数の要素の範囲が適切か(開始と終了が逆になっていないか)
    /// -# 独立変数と従属変数の要素数が一致しているか
    ///
    /// 1 のテストにパスしなかった場合は例外 tadah_fussy::exception::NegativeLength が、2 のテストに
    /// パスしなかった場合は例外 tadah_fussy::exception::NotEqualLength が投げられる。
    ///
    /// テンプレート引数の In1 と In2 はそれぞれ二つのデータ列の反復子の型を表す。
    /// tadah_fussy::exception::NegativeLength のテンプレート引数は、テストにパスしなかったデータ列が [ s1, e1 ) か [ s2, e2 ) かによって変化し、
    /// [ s1, e1 ) なら std::iterator_traits< In1 >::difference_type、[ s2, e2 ) なら std::iterator_traits< In2 >::difference_type となる。
    /// tadah_fussy::exception::NotEqualLength のテンプレート引数は、std::iterator_traits< In1 >::difference_type 側になることに注意。
    ///
    /// 戻り値は、独立変数の反復子に対する距離の型で返される。
    ///
    /// @param s1 独立変数の要素の開始位置
    /// @param e1 独立変数の要素の末尾の次の位置
    /// @param s2 従属変数の要素の開始位置
    /// @param e2 従属変数の要素の末尾の次の位置
    /// @return データ列の長さ
    template< class In1, class In2 >
    typename std::iterator_traits< In1 >::difference_type
    CheckBivariable( In1 s1, In1 e1, In2 s2, In2 e2 )
    {
      const std::string NEGATIVE_LENGTH_INFO = "tadah_fussy::exception::NegativeLength at CheckBivariable";
      const std::string NOT_EQUAL_LENGTH_INFO = "tadah_fussy::exception::NotEqualLength at CheckBivariable";

      return( CheckBivariable( s1, e1, s2, e2, NEGATIVE_LENGTH_INFO, NEGATIVE_LENGTH_INFO, NOT_EQUAL_LENGTH_INFO ) );
    }

  } // namespace error_lib

  /**
     @brief エラー・メッセージ用名前空間

     エラー・メッセージをまとめた名前空間
  **/
  namespace error_message
  {
    /// @brief 処理速度が最大限度を超えたときのエラー・メッセージ
    const std::string EXCESS_LIMIT = "iteration times exceeded the limit";

    /// @brief データ列の長さが負数であった場合のエラー・メッセージ
    const std::string NEGATIVE_LENGTH = "must not be negative length";

    /// @brief データ列の長さが指定値に一致しない場合のエラー・メッセージ
    const std::string NOT_EQUAL_LENGTH = "must be equal to";

    /// @brief 値が負値だった場合のエラー・メッセージ
    const std::string NNNUM = "must not be negative";

    /// @brief 値がゼロ以下だった場合のエラー・メッセージ
    const std::string PNUM = "must be positive";

    /// @brief 値が指定値より大きい場合のエラー・メッセージ
    const std::string LE = "must be less than or equal to";

    /// @brief 値が指定値以上だった場合のエラー・メッセージ
    const std::string LT = "must be less than";

    /// @brief 値が指定値より小さい場合のエラー・メッセージ
    const std::string GE = "must be greater than or equal to";

    /// @brief 値が指定値以下だった場合のエラー・メッセージ
    const std::string GT = "must be greater than";

    /// @brief 値が指定値と等しくない場合のエラー・メッセージ
    const std::string EQ = "must be equal to";

  } // namespace error_message

} // namespace tadah_fussy

#endif
