/**
   @file gadget.h
   @brief ガジェット・プログラム定義用ヘッダ

   @author tadah_fussy
   @date 2015/02/16 コメントを Doxygen 用に変更 
   @date 2019/03/27 参照に対する NULL チェックを外す(コンパイラのワーニング対策)、
                    コメントの見直し
   @date 2019/04/04 StrGadget::Split のバグ修正(NULL文字列が無視されていた)
                    CharArray コピーコンストラクタのバグ修正 (explicit を除外する)
   @date 2021/03/13 名前空間の変更、TimeMeasure の内部仕様を変更
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef GADGET_20210402_H
#define GADGET_20210402_H

#include <string>
#include <vector>
#include <iostream>
#include <sstream>

#include <cstdio>
#include <chrono>
#include <cassert>
#include <cstring>

/**
   @brief 作成者 tadah_fussy
**/
namespace tadah_fussy
{
  /**
     @brief 文字列専用小道具ツール
  **/
  namespace string_lib
  {
    /// @brief string型→char*型への変換
    ///
    /// @param s 対象文字列
    /// @param c 変換結果を登録する char 型配列へのポインタ
    /// @param size cのサイズ
    /// @return なし
    void CharCopy( const std::string& s, char* c, size_t size );

    /// @brief 文字列の両端にある空白文字を除去する
    ///
    /// 空白文字として SPACE, TAB, コントロールコード \\f, \\v, \\n, \\r を対象とする
    /// (isspaceが対象とする文字を除去)
    ///
    /// @param st 対象文字列
    /// @return 処理後の文字列
    std::string Trim( const std::string& st );

    /// @brief 文字列の比較
    ///
    /// 二つの文字列を先頭から順に比較し、不一致を見つけたら、s1 の文字コードの方が s2 より小さければ -1、
    /// 逆の場合は 1 を返す。もしどちらかが末尾まで達した場合、文字列長が等しければ 0、s1 の方が短ければ -1、
    /// 逆の場合は 1 を返す。ignoreCase が true なら、大文字と小文字の違いは無視される。
    ///
    /// @param s1 比較対象文字列 1
    /// @param s2 比較対象文字列 2
    /// @param ignoreCase true 時に大文字・小文字の違いを無視する
    /// @retval s1 < s2 ... -1
    /// @retval s1 > s2 ... 1
    /// @retval s1 = s2 ... 0
    int8_t Compare( const std::string& s1, const std::string& s2, bool ignoreCase = false );

    /// @brief 文字列を指定文字コードで区切る
    ///
    /// @param s 分割対象の文字列
    /// @param c 分割位置を示す文字コード
    /// @param ret 分割した文字列を保持するベクタへのポインタ
    /// @return s が空、または ret が未定義なら false
    bool Split( const std::string& s, char c, std::vector< std::string >* ret );

    /// @brief 文字列のベクタを指定文字コードを付加しながら連結
    ///
    /// @param vecStr 対象となる文字列のベクタ
    /// @param c 区切りとして使用する文字コード
    /// @param ret 連結した文字列を保持する string へのポインタ
    /// @return ret が未定義なら false
    bool Join( const std::vector< std::string >& vecStr, char c, std::string* ret );

    /// @brief 日付を文字列に変換
    ///
    /// 日付のフォーマットは YYYY/MM/DD hh:mm:ss 固定である。
    ///
    /// @param time 日付
    /// @return 変換した文字列
    std::string GetDateTimeFormat( time_t time );

    /**
       @brief 文字コードの配列

       string を C スタイル文字列として扱う必要がある場合に利用する
    **/
    class CharArray
    {
      char* cp_; // char 型配列

      // 空の文字列を構築する
      void createNullStr()
      {
        cp_ = new char[1];
        cp_[0] = '\0';
      }

    public:

      /* コンストラクタ */

      /// @brief デフォルトコンストラクタ
      ///
      /// 空の文字列で初期化する
      CharArray() { createNullStr(); }

      /// @brief コピーする文字列を指定して構築
      ///
      /// @param s 対象の文字列
      explicit CharArray( const std::string& s );

      /// @brief コピー・コンストラクタ
      ///
      /// @param c コピー元の CharArray オブジェクト
      CharArray( const CharArray& c );

      /// @brief デストラクタ
      ///
      /// 構築した char 型配列を解体する
      ~CharArray() { delete[] cp_; }

      /// @brief 代入演算子の多重定義
      ///
      /// @param c 代入する対象の CharArray オブジェクト
      /// @return 自分自身への参照
      CharArray& operator=( const CharArray& c );

      /// @brief 文字コードの配列を返す
      ///
      /// @return 配列の先頭ポインタ
      const char* data() const { return( cp_ ); }

      /// @brief 文字列長を返す
      ///
      /// 関数 strlen の返り値をそのまま返すため、末尾の \\0 は文字列長には含まれない。
      ///
      /// @return 文字列長
      size_t size() const { return( strlen( cp_ ) ); }
    };

    /**
       @brief 文字列比較用関数オブジェクト
    **/
    struct Cmp : public std::binary_function< std::string, std::string, int8_t >
    {
      bool ignoreCase; ///< 大文字・小文字を無視するなら true にする

      /// @brief 大文字・小文字を無視するかを指定して構築
      ///
      /// @param nCase 大文字・小文字を無視するなら true にする
      explicit Cmp( bool nCase = false ) : ignoreCase( nCase ) {}

      /// @brief 二つの文字列を比較する
      ///
      /// 二つの文字列を先頭から順に比較し、不一致を見つけたら、s1 の文字コードの方が s2 より小さければ -1、
      /// 逆の場合は 1 を返す。もしどちらかが末尾まで達した場合、文字列長が等しければ 0、s1 の方が短ければ -1、
      /// 逆の場合は 1 を返す。ignoreCase が true なら、大文字と小文字の違いは無視される。
      ///
      /// @param s1 比較対象文字列 1
      /// @param s2 比較対象文字列 2
      /// @retval s1 < s2 ... -1
      /// @retval s1 > s2 ... 1
      /// @retval s1 = s2 ... 0
      int8_t operator()( const std::string& s1, const std::string& s2 ) const
      { return( Compare( s1, s2, ignoreCase ) ); }
    };

    /**
       @brief 文字列等号比較用関数オブジェクト

       アルゴリズムで利用することを想定している。
       例えば、大文字・小文字を無視した上で二つのシーケンスの文字列を比較したい場合は次のようにすればよい。

       @code
       std::vector< std::string > str1;
       std::vector< std::string > str2;
       :
       // シーケンスの内容が一致したら true を返す
       bool res = std::equal( str1.begin(), str1.end(), str2.begin(), tadah_fussy::string::Eq( true ) );
       @endcode
    **/
    struct Eq : public std::binary_function< std::string, std::string, bool >
    {
      bool ignoreCase; ///< 大文字・小文字を無視するなら true にする

      /// @brief 大文字・小文字を無視するかを指定して構築
      ///
      /// @param nCase 大文字・小文字を無視するなら true にする
      Eq( bool nCase = false ) : ignoreCase( nCase ) {}

      /// @brief 二つの文字列が等しければ true を返す
      ///
      /// @param s1 比較対象文字列 1
      /// @param s2 比較対象文字列 2
      /// @return s1 = s2 なら true を返す
      bool operator()( const std::string& s1, const std::string& s2 ) const
      { return( Compare( s1, s2, ignoreCase ) == 0 ); }
    };

    /**
       @brief 文字列 "より小さい" 比較用関数オブジェクト

       アルゴリズムで利用することを想定している。
       例えば、大文字・小文字を無視して文字列の比較をした上でソートしたい場合は次のようにすればよい。

       @code
       std::vector< std::string > str;
       :
       std::sort( str.begin(), str.end(), tadah_fussy::string::Less( true ) );
       @endcode
    **/
    struct Less : public std::binary_function< std::string, std::string, bool >
    {
      bool ignoreCase; ///< 大文字・小文字を無視するなら true にする

      /// @brief 大文字・小文字を無視するかを指定して構築
      ///
      /// @param nCase 大文字・小文字を無視するなら true にする
      Less( bool nCase = false ) : ignoreCase( nCase ) {}

      /// @brief s1 の方が小さければ true を返す
      ///
      /// 二つの文字列を先頭から順に比較し、不一致を見つけたら、s1 の文字コードの方が s2 より小さければ true、
      /// 逆の場合は false を返す。もしどちらかが末尾まで達した場合、s1 の方が短ければ true、
      /// それ以外の場合は false を返す。ignoreCase が true なら、大文字と小文字の違いは無視される。
      ///
      /// @param s1 比較対象文字列 1
      /// @param s2 比較対象文字列 2
      /// @return s1 < s2 なら true を返す
      bool operator()( const std::string& s1, const std::string& s2 ) const
      { return( Compare( s1, s2, ignoreCase ) < 0 ); }
    };

  } // namespace string_lib

  /**
     数値専用小道具ツール
  **/
  namespace numeric_lib
  {
    /* 叙述関数 */

    /**
       @brief 絶対値の比較用関数オブジェクト
    **/
    template< class T >
    struct LessByAbs
    {
      /// @brief 絶対値の大きさを比較し、|t1| < |t2| なら true を返す
      ///
      /// @param t1 値 1
      /// @param t2 値 2
      /// @return |t1| < |t2| なら true を返す
      bool operator()( T t1, T t2 )
      {
        return( std::less< T >()( std::abs( t1 ), std::abs( t2 ) ) );
      }
    };

    /**
       @brief 非常に近い値は等しいと判断する関数オブジェクト

       テンプレート引数の型 D は比較対象の値の型であり、浮動小数点数での利用を想定している。
    **/
    template< class D >
    class ApproxEq
    {
      D threshold_; // しきい値

    public:

      /// @brief しきい値を指定して構築
      ///
      /// @note しきい値が負数だった場合は assert を実行する
      ///
      /// @param threshold しきい値
      explicit ApproxEq( D threshold )
        : threshold_( threshold ) { assert( threshold >= 0 ); }

      /// @brief 二数の比較
      ///
      /// @param d1 数1
      /// @param d2 数2
      /// @return 二数が近似的に等しい(差の絶対値がしきい値より小さい)場合は true を返す
      bool operator()( D d1, D d2 )
      { return( std::abs( d1 - d2 ) < threshold_ ); }
    };

    /**
       @brief 等値判定をした後で大小比較を行う関数オブジェクト

       テンプレート引数の型 D は比較対象の値の型を表す。

       近似的に等しい場合は false を返したい場合などに利用する
    **/
    template< class D >
    class ApproxLess
    {
      ApproxEq< D > eq_; // 等値判定用関数オブジェクト

    public:

      /// @brief 等値判定用関数オブジェクトを指定して構築
      ///
      /// @note しきい値が負数だった場合は assert を実行する
      ///
      /// @param threshold しきい値(二数の差の絶対値がしきい値より小さい場合は近似的に等しいと判断される)
      explicit ApproxLess( D threshold )
        : eq_( threshold ) {}

      /// @brief 二数の比較
      ///
      /// @param d1 数1
      /// @param d2 数2
      /// @return d1 < d2 なら true を返す。但し、二数の差の絶対値がしきい値より小さい場合は false を返す
      bool operator()( D d1, D d2 )
      {
        if ( eq_( d1, d2 ) )
          return( false );

        return( d1 < d2 );
      }
    };

    /**
       @brief 二変数を保持するオブジェクト
    **/
    template< class T >
    struct Bivariate
    {
      /// @brief 変数の型
      using value_type = T;

      value_type x; ///< 変数1(x)
      value_type y; ///< 変数2(y)

      /// @brief デフォルト・コンストラクタ
      Bivariate()
	: x( value_type() ), y( value_type() ) {}

      /// @brief 二変数の値を指定して構築
      ///
      /// @param value_x 変数1(x)の値
      /// @param value_y 変数2(y)の値
      Bivariate( value_type value_x, value_type value_y )
	: x( value_x ), y( value_y ) {}

      /// @brief 他の二変数を加算する
      ///
      /// @param val 加算する二変数
      /// @return 自分自身への参照
      Bivariate& operator+=( const Bivariate& val )
      {
	x += val.x;
	y += val.y;
	return( *this );
      }

      /// @brief 他の二変数で減算する
      ///
      /// @param val 減算する二変数
      /// @return 自分自身への参照
      Bivariate& operator-=( const Bivariate& val )
      {
	x -= val.x;
	y -= val.y;
	return( *this );
      }

      /// @brief 二変数に同じ値を乗算する
      ///
      /// @param val 乗算する対象の値
      /// @return 自分自身への参照
      Bivariate& operator*=( value_type val )
      {
	x *= val;
	y *= val;
	return( *this );
      }

      /// @brief 二変数に同じ値を除算する
      ///
      /// @param val 除算する対象の値
      /// @return 自分自身への参照
      Bivariate& operator/=( value_type val )
      {
	x /= val;
	y /= val;
	return( *this );
      }
    };

    /// @brief 二変数どうしの加算
    ///
    /// @param val1 加算する二変数1
    /// @param val2 加算する二変数2
    /// @return 加算結果
    template< class T >
    Bivariate< T > operator+( const Bivariate< T >& val1, const Bivariate< T >& val2 )
    {
      Bivariate< T > buff( val1 );
      return( buff += val2 );
    }

    /// @brief 二変数どうしの減算
    ///
    /// @param val1 被減算対象の二変数
    /// @param val2 減算対象の二変数
    /// @return 減算結果
    template< class T >
    Bivariate< T > operator-( const Bivariate< T >& val1, const Bivariate< T >& val2 )
    {
      Bivariate< T > buff( val1 );
      return( buff -= val2 );
    }

    /// @brief 二変数に同じ値を乗算する
    ///
    /// @param val 被乗算対象の二変数
    /// @param mul 乗算する値
    /// @return 乗算結果
    template< class T >
    Bivariate< T > operator*( const Bivariate< T >& val, T mul )
    {
      Bivariate< T > buff( val );
      return( buff *= mul );
    }

    /// @brief 二変数に同じ値を除算する
    ///
    /// @param val 被除算対象の二変数
    /// @param div 除算する値
    /// @return 除算結果
    template< class T >
    Bivariate< T > operator/( const Bivariate< T >& val, T div )
    {
      Bivariate< T > buff( val );
      return( buff /= div );
    }

  } // namespace numeric_lib

  namespace gadget
  {
    /**
       @brief 時間計測用オブジェクト
    **/
    class TimeMeasure
    {
      std::chrono::system_clock::time_point s_; // 計測開始日時
      std::chrono::system_clock::time_point e_; // 計測終了日時

    public:

      /* コンストラクタ */

      /// @brief 計測開始・終了日時を初期化して構築
      TimeMeasure() { s_ = std::chrono::system_clock::now(); e_ = s_; }

      /// @brief 計測開始日時の設定
      ///
      /// 処理開始前に呼び出して利用する
      ///
      /// @return なし
      void start() { s_ = std::chrono::system_clock::now(); }

      /// @brief 計測終了日時の設定
      ///
      /// 処理終了後に呼び出して利用する
      ///
      /// @return なし
      void end() { e_ = std::chrono::system_clock::now(); }

      /// @brief 計測開始・終了間の時間を求める
      ///
      /// テンプレート引数の T には次の型が利用できる
      /// - std::chrono::hours
      /// - std::chrono::minutes
      /// - std::chrono::seconds
      /// - std::chrono::milliseconds
      /// - std::chrono::microseconds
      /// - std::chrono::nanoseconds
      ///
      /// @return 計測開始から終了までの時間
      template< class T >
      typename T::rep time() const
      {
	auto dur = e_ - s_;

	return( std::chrono::duration_cast< T >( dur ).count() );
      }

      /// @brief 計測開始・終了間の時間を sec で求める
      ///
      /// @return 計測開始から終了までの時間(sec)
      double sec() const
      {
	auto t = time< std::chrono::nanoseconds >();

	return( static_cast< double >( t ) / 1E+9 );
      }

      /// @brief 計測開始・終了間の時間を出力する
      ///
      /// テンプレート引数の T には次の型が利用できる
      /// - std::chrono::hours
      /// - std::chrono::minutes
      /// - std::chrono::seconds
      /// - std::chrono::milliseconds
      /// - std::chrono::microseconds
      /// - std::chrono::nanoseconds
      ///
      /// @param header 結果の前に表示する文字列
      /// @param footer 結果の後に出力する文字列
      /// @return なし
      template< class T >
      void print( const std::string& header = "", const std::string& footer = "\n" ) const
      {
	std::cout << header << time< T >() << footer;
      }
    };

  } // namespace gadget

} // namespace tadah_fussy

#endif
