/**
   @file integral.h
   @brief 数値積分法定義用ヘッダ

   @author tadah_fussy
   @date 2016/07/26 テンプレート関数へ変更
   @date 2019/03/28 コメントの見直し
   @date 2021/03/14 名前空間の変更
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef INTEGRAL_20210402_H
#define INTEGRAL_20210402_H

#include <common/error_library.h>

/**
   @brief 作成者 tadah_fussy
**/
namespace tadah_fussy
{
  namespace math_lib
  {
    namespace integral
    {
      /**
	 @brief 中点則 ( Midpoint Rule ) による定積分計算用クラス

	 関数 math_lib::integral::Run へ引数として渡して利用する。

	 @code
	 double f( double x ) { ... }

	 // 0 から 1 までの f の定積分値を求める。分割数は 10000。
	 math_lib::integral::Run( math_lib::integral::Midpoint(), f, 0.0, 1.0, 10000 );
	 @endcode
      **/
      struct Midpoint
      {
	// 積分値の初期値設定
	//
	// f : 積分する関数
	// a, b : 積分する範囲
	// h : 矩形の幅
	template< class T, class F >
	T initValue( F f, T a, T b, T h )
	{ return( f( a + h / 2 ) ); }

	// x の初期値設定
	//
	// a : 積分範囲の開始位置
	// h : 矩形の幅
	template< class T >
	T initX( T a, T h )
	{ return( a + 3 * h / 2 ); }

	// 加算していく積分値の計算
	//
	// f : 積分する関数
	// x : f(x) を計算する x の値
	// h : 矩形の幅
	template< class T, class F >
	T operator()( F f, T x, T h )
	{ return( f( x ) ); }

	// 和を計算した後の補正
	//
	// val : 補正前の積分値
	// h : 矩形の幅
	template< class T >
	T cleanUp( T val, T h )
	{ return( val * h ); }
      };

      /**
	 @brief 台形則 ( Trapezoidal Rule ) による定積分計算用クラス

	 関数 math_lib::integral::Run へ引数として渡して利用する。

	 @code
	 double f( double x ) { ... }

	 // 0 から 1 までの f の定積分値を求める。分割数は 10000。
	 math_lib::integral::Run( math_lib::integral::Trapezoidal(), f, 0.0, 1.0, 10000 );
	 @endcode
      **/
      struct Trapezoidal
      {
	// 積分値の初期値設定
	//
	// f : 積分する関数
	// a, b : 積分する範囲
	// h : 矩形の幅
	template< class T, class F >
	T initValue( F f, T a, T b, T h )
	{ return( ( f( a ) + f( b ) ) / 2 ); }

	// x の初期値設定
	//
	// a : 積分範囲の開始位置
	// h : 矩形の幅
	template< class T >
	T initX( T a, T h )
	{ return( a + h ); }

	// 加算していく積分値の計算
	//
	// f : 積分する関数
	// x : f(x) を計算する x の値
	// h : 矩形の幅
	template< class T, class F >
	T operator()( F f, T x, T h )
	{ return( f( x ) ); }

	// 和を計算した後の補正
	//
	// val : 補正前の積分値
	// h : 矩形の幅
	template< class T >
	T cleanUp( T val, T h )
	{ return( val * h ); }
      };

      /**
	 @brief シンプソン則 ( Composite Simpson's Rule ) による定積分計算用クラス

	 関数 math_lib::integral::Run へ引数として渡して利用する。

	 @code
	 double f( double x ) { ... }

	 // 0 から 1 までの f の定積分値を求める。分割数は 10000。
	 math_lib::integral::Run( math_lib::integral::Simpson(), f, 0.0, 1.0, 10000 );
	 @endcode
      **/
      struct Simpson
      {
	// 積分値の初期値設定
	//
	// f : 積分する関数
	// a, b : 積分する範囲
	// h : 矩形の幅
	template< class T, class F >
	T initValue( F f, T a, T b, T h )
	{ return( f( a ) + 4 * f( b - h / 2 ) + f( b ) ); }

	// x の初期値設定
	//
	// a : 積分範囲の開始位置
	// h : 矩形の幅
	template< class T >
	T initX( T a, T h )
	{ return( a ); }

	// 加算していく積分値の計算
	//
	// f : 積分する関数
	// x : f(x) を計算する x の値
	// h : 矩形の幅
	template< class T, class F >
	T operator()( F f, T x, T h )
	{ return( 4 * f( x + h / 2 ) + 2 * f( x + h ) ); }

	// 和を計算した後の補正
	//
	// val : 補正前の積分値
	// h : 矩形の幅
	template< class T >
	T cleanUp( T val, T h )
	{ return( val * h / 6 ); }
      };

      /**
	 @brief シンプソンの 3/8 則 ( Simpson's 3/8 Rule ) による定積分計算用クラス

	 関数 math_lib::integral::Run へ引数として渡して利用する。

	 @code
	 double f( double x ) { ... }

	 // 0 から 1 までの f の定積分値を求める。分割数は 10000。
	 math_lib::integral::Run( math_lib::integral::Simpson38(), f, 0.0, 1.0, 10000 );
	 @endcode
      **/
      struct Simpson38
      {
	// 積分値の初期値設定
	//
	// f : 積分する関数
	// a, b : 積分する範囲
	// h : 矩形の幅
	template< class T, class F >
	T initValue( F f, T a, T b, T h )
	{ return( f( a ) + 3 * ( f( b - h / 3 ) + f( b - 2 * h / 3 ) ) + f( b ) ); }

	// x の初期値設定
	//
	// a : 積分範囲の開始位置
	// h : 矩形の幅
	template< class T >
	T initX( T a, T h )
	{ return( a ); }

	// 加算していく積分値の計算
	//
	// f : 積分する関数
	// x : f(x) を計算する x の値
	// h : 矩形の幅
	template< class T, class F >
	T operator()( F f, T x, T h )
	{ return( 3 * ( f( x + h / 3 ) + f( x + 2 * h / 3 ) ) + 2 * f( x + h ) ); }

	// 和を計算した後の補正
	//
	// val : 補正前の積分値
	// h : 矩形の幅
	template< class T >
	T cleanUp( T val, T h )
	{ return( val * h / 8 ); }
      };

      /**
	 @brief ブール則 ( Boole's Rule ) による定積分計算用クラス

	 関数 math_lib::integral::Run へ引数として渡して利用する。

	 @code
	 double f( double x ) { ... }

	 // 0 から 1 までの f の定積分値を求める。分割数は 10000。
	 math_lib::integral::Run( math_lib::integral::Boole(), f, 0.0, 1.0, 10000 );
	 @endcode
      **/
      struct Boole
      {
	// 積分値の初期値設定
	//
	// f : 積分する関数
	// a, b : 積分する範囲
	// h : 矩形の幅
	template< class T, class F >
	T initValue( F f, T a, T b, T h )
	{ return( 7 * f( a ) + 32 * ( f( b - 3 * h / 4 ) + f( b - h / 4 ) ) + 12 * f( b - h / 2 ) + 7 * f( b ) ); }

	// x の初期値設定
	//
	// a : 積分範囲の開始位置
	// h : 矩形の幅
	template< class T >
	T initX( T a, T h )
	{ return( a ); }

	// 加算していく積分値の計算
	//
	// f : 積分する関数
	// x : f(x) を計算する x の値
	// h : 矩形の幅
	template< class T, class F >
	T operator()( F f, T x, T h )
	{ return( 32 * ( f( x + h / 4 ) + f( x + 3 * h / 4 ) ) + 12 * f( x + h / 2 ) + 14 * f( x + h ) ); }

	// 和を計算した後の補正
	//
	// val : 補正前の積分値
	// h : 矩形の幅
	template< class T >
	T cleanUp( T val, T h )
	{ return( val * h / 90 ); }
      };

      /// @brief 区分求積法による積分値の計算
      ///
      /// 計算手法 ( 中点則・台形則など ) は引数 op にて指定する。以下の計算手法が現在利用できる。
      ///
      /// - Midpoint 中点則
      /// - Trapezoidal 台形則
      /// - Simpson シンプソン則
      /// - Simpson38 シンプソンの 3/8 則
      /// - Boole ブール則
      ///
      /// 積分範囲の大小関係が逆転している ( a > b の ) 場合、積分値の符号が逆転することに注意。
      ///
      /// - 分割数 n がゼロの場合、例外 exception::NotPositiveNumber< uint_32t > を投げる。
      ///
      /// @param op 積分値を計算する手法
      /// @param f 積分値を計算する対象の関数
      /// @param a,b 積分範囲
      /// @param n 分割数 ( 大きいほど精度が上がるが、処理時間は長くなる )
      template< class T, class F, class Op >
      T Run( Op op, F f, T a, T b, uint32_t n )
      {
	error_lib::PNum( n, "math_lib::integral::Run : n must be positive." );
	bool invSign; // 符号反転するか

	// a > b なら swap して後で符号を反転する
	if ( a > b ) {
	  std::swap( a, b );
	  invSign = true;
	} else {
	  invSign = false;
	}

	T h( ( b - a ) / n ); // 矩形の幅

	// 計算開始

	T ans = op.initValue( f, a, b, h ); // 積分値の初期状態
	T x = op.initX( a, h );       // x の初期値
	for ( uint32_t i = 1 ; i < n ; ++i ) {
	  ans += op( f, x, h ); // f(x) を求めて加算する
	  x += h;
	}

	// 符号反転
	if ( invSign ) ans *= -1;

	// 正しい積分値にする
	return( op.cleanUp( ans, h ) );
      }

    } // namespace integral

  } // namespace math_lib

} // namespace tadah_fussy

#endif
