/**
   @file math_library.h
   @brief 数学関数定義用ヘッダ

   @author tadah_fussy
   @date 2015/10/13 コメントを Doxygen 用に変更
   @date 2019/04/05 NewtonPolynomial のバグ修正
   @date 2019/04/19 DiffMethod_*** クラスの引数チェックエラー時をassert から例外に変更
   @date 2021/03/14 名前空間の変更、SquareRoot, Round, Gamma, Erf の削除など
   @date 2021/03/17 ニュートン・ラフソン法を optimization.h へ移動
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef MATH_LIBRARY_20210402_H
#define MATH_LIBRARY_20210402_H

#include <iostream>
#include <cmath>
#include <limits>
#include <numeric>
#include <map>
#include <vector>
#include <string>

#include <common/error_library.h>
#include <common/matrix.h>

/**
   @brief 作成者 tadah_fussy
**/
namespace tadah_fussy
{
  /**
     @brief 数学関数用名前空間
  **/
  namespace math_lib
  {
    /* --------------------------------------
       数値計算ライブラリ関数オブジェクト
       -------------------------------------- */

    /**
       @brief 差分法名前空間
    **/
    namespace diff_method
    {
      /**
	 @brief 差分法による導関数計算用関数

	 計算には前進差分 ( Forward Difference ) を使う

	 f'(x) = [ f( x + h ) - f( x ) ] / h

	 テンプレート引数の F は導関数計算対象の関数の型、T は関数に渡す変数の型をそれぞれ表す。
	 tadah_fussy::math_lib::diff_method::CreateForward を利用することでテンプレート引数を気にせずに生成することができる。
      **/
      template< class F, class T >
      class Forward
      {
	F f_; // 導関数を計算する対象の関数 f(x)
	T h_; // x の近傍を決めるための微小量

      public:

	/// @brief 関数f(x)と、xの近傍を決めるための微小量を指定して構築
	///
	/// - 微小量 h がゼロ以下の場合、例外 tadah_fussy::exception::NotPositiveNumber を投げる。
	///
	/// @param f 導関数を計算する対象の関数 f(x)
	/// @param h 微小量
	Forward( F f, T h )
	  : f_( f ), h_( h )
	{
	  const std::string FUNC_NAME = "tadah_fussy::math_lib::diff_method::Forward constructor";
	  error_lib::PNum( h_, FUNC_NAME + " : h " + error_message::PNUM );
	}

	/// @brief xでの導関数 f'(x) の近似値を求める
	///
	/// @param x 導関数を求める対象のx値
	/// @return f'(x)の近似値
	T operator()( T x )
	{ return( ( f_( x + h_ ) - f_( x ) ) / h_ ); }
      };

      /**
	 @brief 差分法による導関数計算用関数

	 計算には後退差分 ( Backward Difference ) を使う

	 f'(x) = [ f( x ) - f( x - h ) ] / h

	 テンプレート引数の F は導関数計算対象の関数の型、T は関数に渡す変数の型をそれぞれ表す。
	 tadah_fussy::math_lib::diff_method::CreateBackward を利用することでテンプレート引数を気にせずに生成することができる。
      **/
      template< class F, class T >
      class Backward
      {
	F f_; // 導関数を計算する対象の関数 f(x)
	T h_; // x の近傍を決めるための微小量

      public:

	/// @brief xの近傍を決めるための微小量を指定して構築
	///
	/// - 微小量 h がゼロ以下の場合、例外 tadah_fussy::exception::NotPositiveNumber< T > を投げる。
	///
	/// @param f 導関数を計算する対象の関数 f(x)
	/// @param h 微小量
	Backward( F f, T h )
	  : f_( f ), h_( h )
	{
	  const std::string FUNC_NAME = "tadah_fussy::math_lib::diff_method::Backward constructor";
	  error_lib::PNum( h_, FUNC_NAME + " : h " + error_message::PNUM );
	}

	/// @brief xでの導関数 f'(x) の近似値を求める
	///
	/// @param x 導関数を求める対象のx値
	/// @return f'(x)の近似値
	T operator()( T x )
	{ return( ( f_( x ) - f_( x - h_ ) ) / h_ ); }
      };

      /**
	 @brief 差分法による導関数計算用関数

	 計算には中心差分を使う

	 f'(x) = [ f( x + h / 2 ) - f( x - h / 2 ) ] / h

	 テンプレート引数の F は導関数計算対象の関数の型、T は関数に渡す変数の型をそれぞれ表す。
	 tadah_fussy::math_lib::diff_method::CreateCentral を利用することでテンプレート引数を気にせずに生成することができる。
      **/
      template< class F, class T >
      class Central
      {
	F f_; // 導関数を計算する対象の関数 f(x)
	T h_; // x の近傍を決めるための微小量

      public:

	/// @brief xの近傍を決めるための微小量を指定して構築
	///
	/// - 微小量 h がゼロ以下の場合、例外 tadah_fussy::exception::NotPositiveNumber< T > を投げる。
	///
	/// @param f 導関数を計算する対象の関数 f(x)
	/// @param h 微小量
	Central( F f, T h )
	  : f_( f ), h_( h )
	{
	  const std::string FUNC_NAME = "tadah_fussy::math_lib::diff_method::Central constructor";
	  error_lib::PNum( h_, FUNC_NAME + " : h " + error_message::PNUM );
	}

	/// @brief xでの導関数 f'(x) の近似値を求める
	///
	/// @param x 導関数を求める対象のx値
	/// @return f'(x)の近似値
	T operator()( T x )
	{ return( ( f_( x + h_ / T( 2 ) ) - f_( x - h_ / T( 2 ) ) ) / h_ ); }
      };

      /// @brief 前進差分による導関数計算用関数オブジェクト tadah_fussy::math_lib::diff_method::Forward を生成する
      ///
      /// @param f 導関数を計算する対象の関数 f(x)
      /// @param h 微小量
      template< class F, class T >
      Forward< F, T >
      CreateForward( F f, T h )
      { return( Forward< F, T >( f, h ) ); }

      /// @brief 後退差分による導関数計算用関数オブジェクト tadah_fussy::math_lib::diff_method::Backward を生成する
      ///
      /// @param f 導関数を計算する対象の関数 f(x)
      /// @param h 微小量
      template< class F, class T >
      Backward< F, T >
      CreateBackward( F f, T h )
      { return( Backward< F, T >( f, h ) ); }

      /// @brief 中心差分による導関数計算用関数オブジェクト tadah_fussy::math_lib::diff_method::Central を生成する
      ///
      /// @param f 導関数を計算する対象の関数 f(x)
      /// @param h 微小量
      template< class F, class T >
      Central< F, T >
      CreateCentral( F f, T h )
      { return( Central< F, T >( f, h ) ); }

    } // namespace diff_method

    /// @brief t の n 乗を計算する関数
    ///
    /// テンプレート引数の U は、乗法・除法・減算・剰余・等号演算子が使える型である必要がある。
    /// 剰余演算子を使うため、整数型であることを前提としている。
    ///
    /// 0 の 0 乗は 0 ではなく 1 を返す。
    ///
    /// - べき数が負数の場合は例外 tadah_fussy::exception::NegativeNumber< U > を投げる。
    ///
    /// @param t べき乗される数
    /// @param n べき数
    /// @return 計算結果
    template< class U > U Pow( const U& t, const U& n )
    {
      const std::string FUNC_NAME = "tadah_fussy::math_lib::Pow";

      error_lib::NNNum( n, FUNC_NAME + " : n " + error_message::NNNUM );

      if ( n == U() ) return( 1 );
      if ( n % 2 == U() ) {
        U t2 = Pow( t, n / 2 );
        return( t2 * t2 );
      } else {
        return( t * Pow( t, n - 1 ) );
      }
    }

    /// @brief c の e 乗根の近似値を求める
    ///
    /// ニュートン-ラフソン法を利用したべき乗根算出ルーチン。
    /// c は符号なし整数であることを前提とし、c の e 乗根を超えない最大数を返す。
    /// 但し、e が U() の場合は U() を返す。
    ///
    /// @param c べき乗根を求める対象の数
    /// @param e べき数
    /// @return c の e 乗根
    template< class U >
    U PowerRoot( const U& c, const U& e )
    {
      if ( e == U() ) return( U() );
      if ( e == 1 ) return( c );

      U a = c;
      U b = ( c + 1 ) / 2;

      while ( b < a ) {
        a = b;
        b = ( ( e - 1 ) * Pow( a, e ) + c ) / ( e * Pow( a, e - 1 ) );
      }

      return( a );
    }

    /// @brief 繰り返し自乗法を使った法 n のべき乗計算( a の k 乗を n で割った余りを求める )
    ///
    /// テンプレート引数の U は、乗法・剰余・ビットシフト・AND・比較演算子が使える型である必要がある。
    /// 剰余演算子やビットシフトなどを使うため、符号なし整数型であることを前提としている。
    ///
    /// - 底 a や法 n が 0 の場合は 0 を返す。
    /// - 指数が 0 の場合は法 n における 1 を返す。
    ///
    /// @param a 底
    /// @param k 指数
    /// @param n 法
    /// @return べき乗
    template< class U >
    U ModularPower( const U& a, U k, const U& n )
    {
      if ( a == U() || n == U() ) return( U() ); // 底や法が 0 の場合は 0 を返す
      if ( k == U() ) return( 1 % n );       // 指数が 0 の場合は法 n における 1

      U mod( a % n ); // n を法とした a
      U ans( ( ( k & 1 ) > U() ) ? mod : 1 ); // 求める値

      for ( k >>= 1 ; k > U() ; k >>= 1 ) {
        mod = ( mod * mod ) % n;
        if ( ( k & 1 ) > U() )
          ans = ( ans * mod ) % n;
      }

      return( ans );
    }

    /// ユークリッドの互除法を使った最大公約数の計算
    ///
    /// テンプレート引数の U は、剰余・比較演算子が使える型である必要がある。
    /// 剰余演算子を使うため、符号なし整数型であることを前提としている。
    ///
    /// - いずれかがゼロの場合、もう一方の数をそのまま返す。
    ///
    /// @param a, b 最大公約数を求める二つの自然数
    /// @return 最大公約数
    template< class U >
    U GCD( U a, U b )
    {
      if ( a < b ) std::swap( a, b );
      if ( b == U() ) return( a );

      do {
        U r = a % b;
        a = b;
        b = r;
      } while ( b != U() );

      return( a );
    }

    /// @brief ヤコビ記号を求める
    ///
    /// a が p を法として 0 でない数の平方数と合同になるとき、a を「p を法とした平方剰余」という。
    ///
    /// a が p を法として平方剰余(1)か平方非剰余(-1)かを判定する。この判定値を「ヤコビ記号」という。
    /// 但し、次の場合は 0 を返す。
    ///
    /// - a と p が互いに素でない場合
    /// - p が 3 以上の奇数でない場合
    ///
    /// @param a 判定対象の数
    /// @param p 法
    /// @return 1 ... 平方剰余 ; -1 ... 平方非剰余 ; 0 ... GCD(a,p) != 1
    template< class U >
    int8_t Jacobi( U a, U p )
    {
      if ( ( p == 1 ) || ( ( p & 1 ) == U() ) ) return( 0 ); // pが3以上の奇数でなければ 0を返す
      if ( GCD( a, p ) != 1 ) return( 0 );

      // (1/p) = 1
      if ( a == 1 ) return( 1 );

      // (2/p) =  1, ( p ≡ 1 or 7 (mod 8) )
      //       = -1, ( p ≡ 3 or 5 (mod 8) )
      if ( a == 2 ) {
        U modP = p % 8;
        if ( modP == 1 || modP == 7 )
          return( 1 );
        else
          return( -1 );
      }

      // aから因数2を除外する
      U evenCnt{};
      while ( ( a & 1 ) == 0 ) {
        a >>= 1;
        ++evenCnt;
      }
      int8_t ans = ( ( evenCnt & 1 ) != 0 ) ? Jacobi( U( 2 ), p ) : 1;

      // (a/p) =  (p/a), ( p ≡ 1 or a ≡ 1 (mod 4) )
      //       = -(p/a), ( p ≡ a ≡ 3 (mod 4) )
      if ( p % 4 == 3 && a % 4 == 3 )
        ans *= -1;
      if ( a == 1 ) return( ans );

      p %= a;

      return( ans * Jacobi( p, a ) );
    }

    /// @brief 円周率を返す関数
    ///
    /// 定数 M_PI(double/float用) と M_PIl(long double用) を返す。明示的にテンプレート引数を記述する必要がある。
    ///
    /// @return 円周率
    template< class T > T Pi();
    template<> inline double Pi< double >()
    { return( M_PI ); }
    template<> inline long double Pi< long double >()
    { return( M_PIl ); }
    template<> inline float Pi< float >()
    { return( M_PI ); }

    /// @brief 順列 ( n 個の要素から r 個を並べる場合の数 ) の計算
    ///
    /// nPr = n! / ( n - r )! = n( n - 1 )...( n - r + 1 ) を正確に計算する。
    /// 計算可能な範囲は戻り値の型 Res に依存し、Res の取りうる値の範囲を超えた場合は正しい結果は得られない。
    /// テンプレート引数 Res は明示的に記述する必要があることに注意。
    ///
    /// - n < r または r < 0 ならゼロを返す。
    ///
    /// @param n 全要素数
    /// @param r 並べる要素の数
    /// @return 順列の計算結果(n < r または r < 0 ならゼロ)
    template< class Res, class I > Res AccuratePermutation( I n, I r )
    {
      if ( n < r ) return( Res() );  // n < r なら nPr = 0
      if ( r < 0 ) return( Res() );  // nP-r = 0
      if ( r == 0 ) return( Res( 1L ) ); // nP0 = 1

      Res ans = n; // 計算結果

      // n( n - 1 )...( n - r + 1 ) を求める
      for ( Res t = n - r + 1 ; t < n ; ++t )
        ans *= t;

      return( ans );
    }

    /// @brief 順列 ( n 個の要素から r 個を並べる場合の数 ) の計算
    ///
    /// nPr = n! / ( n - r )! = n( n - 1 )...( n - r + 1 ) の近似値を計算する。
    ///
    /// - n < r または r < 0 ならゼロを返す。
    ///
    /// @param n 全要素数
    /// @param r 並べる要素の数
    /// @return 順列の計算結果(n < r または r < 0 ならゼロ)
    template< class T > T Permutation( T n, T r )
    {
      return( ( n < r || r < 0 ) ? T() : std::tgamma( n + 1 ) / std::tgamma( n - r + 1 ) );
    }

    /// @brief m 個のグループからそれぞれ k[1] 〜 k[m] 個ずつ合計 n 個並べる場合の数(多項係数)を求める
    ///
    /// n! / k[1]!k[2]!...k[m]! を正確に計算する。但し、n は k[1] から k[m] までの数の和を意味する。
    /// これは多項定理における項 Πi( x^k[i] ) の係数を表す。
    /// 計算可能な範囲は戻り値の型 Res に依存し、Res の取りうる値の範囲を超えた場合は正しい結果は得られない。
    ///
    /// k[1] から k[m] の中に一つでも負数があった場合はゼロを返す。
    ///
    /// - 配列 [ s, e ) の位置が逆転していた場合は例外 tadah_fussy::exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
    ///
    /// @param s 各グループの要素数からなる配列の開始位置
    /// @param e 各グループの要素数からなる配列の末尾の次の位置
    /// @return 多項係数の計算結果(要素数に負数が含まれていた場合はゼロ)
    template< class Res, class In >
    Res AccurateMultiCoef( In s, In e )
    {
      const std::string FUNC_NAME = "AccurateMultiCoef";

      auto diff = error_lib::ArrayLength( s, e, FUNC_NAME + " : [ s, e ) " + error_message::NEGATIVE_LENGTH );

      if ( diff == 0 ) return( Res() );    // 要素数がゼロなら結果もゼロ

      Res n = Res(); // 並べる合計の数
      std::map< Res, Res > mapFact; // 0! 〜 n! の階乗リスト
      for ( In i = s ; i != e ; ++i ) {
        if ( *i < 0 ) return( 0 ); // 一つでも要素数が負数のものがあった場合はゼロを返す
        if ( *i == 0 ) continue; // 要素数がゼロなら結果には影響しないので無視
        n += *i;
        if ( mapFact.find( *i ) == mapFact.end() )
          mapFact[*i] = Res();
      }

      // 階乗の計算
      Res fact = Res( 1 ); // 最後に n! が得られる
      for ( Res r = 1 ; r <= n ; ++r ) {
        fact *= r;
        if ( mapFact.find( r ) != mapFact.end() )
          mapFact[r] = fact;
      }

      // 各要素数の階乗で徐算
      for ( ; s != e ; ++s )
        if ( *s > 0 ) fact /= mapFact[*s];

      return( fact );
    }

    /// @brief m 個のグループからそれぞれ k[1] 〜 k[m] 個ずつ合計 n 個並べる場合の数(多項係数)を求める
    ///
    /// n! / k[1]!k[2]!...k[m]! の近似値を計算する。但し、n は k[1] から k[m] までの数の和を意味する。
    /// これは多項定理における項 Πi( x^k[i] ) の係数を表す。
    /// 計算可能な範囲は戻り値の型 Res に依存し、Res の取りうる値の範囲を超えた場合は正しい結果は得られない。
    ///
    /// - k[1] から k[m] の中に一つでも負数があった場合はゼロを返す。
    ///
    /// @param s 各グループの要素数からなる配列の開始位置
    /// @param e 各グループの要素数からなる配列の末尾の次の位置
    /// @return 多項係数の計算結果(要素数に負数が含まれていた場合はゼロ)
    template< class Res, class In >
    Res MultiCoef( In s, In e )
    {
      const std::string FUNC_NAME = "MultiCoef";

      auto diff = error_lib::ArrayLength( s, e, FUNC_NAME + " : [ s, e ) " + error_message::NEGATIVE_LENGTH );

      if ( diff == 0 ) return( Res() );    // 要素数がゼロなら結果もゼロ

      Res n = Res(); // 並べる合計の数
      Res product = Res( 1 ); // 分母の値
      for ( In i = s ; i != e ; ++i ) {
        if ( *i < 0 ) return( 0 ); // 一つでも要素数が負数のものがあった場合はゼロを返す
        if ( *i == 0 ) continue; // 要素数がゼロなら結果には影響しないので無視
        n += *i;
        product *= std::tgamma< Res >( *i + 1 );
      }

      return( std::tgamma< Res >( n + 1 ) / product );
    }

    /// @brief 組み合わせ ( n 個の要素から r 個を選ぶ場合の数 ) の計算
    ///
    /// nCr = n! / r!( n - r )! を正確に計算する。
    /// 計算可能な範囲は戻り値の型 Res に依存し、Res の取りうる値の範囲を超えた場合は正しい結果は得られない。
    ///
    /// - n < r ならゼロを返す。
    /// - r < 0 ならゼロを返す。
    ///
    /// @param n 要素数
    /// @param r 要素から選択する数
    /// @return 組み合わせの計算結果
    template< class Res, class T >
    Res AccurateCombination( T n, T r )
    {
      if ( n < r ) return( Res() );   // n < r なら nCr = 0
      if ( r < 0 ) return( Res() );   // nC-r = 0

      if ( r > n - r ) r = n - r; // nCn-r = nCr

      if ( r == 0 ) return( Res( 1 ) );  // nC0 = 1
      if ( r == 1 ) return( Res( n ) );  // nC1 = n

      // まずは順列 n( n - 1 )...( n - r + 1 ) を求める
      Res ans = n;
      for ( T t = n - r + 1 ; t < n ; ++t )
        ans *= t;

      // 階乗 r! を求める
      Res fact = r;
      for ( --r ; r > 1 ; --r )
        fact *= r;

      ans /= fact;

      return( ans );
    }

    /// @brief 一般の二項係数の計算
    ///
    /// C( a, r ) = Γ( a + 1 ) / Γ( r + 1 )Γ( a - r + 1 ) を近似的に計算する。このとき、
    ///
    /// ( 1 + x )^a = 1 + C( a, 1 )x + C( a, 2 )x^2 + ... + C( a, r )x^r + ...
    ///
    /// が成り立つ。
    ///
    /// - r < 0 ならゼロを返す。
    /// - a + 1 <= r ならゼロを返す
    ///
    /// @param a 指数部
    /// @param r 次数
    /// @return 二項係数の計算結果
    template< class D, class I > D BinCoef( D a, I r )
    {
      if ( r < 0 ) return( D() ); // C( a, -r ) = 0

      // C( -a, r ) = C( a + r - 1, r )
      if ( a < 0 )
        return( BinCoef< D, I >( -a + r - 1, r ) * ( ( r % 2 != 0 ) ? -1 : 1 ) );

      // 展開後の項に対する次数 r が左辺の次数 a ( >0 ) より大きい場合はゼロとする
      if ( a + 1 <= r ) return( D() );

      if ( r == 1 ) return( a ); // C( a, 1 ) = a

      return( std::tgamma( a + 1 ) / ( std::tgamma( a - r + 1 ) * std::tgamma( r + 1 ) ) );
    }

    /// @brief ベルヌーイ数 Bn を求める
    ///
    /// ベルヌーイ数の計算に利用する組み合わせ(二項係数)は math_lib::AccurateCombination を使って正確に求めている。
    ///
    /// @param n 何番目のベルヌーイ数か？
    /// @return ベルヌーイ数 Bn
    template< class Res, class I >
    Res AccurateBernoulliNum( I n )
    {
      if ( n < I() ) return( Res() );
      if ( n == I() ) return( Res( 1 ) );

      I np( n + 1 );
      Res b = Res();
      for ( I i = I() ; i < n ; ++i )
        b -= Res( AccurateCombination< Res, I >( np, i ) ) * AccurateBernoulliNum< Res, I >( i );

      b /= Res( np );

      return( b );
    }

    /// @brief ベルヌーイ数 Bn を求める
    ///
    /// ベルヌーイ数の計算に利用する組み合わせ(二項係数)は math_lib::BinCoef を使って近似的に求めている。
    ///
    /// @param n 何番目のベルヌーイ数か？
    /// @return ベルヌーイ数 Bn
    template< class Res, class I >
    Res BernoulliNum( I n )
    {
      if ( n < I() ) return( Res() );
      if ( n == I() ) return( Res( 1 ) );

      I np( n + 1 );
      Res b = Res();
      for ( I i = I() ; i < n ; ++i )
        b -= BinCoef< Res, I >( np, i ) * BernoulliNum< Res, I >( i );

      b /= Res( np );

      return( b );
    }

    /**
       @brief 多項式定義用名前空間
    **/
    namespace polynomial
    {
      /**
	 @brief Bernstein多項式

	 Bernstein多項式は以下の基底で定義される多項式である。

	 b( k, N ; t ) = C( N, k )・t^k・( 1 - t )^(N-k)

	 但し、C( N, k ) は二項係数 C( N , k ) = N! / k!( N - k )! を表す。
      **/
      template< class T >
      class Bernstein
      {
      public:

	using size_type = typename std::vector< T >::size_type; ///< 要素へアクセスするときの添字の型

      private:

	size_type n_; // 次数
	std::vector< T > binCoeff_; // 二項係数 ( 0-n_ までの n_+1 個 )

      public:

	/// @brief 次数 N を指定して構築
	///
	/// @param n 次数
	Bernstein( size_type n )
	{ reset( n ); }

	/// @brief 次数を変更する(変数の初期化を行う)
	///
	/// @param n 次数
	void reset( size_type n )
	{
	  n_ = n;
	  binCoeff_.resize( n_ + 1 );
	  for ( size_type k = 0 ; k <= n_ ; ++k )
	    binCoeff_[k] = BinCoef< T, size_type >( n_, k );
	}

	/// @brief Bernstein基底関数 b( k, N ; t ) の値を求める
	///
	/// - パラメータ k が次数より大きい場合はゼロを返す。
	///
	/// @param k パラメータ k
	/// @param t 変数 t
	/// @return 計算結果
	T operator()( size_type k, T t )
	{ return( ( k > n_ ) ? 0 : binCoeff_[k] * pow( t, k ) * pow( 1 - t, n_ - k ) ); }
      };

      /**
	 @brief Lagrange多項式

	 Lagrange多項式は以下の多項式とサンプル点の積和で表される。

	 <pre>
	 l_i( x ) = Πj,j≠i ( x - x_j ) / ( x_i - x_j )
	 = ( x - x_0 )( x - x_1 )...( x - x_i-1 )( x - x_i+1 )...
	 / ( x_i - x_0 )( x_i - x_1 )...( x_i - x_i-1 )( x_i - x_i+1 )...
	 L( x ) = Σi y_i・l_i( x )
	 </pre>

	 ( x_i, y_i ) がサンプル点を表す。

	 @note x_i &ne; x_j ( i &ne; j ) を満たす必要があることに注意。内部ではそのチェックは行っていない。
      **/
      template< class T >
      class Lagrange
      {
	// 添字の差分の型
	using size_type = typename std::vector< T >::size_type;

	std::vector< T > xi_; // サンプル点の x 座標
	std::vector< T > yi_; // サンプル点の y 座標

      public:

	/// @brief サンプル点の座標を指定して構築
	///
	/// 指定したサンプル点に対して以下のチェックを行う。
	///
	/// -# X, Y の要素の範囲が適切か(開始と終了が逆になっていないか)
	/// -# X と Y の要素数が一致しているか
	///
	/// 1 のテストにパスしなかった場合は例外 tadah_fussy::exception::NegativeLength が、2 のテストに パスしなかった場合は例外 tadah_fussy::exception::NotEqualLength が投げられる。
	///
	/// テンプレート引数の In1 と In2 はそれぞれ二つのデータ列の反復子の型を表す。
	/// tadah_fussy::exception::NegativeLength のテンプレート引数は、[ sx, ex ) が不正だった場合 std::iterator_traits< In1 >::difference_type に、
	/// [ sy, ey ) が不正だった場合 std::iterator_traits< In2 >::difference_type になる。
	/// また、tadah_fussy::exception::NotEqualLength のテンプレート引数は必ず std::iterator_traits< In1 >::difference_type になる。
	///
	/// @param sx,ex サンプル点の X 座標の数列
	/// @param sy,ey サンプル点の Y 座標の数列
	template< class In1, class In2 >
	Lagrange( In1 sx, In1 ex, In2 sy, In2 ey )
	{
	  const std::string FUNC_NAME = "LagrangePolynomial constructor";

	  error_lib::CheckBivariable( sx, ex, sy, ey,
				      FUNC_NAME + " : [ sx, ex ) " + error_message::NEGATIVE_LENGTH,
				      FUNC_NAME + " : [ sy, ey ) " + error_message::NEGATIVE_LENGTH,
				      FUNC_NAME + " : [ sx, ex ) and [ sy, ey ) " + error_message::NOT_EQUAL_LENGTH );
	  xi_.assign( sx, ex );
	  yi_.assign( sy, ey );
	}

	/// @brief 指定した X 座標での Y の値を返す
	///
	/// @param x X 座標
	/// @return Y 座標の値
	T operator()( T x ) const
	{
	  size_type n = yi_.size();

	  T ans = T();
	  for ( size_type i = 0 ; i < n ; ++i ) {
	    T pi( 1 );
	    for ( size_type j = 0 ; j < n ; ++j ) {
	      if ( j == i ) continue;
	      pi *= ( x - xi_[j] ) / ( xi_[i] - xi_[j] );
	    }
	    ans += yi_[i] * pi;
	  }

	  return( ans );
	}
      };

      /**
	 @brief Newton多項式

	 Newton多項式は以下の多項式とサンプル点の積和で表される。
	 <pre>
	 N( t ) = Σi [y_0,y_1,...y_i]Π0≦j≦i-1 ( x - x_j )
	 </pre>
	 [y_0,y_1,...y_i] は i 階差分商といい、以下の漸化式が成り立つ。
	 <pre>
	 [y_0] = y_0, [y_0,y_1] = ( y_1 - y_0 ) / ( x_1 - x_0 )
	 [y_0,y_1,...y_i] = { [y_1,y_2,...y_i] - [y_0,y_1,...y_i-1] } / ( x_i - x_0 ) 
	 </pre>

	 @note x_i &ne; x_j ( i &ne; j ) を満たす必要があることに注意。内部ではそのチェックは行っていない。
      **/
      template< class T >
      class Newton
      {
	using difference_type = typename std::vector< T >::difference_type;

	std::vector< T > xi_;   // X 座標
	std::vector< T > diff_; // 差分商

      public:

	/// @brief サンプル点の座標を指定して構築
	///
	/// 指定したサンプル点に対して以下のチェックを行う。
	///
	/// -# X, Y の要素の範囲が適切か(開始と終了が逆になっていないか)
	/// -# X と Y の要素数が一致しているか
	///
	/// 1 のテストにパスしなかった場合は例外 tadah_fussy::exception::NegativeLength が、2 のテストに パスしなかった場合は例外 tadah_fussy::exception::NotEqualLength が投げられる。
	///
	/// テンプレート引数の In1 と In2 はそれぞれ二つのデータ列の反復子の型を表す。
	/// tadah_fussy::exception::NegativeLength のテンプレート引数は、[ sx, ex ) が不正だった場合 std::iterator_traits< In1 >::difference_type に、
	/// [ sy, ey ) が不正だった場合 std::iterator_traits< In2 >::difference_type になる。
	/// また、tadah_fussy::exception::NotEqualLength のテンプレート引数は必ず std::iterator_traits< In1 >::difference_type になる。
	///
	/// @param sx,ex サンプル点の X 座標の数列
	/// @param sy,ey サンプル点の Y 座標の数列
	template< class In1, class In2 >
	Newton( In1 sx, In1 ex, In2 sy, In2 ey )
	{
	  const std::string FUNC_NAME = "NewtonPolynomial constructor";

	  difference_type n = error_lib::CheckBivariable( sx, ex, sy, ey,
							  FUNC_NAME + " : [ sx, ex ) " + error_message::NEGATIVE_LENGTH,
							  FUNC_NAME + " : [ sy, ey ) " + error_message::NEGATIVE_LENGTH,
							  FUNC_NAME + " : [ sx, ex ) and [ sy, ey ) " + error_message::NOT_EQUAL_LENGTH );

	  // X 座標と差分商の初期化
	  xi_.assign( sx, ex );
	  diff_.assign( sy, ey );

	  for ( difference_type i = 1 ; i < n ; ++i ) {
	    // 差分商の計算
	    for ( difference_type j = n ; j > i ; --j )
	      diff_[j-1] = ( diff_[j-1] - diff_[j-2] ) / ( xi_[j-1] - xi_[j-i-1] );
	  }
	}

	/// @brief 指定した X 座標での Y の値を返す
	///
	/// @param x X 座標
	/// @return Y 座標の値
	T operator()( T x ) const
	{
	  difference_type n = diff_.size();
	  if ( n == 0 ) return( T() );
	  T ans = diff_[0];

	  T d = x - xi_[0];
	  for ( difference_type i = 1 ; i < n ; ++i ) {
	    ans += diff_[i] * d;
	    d *= x - xi_[i];
	  }

	  return( ans );
	}
      };

    } // namespace polynomial

    /* 総和計算用関数 */

    /**
       @brief 二つの値の和を計算するための関数オブジェクト

       渡された数を加算して返す。std::plus と同様の操作を行うが、
       std::pair に対して加算処理ができるよう、個別に作成している。

       ループ処理により和の計算を行うことを想定した関数オブジェクトであり、例えば

       @code
       std::accumulate( s, e, T(), SumOp< T >() );
       @endcode

       とすれば、T の型が std::pair であっても処理できる。
    **/
    template< class T > class SumOp
    {
    public:

      /// @brief 加算を行う
      ///
      /// @param init 被加算値
      /// @param value 加算する値
      /// @return 加算結果
      T operator()( T init, T value )
      { return( init + value ); }
    };

    /**
       @brief 二つの値の和を計算するための関数オブジェクト( std::pair用 )

       渡された std::pair の first, second それぞれを加算して返す。

       ループ処理により和の計算を行うことを想定した関数オブジェクトであり、例えば

       @code
       std::accumulate( s, e, T(), SumOp< T >() );
       @endcode

       とすれば、T の型が std::pair であっても処理できる。
    **/
    template< class T, class U > class SumOp< std::pair< T, U > >
    {
    public:

      /// @brief 加算を行う
      ///
      /// @param init 被加算値
      /// @param value 加算する値
      /// @return 加算結果
      std::pair< T, U > operator()( std::pair< T, U > init, std::pair< T, U > value )
      { return( std::pair< T, U >( init.first + value.first, init.second + value.second ) ); }
    };

    /**
       @brief Kahanの加算アルゴリズムを使って和を計算するための関数オブジェクト

       渡された数を加算して返す。誤差を軽減するために Kahan の加算アルゴリズムを利用

       ループ処理により和の計算を行うことを想定した関数オブジェクトであり、例えば

       @code
       std::accumulate( s, e, T(), KahanSumOp< T >() );
       @endcode

       とすれば、T の型が std::pair であっても処理できる。
    **/
    template< class T > class KahanSumOp
    {
      T r_; // 積み残し

    public:

      /// @brief デフォルト・コンストラクタ
      KahanSumOp()
	: r_( T() ) {}

      /// @brief 加算結果の初期化
      ///
      /// @return なし
      void reset()
      { r_ = T(); }

      /// @brief 加算を行う
      ///
      /// @param init 被加算値
      /// @param value 加算する値
      /// @return 加算結果
      T operator()( T init, T value )
      {
	T y = value - r_;  // 計算値から積み残しを引いた値
	T buff = init + y; // 今までの合計値に y を加える
	r_ = ( buff - init ) - y; // 積み残しを計算
	return( buff );
      }
    };

    /**
       @brief Kahanの加算アルゴリズムを使って和を計算するための関数オブジェクト(std::pair用)

       渡された std::pair の first, second それぞれを加算して返す。
       誤差を軽減するために Kahan の加算アルゴリズムを利用

       ループ処理により和の計算を行うことを想定した関数オブジェクトであり、例えば

       @code
       std::accumulate( s, e, T(), KahanSumOp< T >() );
       @endcode

       とすれば、T の型が std::pair であっても処理できる。
    **/
    template< class T, class U > class KahanSumOp< std::pair< T, U > >
    {
      std::pair< T, U > r_; // 積み残し

    public:

      /// @brief デフォルト・コンストラクタ
      KahanSumOp()
	: r_( std::pair< T, U >( T(), U() ) ) {}

      /// @brief 加算結果の初期化
      /// @return なし
      void reset()
      { r_.first = T(); r_.second = U(); }

      /// @brief 加算を行う
      ///
      /// @param init 被加算値
      /// @param value 加算する値
      /// @return 加算結果
      std::pair< T, U > operator()( std::pair< T, U > init, std::pair< T, U > value )
      {
	std::pair< T, U > y( value.first - r_.first, value.second - r_.second ); // 計算値から積み残しを引いた値
	std::pair< T, U > buff( init.first + y.first, init.second + y.second );  // 今までの合計値に y を加える

	// 積み残しを計算
	r_.first = ( buff.first - init.first ) - y.first;
	r_.second = ( buff.second - init.second ) - y.second;

	return( buff );
      }
    };

    /**
       @brief 単項関数で値を処理してから加算するための関数オブジェクト

       渡された数を処理してから加算して返す。
       ループ処理により和の計算を行うことを想定した関数オブジェクトであり、例えば

       @code
       using Func = double ( * )( double, double );

       std::accumulate( s, e, 0.0,
       math_lib::OpBind(
       std::bind( std::ptr_fun( static_cast< Func >( std::pow ) ), _1, 2.0 )
       )
       );
       @endcode

       とすれば、[ s, e ) に対し二乗和を計算することができる。

       テンプレート引数の Res は戻り値の型、Op は加算前の処理を行なう関数を表す。
    **/
    template< class Op > class AddBinder
    {
      Op op_; // 加算前に処理を行なうための関数

    public:

      /// @brief 処理用の関数を指定して構築
      ///
      /// @param op 加算する値を処理するための単項関数
      AddBinder( Op op ) : op_( op ) {}

      /// @brief 値 value を処理してから init に加算する
      ///
      /// テンプレート引数の Res は戻り値及び初期値の型、T は単項関数で処理する値の型をそれぞれ表す。
      ///
      /// @param init 初期値
      /// @param value 処理する値
      /// @return 加算結果
      template< class Res, class T >
      Res operator()( Res init, T value )
      { return( init + op_( value ) ); }
    };

    /**
       @brief 単項関数と二項関数の組み合わせにより値を処理するための関数オブジェクト

       渡された数を単項関数で処理してから加算などの二項関数を使って初期値に作用させる。
       ループ処理により和の計算などを行うことを想定した関数オブジェクトであり、例えば

       @code
       using Func = double ( * )( double, double );

       std::accumulate( s, e, 0.0,
       math_lib::OpBind< double >(
       std::bind( std::ptr_fun( static_cast< Func >( std::pow ) ), _1, 2.0 ),
       KahanSumOp< double >()
       )
       );
       @endcode

       とすれば、[ s, e ) に対し KahanSumOp を使って二乗和を計算することができる。

       テンプレート引数の Res は戻り値の型、UnOp は単項関数、BinOp は二項関数を表す。
    **/
    template< class Res, class UnOp, class BinOp > class OpBinder
    {
      UnOp unOp_;   // 単項関数
      BinOp binOp_; // 二項関数

    public:

      /// @brief 処理用の関数と加算用の関数を指定して構築
      ///
      /// @param unOp 単項関数
      /// @param binOp 二項関数
      OpBinder( UnOp unOp, BinOp binOp ) : unOp_( unOp ), binOp_( binOp ) {}

      /// @brief 値 value を単項関数で処理してから二項関数により init に作用させる
      ///
      /// テンプレート引数の T は初期値の型、U は処理する値の型をそれぞれ表す。
      /// 単項関数が U 型の値を処理して T 型の値を出力し、二項関数によって T 型の値から Res 型の値が出力されることを想定している。
      ///
      /// @param init 初期値
      /// @param value 処理する値
      /// @return 処理結果
      template< class T, class U >
      Res operator()( T init, U value )
      { return( binOp_( init, unOp_( value ) ) ); }
    };

    /* ------------------------------
       関数オブジェクト生成ヘルパ関数
       ------------------------------ */

    /// @brief 処理用単項関数から AddBinder 関数オブジェクトを生成する
    ///
    /// @param op 処理用単項関数
    /// @return AddBinder 関数オブジェクト
    template< class Op >
    AddBinder< Op > OpBind( Op op )
    {
      return( AddBinder< Op >( op ) );
    }

    /// @brief 処理用単項関数と二項関数から OpBinder 関数オブジェクトを生成する
    ///
    /// AddBinder の場合と異なり、戻り値の Res を明示する必要があることに注意
    ///
    /// @param unOp 処理用単項関数
    /// @param binOp 二項関数
    /// @return OpBinder 関数オブジェクト
    template< class Res, class UnOp, class BinOp >
    OpBinder< Res, UnOp, BinOp > OpBind( UnOp unOp, BinOp binOp )
    {
      return( OpBinder< Res, UnOp, BinOp >( unOp, binOp ) );
    }

    /* ---------------
       総和計算用関数
       --------------- */

    /// @brief データ列の s から e の直前までの範囲の総和を求める
    ///
    /// テンプレート引数の In はデータ列の反復子の型を表す。
    /// 処理内容としては、初期値をゼロとして std::accumulate を呼び出すだけである。
    ///
    /// @code
    /// return std::accumulate( s, e, Res() );
    /// @endcode
    ///
    /// 初期値を渡さないため、Sum< Res >( ... ) のように戻り値の型を明示する必要があることに注意。
    ///
    /// - s, e からデータ列の長さを調べ、長さが負数となっている場合は例外
    ///   tadah_fussy::exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
    ///
    /// @param s 対象のデータ列の開始反復子
    /// @param e 対象のデータ列の末尾の次の反復子
    /// @return 求めた総和
    template< class Res, class In >
    Res Sum( In s, In e )
    {
      const std::string FUNC_NAME = "math_lib::Sum";

      error_lib::CheckArrayLength( s, e, FUNC_NAME + " : [ s, e ) " + error_message::NEGATIVE_LENGTH );

      return( std::accumulate( s, e, Res() ) );
    }

    /// @brief データ列の s から e の直前までの範囲の総和を求める
    ///
    /// テンプレート引数の In はデータ列の反復子、SUM_Op は総和計算用の二項関数オブジェクトの型をそれぞれ表す。
    /// 処理内容としては、初期値をゼロとして std::accumulate を呼び出すだけである。
    ///
    /// @code
    /// return std::accumulate( s, e, Res(), sumOp );
    /// @endcode
    ///
    /// 初期値を渡さないため、Sum< Res >( ... ) のように戻り値の型を明示する必要があることに注意。
    ///
    /// - s, e からデータ列の長さを調べ、長さが負数となっている場合は例外
    /// tadah_fussy::exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
    ///
    /// @param s 対象のデータ列の開始反復子
    /// @param e 対象のデータ列の末尾の次の反復子
    /// @param sumOp 和の二項計算用関数オブジェクト
    /// @return 求めた総和
    template< class Res, class In, class SUM_Op >
    Res Sum( In s, In e, SUM_Op sumOp )
    {
      const std::string FUNC_NAME = "math_lib::Sum";

      error_lib::CheckArrayLength( s, e, FUNC_NAME + " : [ s, e ) " + error_message::NEGATIVE_LENGTH );

      return( std::accumulate( s, e, Res(), sumOp ) );
    }

    /// @brief 全データの総和を求める
    ///
    /// テンプレート引数の Container はデータ列の型を表す。
    ///
    /// 初期値を渡さないため、SumAll< Res >( ... ) のように戻り値の型を明示する必要があることに注意。
    ///
    /// @param c 対象のデータ列への参照
    /// @return 求めた総和
    template< class Res, class Container >
    Res SumAll( const Container& c )
    {
      return( Sum< Res >( c.begin(), c.end() ) );
    }

    /// @brief 全データの総和を求める
    ///
    /// テンプレート引数の Container はデータ列、SUM_Op は総和計算用の二項関数オブジェクトの型をそれぞれ表す。
    ///
    /// 初期値を渡さないため、SumAll< Res >( ... ) のように戻り値の型を明示する必要があることに注意。
    ///
    /// @param c 対象のデータ列への参照
    /// @param sumOp 和の二項計算用関数オブジェクト
    /// @return 求めた総和
    template< class Res, class Container, class SUM_Op >
    Res SumAll( const Container& c, SUM_Op sumOp )
    {
      return( Sum< Res >( c.begin(), c.end(), sumOp ) );
    }

    /* --------------
       正規化関数
       -------------- */

    /// @brief データ列の s から e の直前までの範囲を正規化する
    ///
    /// テンプレート引数の In はデータ列の反復子の型を表す。
    ///
    /// データ列のユークリッド・ノルムを 1 にする正規化であり、平均ゼロ・標準偏差 1 とする正規化とは異なる。
    /// 後者の正規化は Statistics::Normalize を参照。
    ///
    /// - s, e からデータ列の長さを調べ、長さが負数となっている場合は例外
    /// tadah_fussy::exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
    ///
    /// @param s 対象のデータ列の開始反復子
    /// @param e 対象のデータ列の末尾の次の反復子
    /// @return ユークリッド・ノルムがゼロの場合は false を返す (正規化は行われない)
    template< class In >
    bool Normalize( In s, In e )
    {
      using value_type = typename std::iterator_traits< In >::value_type;

      const std::string FUNC_NAME = "math_lib::Normalize";

      error_lib::CheckArrayLength( s, e, FUNC_NAME + " : [ s, e ) " + error_message::NEGATIVE_LENGTH );
      value_type norm = std::sqrt( std::inner_product( s, e, s, value_type() ) );
      if ( norm == value_type() )
        return( false );
      for ( ; s != e ; ++s )
        *s /= norm;

      return( true );
    }

  } // namespace math_lib

} // namespace tadah_fussy

#endif
