/**
   @file matrix.h
   @brief 行列定義用ヘッダ

   @author tadah_fussy
   @date 2015/10/17 コメント追加
   @date 2019/03/29 コメントの見直し
   @date 2019/04/08
                    - コピー・コンストラクタ、代入演算子の見直し
                    - assign( In s, In e, size_type sz, bool inRows ) において、
                      行列数をゼロとしていた場合は空の行列を生成する(例外は投げない)ように変更
   @date 2019/05/09 const_indexer_iterator の end 関数に const 句を付加 ( 付け忘れ )
   @date 2021/03/08 Indexer/Iteratorにデフォルト・コンストラクタを追加
   @date 2021/03/13 名前空間の変更
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef MATRIX_20210402_H
#define MATRIX_20210402_H

#include <cassert>

#include <iterator>
#include <valarray>
#include <numeric>
#include <utility>
#include <stdexcept>

#include <common/error_library.h>

/**
   @brief 作成者 tadah_fussy
**/
namespace tadah_fussy
{
  /**
     @brief 添字演算子用名前空間
  **/
  namespace indexer
  {
    /* ------------------------------------------------------------------
       Slice : slice を利用した添字演算子
       ------------------------------------------------------------------ */

    /**
       @brief slice を利用した添字演算子

       コンテナやランダム・アクセス反復子を二次元配列として扱い、行・列方向の添字を表現する。例えば

       @code
       indexer::Slice p( 10, 5, 2 )
       @endcode

       なら p( 0 ) = 10, p( 1 ) = 12 ... p( 4 ) = 18 となる。
    **/
    class Slice
    {
      std::slice s_; // slice (先頭・要素数・飛び幅を定義した抽象クラス)

    public:

      /// @brief デフォルト・コンストラクタ
      ///
      /// 先頭位置・要素数・要素間の幅はすべてゼロで初期化する
      Slice()
	: s_( 0, 0, 0 ) {}

      /// @brief slice の初期化引数 ( 先頭・要素数・飛び幅 ) を指定して構築
      ///
      /// 先頭位置とするコンテナ上の添字を start、要素の数を size、アクセスする要素間の幅を stride として構築する。
      /// 例えば、
      /// @code
      /// indexer::Slice p( 10, 5, 2 )
      /// @endcode
      /// なら 10 番目の要素を先頭に、一つおきに 5 つの要素にアクセスできるよう添字を変換するので、
      /// p( 0 ) = 10, p( 1 ) = 12 ... p( 4 ) = 18 となる。
      ///
      /// @param start 先頭のインデックス
      /// @param size 要素数
      /// @param stride 飛び幅
      Slice( size_t start, size_t size, size_t stride )
	: s_( start, size, stride ) {}

      /// @brief 指定した添字に対し、実際にアクセスする添字を返す
      ///
      /// - 範囲外を参照しようとした場合は assert を返す
      ///
      /// @param i 添字
      /// @return コンテナ上で実際にアクセスする対象の添字
      size_t operator()( size_t i ) const;

      /// @brief 要素数 ( アクセス可能な添字の最大値 + 1 ) を返す
      ///
      /// @note 変換後の添字の最大値ではなく、指定した要素数 ( size ) を返すことに注意。
      ///
      /// @return 要素数
      size_t size() const
      { return( s_.size() ); }

      /// @brief 等号演算子
      ///
      /// 他の indexer::Slice と値を比較して等しければ true を返す。比較対象は先頭位置と飛び幅で、要素数は対象外とする。
      ///
      /// @return 等しい場合は true を返す。
      bool operator==( const Slice& si ) const
      { return( s_.start() == si.s_.start() && s_.stride() == si.s_.stride() ); }

      /// @brief 不等号演算子
      ///
      /// 他の indexer::Slice と値を比較して等しくなければ true を返す。比較対象は先頭位置と飛び幅で、要素数は対象外とする。
      ///
      /// @return 等しくない場合は true を返す。
      bool operator!=( const Slice& si ) const
      { return( s_.start() != si.s_.start() || s_.stride() != si.s_.stride() ); }
    };

    /* ------------------------------------------------------------------
       SymMat : 対称行列に対する行・列へのアクセスを想定した添字演算子
       ------------------------------------------------------------------ */

    /**
       @brief 対称行列に対する行・列へのアクセスを想定した添字演算子

       コンテナやランダム・アクセス反復子を下三角行列として扱い(上三角成分は対角成分に対して対称とする)、行・列方向の添字を表現する。例えば、

       <pre>
       [0]  1   2   3
       [1] [4]  5   6
       [2] [5] [7]  8
       [3] [6] [8] [9]
       </pre>

       の場合、実際の値は [] 付きの要素のみでコンテナは

       { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 }

       となり、2 行目の添字は { 1, 4, 5, 6 } となる。
    **/
    class SymMat
    {
      size_t start_;  // 先頭位置
      size_t cols_;   // 行方向に進む列数
      size_t size_;   // アクセスする要素数
      size_t stride_; // 行方向に進む要素間の幅の初期値
      size_t step_;   // 添字の飛び幅(対称行列上での要素の飛び幅)

    public:

      /// @brief デフォルト・コンストラクタ
      ///
      /// 全パラメータはゼロで初期化される
      SymMat()
	: start_( 0 ), cols_( 0 ), size_( 0 ), stride_( 0 ), step_( 0 ) {}

      /// @brief 先頭・行列方向に進む要素数・行方向の飛び幅を指定して構築
      ///
      /// 先頭位置とするコンテナ上の添字を start、行方向に進む列数を cols、アクセスする要素数を size、
      /// 行方向に進む要素間の幅の初期値を stride、添字の飛び幅を step として構築する。
      /// 例えば、
      /// @code
      /// indexer::SymMat p( 2, 3, 4, 3, 1 )
      /// @endcode
      /// なら 2 番目の要素を先頭に、最初に 3 つ、次に 2 つ先へアクセスした後、残りは 1 つずつ進むような添字になるので、
      /// p( 0 ) = 2, p( 1 ) = 5, p( 2 ) = 7, p( 3 ) = 8 となる。このとき、下記のような対称行列に対し、()で示した経路を進む。
      /// 但し、実体のある要素は [] 内のみであることを前提としている。
      ///
      /// <pre>
      /// [0]  1   2   3
      /// [1] [4]  5   6
      /// (2)-(5)-(7)  8
      /// [3] [6] (8) [9]
      /// </pre>
      ///
      /// step = 2 の場合、添字は 2 つずつ進むことになる。先と同じ経路を一つおきに進める場合、
      /// @code
      /// indexer::SymMat p( 2, 3, 2, 3, 2 )
      /// @endcode
      /// とすれば p( 0 ) = 2, p( 1 ) = 7 となる。アクセスする要素数は二つなので size は半分になるのに対し、
      /// cols の値は変化しないことに注意。
      ///
      /// アクセス開始位置を変える場合は start を変更する。
      /// @code
      /// indexer::SymMat p( 5, 2, 3, 2, 1 )
      /// @endcode
      /// とすれば p( 0 ) = 5, p( 1 ) = 7, p( 2 ) = 8 となる。
      /// cols の値は変化しないことに注意。
      ///
      /// @param start 先頭のインデックス
      /// @param cols 行方向に進む列数
      /// @param size アクセスする要素数
      /// @param stride 行方向に進む要素間の幅の初期値
      /// @param step 飛び幅(対称行列上での要素の飛び幅)
      SymMat( size_t start, size_t cols, size_t size, size_t stride, size_t step )
	: start_( start ), cols_( cols ), size_( size ), stride_( stride ), step_( step )  {}

      /// @brief 指定した添字に対し、実際にアクセスする添字を返す
      ///
      /// - 範囲外を参照しようとした場合は assert を返す
      ///
      /// @param i 添字
      /// @return コンテナ上で実際にアクセスする対象の添字
      size_t operator()( size_t i ) const;

      /// @brief 要素数 ( アクセス可能な添字の最大値 + 1 ) を返す
      ///
      /// @note 変換後の添字の最大値ではなく、指定した要素数 ( size ) を返すことに注意。
      ///
      /// @return 要素数
      size_t size() const
      { return( size_ ); }

      /// @brief 等号演算子
      ///
      /// 他の indexer::SymMat と比較して値が等しければ true を返す。比較対象は要素数以外の全てとする。
      ///
      /// @return 等しい場合は true を返す。
      bool operator==( const SymMat& si ) const
      { return( start_ == si.start_ && cols_ == si.cols_ && stride_ == si.stride_ && step_ == si.step_ ); }

      /// @brief 不等号演算子
      ///
      /// 他の indexer::SymMat と比較して値が等しくなければ true を返す。比較対象は要素数以外の全てとする。
      ///
      /// @return 等しくない場合は true を返す。
      bool operator!=( const SymMat& si ) const
      { return( start_ != si.start_ || cols_ != si.cols_ || stride_ != si.stride_ || step_ != si.step_ ); }
    };

  } // namespace indexer

  namespace indexer_iterator
  {
    /* ------------------------------------------------------------------
       Base : indexer を利用した反復子の基底クラス
       ------------------------------------------------------------------ */

    /**
       @brief indexer を利用した反復子

       コンテナやランダム・アクセス反復子の添字演算子を利用して反復処理を行うための反復子。
       コンテナはランダム・アクセス反復子を持つことを前提としている( コンテナの添字反復子 ( operator[] ) を利用する )。

       テンプレート引数の Value はコンテナの値、Container は対象のコンテナ、
       Pointer はポインタの型、Reference は参照の型、Indexer は添字計算用関数オブジェクトの型をそれぞれ表す。

       indexer として indexer::Slice または indexer::SymMat の利用を想定している。

       @note indexer_iterator::Base 自身はそのままでは利用できず、派生クラスを定義して利用する。
    **/
    template
    < class Value,
      class Container,
      class Pointer,
      class Reference,
      class Indexer >
    class Base
      : public std::iterator
    < std::random_access_iterator_tag,
      Value,
      std::ptrdiff_t,
      Pointer,
      Reference >
    {
    public:

      /// 要素の値の型
      using value_type = Value;
      /// 反復子どうしの距離の型
      using difference_type = std::ptrdiff_t;
      /// サイズの型
      using size_type = size_t;
      /// ポインタの型
      using pointer = Pointer;
      /// 参照の型
      using reference = Reference;

    protected:

      Container* container_; ///< 対象のコンテナ
      Indexer indexer_;      ///< 添字計算用関数オブジェクト
      difference_type ci_;   ///< 現在位置

      /// @brief デストラクタ
      ///
      /// デストラクタのアクセス指定子は protected であり、派生クラスしか呼び出しができない。 
      /// そのため、indexer_iterator::Base 自身を構築することはできず、利用できるのはその派生クラスのみとなる。
      ~Base() {}

    private:

      // i 番目の要素への参照を返す(範囲チェックなし)
      reference refUnsafe( difference_type i ) const
      {
	assert( container_ != nullptr );

	return( ( *container_ )[indexer_( ci_ + i )] );
      }

      // i 番目の要素への参照を返す
      // 範囲外を参照しようとした場合は例外を返す
      reference ref( difference_type i ) const
      {
	const std::string FUNC_NAME = "indexer_iterator::Base::ref";

	assert( container_ != nullptr );

	// 範囲外を参照しようとした場合は例外を返す
	if ( ci_ + i < 0 || ci_ + i >= indexer_.size() )
	  throw( std::out_of_range( FUNC_NAME + " : Out of range." ) );

	return( ( *container_ )[indexer_( ci_ + i )] );
      }

      // i 番目の要素へのポインタを返す(範囲チェックなし)
      pointer pUnsafe( difference_type i ) const
      {
	assert( container_ != nullptr );

	return( &( ( *container_ )[indexer_( ci_ + i )] ) );
      }

      // i 番目の要素へのポインタを返す
      // 範囲外を参照しようとした場合は例外を返す
      pointer p( difference_type i ) const
      {
	const std::string FUNC_NAME = "indexer_iterator::Base::p";

	assert( container_ != nullptr );

	if ( ci_ + i < 0 || ci_ + i >= indexer_.size() )
	  throw( std::out_of_range( FUNC_NAME + " : Out of range." ) );

	return( &( ( *container_ )[indexer_( ci_ + i )] ) );
      }

    public:

      /// @brief デフォルト・コンストラクタ
      ///
      /// コンテナへの参照はないため、このまま利用することはできない。
      Base()
	: container_( nullptr ), indexer_(), ci_( 0 ) {}

      /// @brief 対象のコンテナ Container と添字計算用関数オブジェクト Indexer を指定して構築
      ///
      /// @param container 対象のコンテナへの参照(Container)
      /// @param indexer 添字計算用関数オブジェクト(Indexer)
      Base( Container& container, Indexer indexer )
	: container_( &container ), indexer_( indexer ), ci_( 0 ) {}

      /// @brief 指定した添字の値への参照を返す(添字演算子)
      ///
      /// @note 範囲チェックは行わないことに注意。
      ///       チェックする(範囲外なら std::out_of_range 例外を投げる)場合は at( difference_type i ) を使う。
      ///
      /// - コンテナが未定義の場合は assert を呼び出す。
      ///
      /// @param i 要素を取得する対象の添字
      /// @return 対象の要素への参照
      reference operator[]( difference_type i )
      { return( refUnsafe( i ) ); }

      /// @brief 指定した添字の値への参照を返す(範囲チェック付き)
      ///
      /// - コンテナが未定義の場合は assert を呼び出す。
      /// - 範囲外の添字を指定した場合は out_of_range 例外を投げる。
      ///
      /// @param i 要素を取得する対象の添字
      /// @return 対象の要素への参照
      reference at( difference_type i )
      { return( ref( i ) ); }

      /// @brief 現在位置の値への参照を返す(参照演算子)
      ///
      /// @note 範囲チェックは行わないことに注意。
      ///       チェックする(範囲外なら std::out_of_range 例外を投げる)場合は at( void ) を使う。
      ///
      /// - コンテナが未定義の場合は assert を呼び出す。
      ///
      /// @return 現在位置の要素への参照
      reference operator*() const
      { return( refUnsafe( 0 ) ); }

      /// @brief 現在位置の値への参照を返す(範囲チェック付き)
      ///
      /// - コンテナが未定義の場合は assert を呼び出す。
      /// - 範囲外を指し示していた場合は out_of_range 例外を投げる。
      ///
      /// @return 現在位置の要素への参照
      reference at() const
      { return( ref( 0 ) ); }

      /// @brief ポインタ演算子
      ///
      /// @note 範囲チェックは行わないことに注意。
      ///       チェックする(範囲外なら std::out_of_range 例外を投げる)場合は pointTo() を使う。
      ///
      /// - コンテナが未定義の場合は assert を呼び出す。
      ///
      /// @return 現在位置へのポインタ
      pointer operator->() const
      { return( pUnsafe( 0 ) ); }

      /// @brief ポインタ演算子(範囲チェック付き)
      ///
      /// - コンテナが未定義の場合は assert を呼び出す。
      /// - 範囲外を指し示していた場合は out_of_range 例外を投げる。
      ///
      /// @return 現在位置へのポインタ
      pointer pointTo() const
      { return( p( 0 ) ); }

      /// @brief 等号演算子
      ///
      /// 現在位置だけではなく Indexer も比較対象とする。
      ///
      /// - 対象のコンテナが異なる場合は assert を呼び出す。
      ///
      /// @param p 比較対象の Base
      /// @return 等しければ true を返す
      bool operator==( const Base& p ) const
      {
	assert( container_ == p.container_ );

	return( ci_ == p.ci_ && indexer_ == p.indexer_ );
      }

      /// @brief 不等号演算子
      ///
      /// 現在位置だけではなく Indexer も比較対象とする。
      ///
      /// - 対象のコンテナが異なる場合は assert を呼び出す。
      ///
      /// @param p 比較対象の Base
      /// @return 等しくなければ true を返す
      bool operator!=( const Base& p ) const
      { return( ! operator==( p ) ); }

      /// @brief 比較演算子(より小さい)
      ///
      /// Indexer どうしが等しくないと判断された場合は常に false となる。
      ///
      /// - 対象のコンテナが異なる場合は assert を呼び出す。
      ///
      /// @param p 比較対象の Base
      /// @return 自分のほうが小さければ true を返す
      bool operator<( const Base& p ) const
      {
	assert( container_ == p.container_ );

	if ( indexer_ != p.indexer_ ) return( false );
	return( ( ci_ - p.ci_ ) < 0 );
      }

      /// @brief 比較演算子(より大きい)
      ///
      /// Indexer どうしが等しくないと判断された場合は常に false となる。
      ///
      /// - 対象のコンテナが異なる場合は assert を呼び出す。
      ///
      /// @param p 比較対象の Base
      /// @return 自分のほうが大きければ true を返す
      bool operator>( const Base& p ) const
      {
	assert( container_ == p.container_ );

	if ( indexer_ != p.indexer_ ) return( false );
	return( ( ci_ - p.ci_ ) > 0 );
      }

      /// @brief 比較演算子(以下)
      ///
      /// Indexer どうしが等しくないと判断された場合は常に false となる。
      ///
      /// - 対象のコンテナが異なる場合は assert を呼び出す。
      ///
      /// @param p 比較対象の Base
      /// @return 自分のほうが小さいか等しければ true を返す
      bool operator<=( const Base& p ) const
      {
	assert( container_ == p.container_ );

	if ( indexer_ != p.indexer_ ) return( false );
	return( ( ci_ - p.ci_ ) <= 0 );
      }

      /// @brief 比較演算子(以上)
      ///
      /// Indexer どうしが等しくないと判断された場合は常に false となる。
      ///
      /// - 対象のコンテナが異なる場合は assert を呼び出す。
      ///
      /// @param p 比較対象の Base
      /// @return 自分のほうが大きいか等しければ true を返す
      bool operator>=( const Base& p ) const
      {
	assert( container_ == p.container_ );

	if ( indexer_ != p.indexer_ ) return( false );
	return( ( ci_ - p.ci_ ) >= 0 );
      }

      /// @brief 二つの反復子間の距離を返す
      ///
      /// 自分の方が先頭側にある場合は負数になる。
      ///
      /// @note 両反復子の対象コンテナが一致しているかはチェックしていない。
      ///       単純に両反復子の持つ添字の差分を計算しているだけなので、
      ///       コンテナが異なる場合、計算結果は実際の距離と異なることに注意。
      ///
      /// @param p 相手側の indexer_iterator::Base
      /// @return 二つの反復子間の距離
      difference_type operator-( const Base& p ) const
      { return( ci_ - p.ci_ ); }
    };

    /// @cond : Doxygen で二重に表示されるのを防ぐためのコメント

    // indexer_iterator::Constant の前方宣言
    template< class Container, class Indexer, class Value >
    struct Constant;

    /// @endcond

    /* ------------------------------------------------------------------
       indexer_iterator::Normal : 添字演算用関数オブジェクト(indexer)を利用した反復子
       ------------------------------------------------------------------ */

    /**
       @brief 添字演算用関数オブジェクト(indexer)を利用した反復子 (indexer_iterator::Base_からの派生)

       コンテナやランダム・アクセス反復子の添字演算子を利用して反復処理を行うための反復子。
       コンテナはランダム・アクセス反復子を持つことを前提としている( コンテナの添字反復子 ( operator[] ) を利用する )。

       実体は indexer_iterator::Base のテンプレート引数 Pointer と Reference を、
       コンテナの要素に対するポインタ・参照としたものである。

       Indexer として indexer::Slice または indexer::SymMat の利用を想定している。
    **/
    template< class Container, class Indexer, class Value = typename Container::value_type >
    class Normal
      : public Base
    < Value,
      Container,
      Value*,
      Value&,
      Indexer >
    {
      using base_iterator = Base
	< Value,
	  Container,
	  Value*,
	  Value&,
	  Indexer >;

    public:

      /// @brief デフォルト・コンストラクタ
      ///
      /// コンテナへの参照はないため、このまま利用することはできない。
      Normal()
      : base_iterator() {}

      /// @brief 対象のコンテナ Container と添字計算用関数オブジェクト Indexer を指定して構築
      ///
      /// @param container 対象のコンテナへの参照(Container)
      /// @param indexer 添字計算用関数オブジェクト(Indexer)
      Normal( Container& container, Indexer indexer )
	: base_iterator( container, indexer ) {}

      /// @brief indexer_iterator::Base からのコピー・コンストラクタ
      ///
      /// @param it 対象の base_indexer_iterator
      explicit Normal( const Base< Value, Container, Value*, Value&, Indexer >& it )
	: base_iterator( it ) {}

      /// @brief 末尾の次の反復子を返す
      ///
      /// @return 末尾の次の反復子
      Normal end() const
      {
	Normal it( *this );
	it.ci_ = base_iterator::indexer_.size();

	return( it );
      }

      /// @brief インクリメント演算子(前置演算子)
      ///
      /// @note 添字が末尾の次にまで達しているかはチェックしていない。
      ///       インクリメントによりいくらでも添字は大きくすることができる。
      ///
      /// @return インクリメント後の反復子
      Normal& operator++()
      {
	++base_iterator::ci_;

	return( *this );
      }

      /// @brief インクリメント演算子(後置演算子)
      ///
      /// @note 添字が末尾の次にまで達しているかはチェックしていない。
      ///       インクリメントによりいくらでも添字は大きくすることができる。
      ///
      /// @return インクリメント前の反復子
      Normal operator++( int )
      {
	Normal it( *this );
	++base_iterator::ci_;

	return( it );
      }

      /// @brief デクリメント演算子(前置演算子)
      ///
      /// @note 添字が先頭にまで達しているかはチェックしていない。
      ///       デクリメントによりいくらでも添字は小さくすることができる。
      ///
      /// @return デクリメント後の反復子
      Normal& operator--()
      {
	--base_iterator::ci_;

	return( *this );
      }

      /// @brief デクリメント演算子(後置演算子)
      ///
      /// @note 添字が先頭にまで達しているかはチェックしていない。
      ///       デクリメントによりいくらでも添字は小さくすることができる。
      ///
      /// @return デクリメント前の反復子
      Normal operator--( int )
      {
	Normal it( *this );
	--base_iterator::ci_;

	return( it );
      }

      /// @brief 加算・代入演算子
      ///
      /// @note 加算後の添字がオーバーフローしているかはチェックしていない。
      ///
      /// @param i 加算数
      /// @return 加算後の反復子
      Normal& operator+=( typename base_iterator::difference_type i )
      {
	base_iterator::ci_ += i;

	return( *this );
      }

      /// @brief 減算・代入演算子
      ///
      /// @note 減算後の添字がオーバーフローしているかはチェックしていない。
      ///
      /// @param i 減算数
      /// @return 減算後の反復子
      Normal& operator-=( typename base_iterator::difference_type i )
      {
	base_iterator::ci_ -= i;

	return( *this );
      }

      friend Constant< Container, Indexer, Value >::Constant( const Normal< Container, Indexer, Value >& it );
    };

    /* ------------------------------------------------------------------
       indexer_iterator::Normal 用ヘルパ関数
       ------------------------------------------------------------------ */

    /// @brief 加算演算子
    ///
    /// @note 加算後の添字がオーバーフローしているかはチェックしていない。
    ///
    /// @param it 加算対象の反復子
    /// @param i 加算数
    /// @return 加算後の反復子
    template< class Container, class Indexer, class Value >
    Normal< Container, Indexer, Value >
    operator+
    ( const Normal< Container, Indexer, Value >& it,
      typename Normal< Container, Indexer, Value >::difference_type i )
    {
      Normal< Container, Indexer, Value > res = it;
      res += i;

      return( res );
    }

    /// @brief 減算演算子
    ///
    /// @note 減算後の添字がオーバーフローしているかはチェックしていない。
    ///
    /// @param it 加算対象の反復子
    /// @param i 減算数
    /// @return 減算後の反復子
    template< class Container, class Indexer, class Value >
    Normal< Container, Indexer,Value >
    operator-
    ( const Normal< Container, Indexer, Value >& it,
      typename Normal< Container, Indexer, Value >::difference_type i )
    {
      Normal< Container, Indexer, Value > res = it;
      res -= i;

      return( res );
    }

    /* ------------------------------------------------------------------
       indexer_iterator::Constant :  indexer を利用した定数反復子
       ------------------------------------------------------------------ */

    /**
       @brief 添字演算用関数オブジェクト(indexer)を利用した定数反復子 (Baseからの派生)

       コンテナやランダム・アクセス反復子の添字演算子を利用して反復処理を行うための反復子。
       コンテナはランダム・アクセス反復子を持つことを前提としている( コンテナの添字反復子 ( operator[] ) を利用する )。

       実体は indexer_iterator::Base のテンプレート引数 Pointer と Reference を、
       コンテナの要素に対する const ポインタ・const 参照としたものである。

       Indexer として indexer::Slice または indexer::SymMat の利用を想定している。
    **/
    template< class Container, class Indexer, class Value = typename Container::value_type >
    class Constant
      : public Base
    < Value,
      const Container,
      const Value*,
      const Value&,
      Indexer >
    {
      using base_iterator = Base
	< Value,
	  const Container,
	  const Value*,
	  const Value&,
	  Indexer >;

    public:

      /// @brief デフォルト・コンストラクタ
      ///
      /// コンテナへの参照はないため、このまま利用することはできない。
      Constant()
      : base_iterator() {}

      /// @brief 対象のコンテナ Container と添字計算用関数オブジェクト Indexer を指定して構築
      ///
      /// @note 対象のコンテナに固定の配列を利用する場合、Container として "Value *const" を指定する必要がある。
      ///       例えば indexer_iterator< int *const, SliceIndexer, int > など。
      ///
      /// @param container 対象のコンテナへの参照(Container)
      /// @param indexer 添字計算用関数オブジェクト(Indexer)
      Constant( const Container& container, Indexer indexer )
	: base_iterator( container, indexer ) {}

      /// @brief indexer_iterator::Base からのコピー・コンストラクタ
      ///
      /// @param it 対象の indexer_iterator::Base
      explicit Constant( const Base< Value, const Container, const Value*, const Value&, Indexer >& it )
	: base_iterator( it ) {}

      /// @brief indexer_iterator からのコピー・コンストラクタ
      ///
      /// 通常の indexer_iterator を定数反復子に変換するために必要となる。
      ///
      /// @param it 対象の indexer_iterator
      Constant( const Normal< Container, Indexer, Value >& it )
	: base_iterator( *( it.container_ ), it.indexer_ )
      { this->ci_ = it.ci_; }

      /// @brief 末尾の次の反復子を返す
      ///
      /// @return 末尾の次の反復子
      Constant end() const
      {
	Constant it( *this );
	it.ci_ = base_iterator::indexer_.size();

	return( it );
      }

      /// @brief インクリメント演算子(前置演算子)
      ///
      /// @note 添字が末尾の次にまで達しているかはチェックしていない。
      ///       インクリメントによりいくらでも添字は大きくすることができる。
      ///
      /// @return インクリメント後の反復子
      Constant& operator++()
      {
	++base_iterator::ci_;

	return( *this );
      }

      /// @brief インクリメント演算子(後置演算子)
      ///
      /// @note 添字が末尾の次にまで達しているかはチェックしていない。
      ///       インクリメントによりいくらでも添字は大きくすることができる。
      ///
      /// @return インクリメント前の反復子
      Constant operator++( int )
      {
	Constant it( *this );
	++base_iterator::ci_;

	return( it );
      }

      /// @brief デクリメント演算子(前置演算子)
      ///
      /// @note 添字が先頭にまで達しているかはチェックしていない。
      ///       デクリメントによりいくらでも添字は小さくすることができる。
      ///
      /// @return デクリメント後の反復子
      Constant& operator--()
      {
	--base_iterator::ci_;

	return( *this );
      }

      /// @brief デクリメント演算子(後置演算子)
      ///
      /// @note 添字が先頭にまで達しているかはチェックしていない。
      ///       デクリメントによりいくらでも添字は小さくすることができる。
      ///
      /// @return デクリメント前の反復子
      Constant operator--( int )
      {
	Constant it( *this );
	--base_iterator::ci_;

	return( it );
      }

      /// @brief 加算・代入演算子
      ///
      /// @note 加算後の添字がオーバーフローしているかはチェックしていない。
      ///
      /// @param i 加算数
      /// @return 加算後の反復子
      Constant& operator+=( typename base_iterator::difference_type i )
      {
	base_iterator::ci_ += i;

	return( *this );
      }

      /// @brief 減算・代入演算子
      ///
      /// @note 減算後の添字がオーバーフローしているかはチェックしていない。
      ///
      /// @param i 減算数
      /// @return 減算後の反復子
      Constant& operator-=( typename base_iterator::difference_type i )
      {
	base_iterator::ci_ -= i;

	return( *this );
      }
    };

    /* ------------------------------------------------------------------
       indexer_iterator::Constant 用ヘルパ関数
       ------------------------------------------------------------------ */

    /// @brief 加算演算子
    ///
    /// @note 加算後の添字がオーバーフローしているかはチェックしていない。
    ///
    /// @param it 加算対象の反復子
    /// @param i 加算数
    /// @return 加算後の反復子
    template< class Container, class Indexer, class Value >
    Constant< Container, Indexer, Value >
    operator+
    ( const Constant< Container, Indexer, Value >& it,
      typename Constant< Container, Indexer, Value >::difference_type i )
    {
      Constant< Container, Indexer, Value > res = it;
      res += i;

      return( res );
    }

    /// @brief 減算演算子
    ///
    /// @note 減算後の添字がオーバーフローしているかはチェックしていない。
    ///
    /// @param it 加算対象の反復子
    /// @param i 減算数
    /// @return 減算後の反復子
    template< class Container, class Indexer, class Value >
    Constant< Container, Indexer, Value >
    operator-
    ( const Constant< Container, Indexer, Value >& it,
      typename Constant< Container, Indexer, Value >::difference_type i )
    {
      Constant< Container, Indexer, Value > res = it;
      res -= i;

      return( res );
    }

    /* ------------------------------------------------------------------
       indexer_iterator::Normal と indexer_iterator::Constant 間での比較関数
       ------------------------------------------------------------------ */

    /// @brief 等号演算子
    ///
    /// 通常・定数反復子での比較を行うためのヘルパ関数
    ///
    /// @param c 比較対象の indexer_iterator::Constant
    /// @param p 比較対象の indexer_iterator::Normal
    /// @return 等しければ true を返す
    template< class Container, class Indexer, class Value >
    bool operator==( const Constant< Container, Indexer, Value >& c, const Normal< Container, Indexer, Value >& p )
    { return( c == Constant< Container, Indexer, Value >( p ) ); }

    /// @brief 等号演算子
    ///
    /// 通常・定数反復子での比較を行うためのヘルパ関数
    ///
    /// @param p 比較対象の indexer_iterator::Normal
    /// @param c 比較対象の indexer_iterator::Constant
    /// @return 等しければ true を返す
    template< class Container, class Indexer, class Value >
    bool operator==( const Normal< Container, Indexer, Value >& p, const Constant< Container, Indexer, Value >& c )
    { return( c == Constant< Container, Indexer, Value >( p ) ); }

    /// @brief 不等号演算子
    ///
    /// 通常・定数反復子での比較を行うためのヘルパ関数
    ///
    /// @param c 比較対象の indexer_iterator::Constant
    /// @param p 比較対象の indexer_iterator::Normal
    /// @return 等しくなければ true を返す
    template< class Container, class Indexer, class Value >
    bool operator!=( const Constant< Container, Indexer, Value >& c, const Normal< Container, Indexer, Value >& p )
    { return( c != Constant< Container, Indexer, Value >( p ) ); }

    /// @brief 不等号演算子
    ///
    /// 通常・定数反復子での比較を行うためのヘルパ関数
    ///
    /// @param p 比較対象の indexer_iterator::Normal
    /// @param c 比較対象の indexer_iterator::Constant
    /// @return 等しくなければ true を返す
    template< class Container, class Indexer, class Value >
    bool operator!=( const Normal< Container, Indexer, Value >& p, const Constant< Container, Indexer, Value >& c )
    { return( c != Constant< Container, Indexer, Value >( p ) ); }

    /// @brief 比較演算子(より小さい)
    ///
    /// 通常・定数反復子での比較を行うためのヘルパ関数
    ///
    /// @param c 比較対象の indexer_iterator::Constant
    /// @param p 比較対象の indexer_iterator::Normal
    /// @return c の方が小さければ true を返す
    template< class Container, class Indexer, class Value >
    bool operator<( const Constant< Container, Indexer, Value >& c, const Normal< Container, Indexer, Value >& p )
    { return( c < Constant< Container, Indexer, Value >( p ) ); }

    /// @brief 比較演算子(より小さい)
    ///
    /// 通常・定数反復子での比較を行うためのヘルパ関数
    ///
    /// @param p 比較対象の indexer_iterator::Normal
    /// @param c 比較対象の indexer_iterator::Constant
    /// @return p の方が小さければ true を返す
    template< class Container, class Indexer, class Value >
    bool operator<( const Normal< Container, Indexer, Value >& p, const Constant< Container, Indexer, Value >& c )
    { return( Constant< Container, Indexer, Value >( p ) < c ); }

    /// @brief 比較演算子(より大きい)
    ///
    /// 通常・定数反復子での比較を行うためのヘルパ関数
    ///
    /// @param c 比較対象の indexer_iterator::Constant
    /// @param p 比較対象の indexer_iterator::Normal
    /// @return c の方が大きければ true を返す
    template< class Container, class Indexer, class Value >
    bool operator>( const Constant< Container, Indexer, Value >& c, const Normal< Container, Indexer, Value >& p )
    { return( c > Constant< Container, Indexer, Value >( p ) ); }

    /// @brief 比較演算子(より大きい)
    ///
    /// 通常・定数反復子での比較を行うためのヘルパ関数
    ///
    /// @param p 比較対象の indexer_iterator::Normal
    /// @param c 比較対象の indexer_iterator::Constant
    /// @return p の方が大きければ true を返す
    template< class Container, class Indexer, class Value >
    bool operator>( const Normal< Container, Indexer, Value >& p, const Constant< Container, Indexer, Value >& c )
    { return( Constant< Container, Indexer, Value >( p ) > c ); }

    /// @brief 比較演算子(以下)
    ///
    /// 通常・定数反復子での比較を行うためのヘルパ関数
    ///
    /// @param c 比較対象の indexer_iterator::Constant
    /// @param p 比較対象の indexer_iterator::Normal
    /// @return c が p 以下ならば true を返す
    template< class Container, class Indexer, class Value >
    bool operator<=( const Constant< Container, Indexer, Value >& c, const Normal< Container, Indexer, Value >& p )
    { return( c <= Constant< Container, Indexer, Value >( p ) ); }

    /// @brief 比較演算子(以下)
    ///
    /// 通常・定数反復子での比較を行うためのヘルパ関数
    ///
    /// @param p 比較対象の indexer_iterator::Normal
    /// @param c 比較対象の indexer_iterator::Constant
    /// @return p が c 以下ならば true を返す
    template< class Container, class Indexer, class Value >
    bool operator<=( const Normal< Container, Indexer, Value >& p, const Constant< Container, Indexer, Value >& c )
    { return( Constant< Container, Indexer, Value >( p ) <= c ); }

    /// @brief 比較演算子(以上)
    ///
    /// 通常・定数反復子での比較を行うためのヘルパ関数
    ///
    /// @param c 比較対象の indexer_iterator::Constant
    /// @param p 比較対象の indexer_iterator::Normal
    /// @return c が p 以上ならば true を返す
    template< class Container, class Indexer, class Value >
    bool operator>=( const Constant< Container, Indexer, Value >& c, const Normal< Container, Indexer, Value >& p )
    { return( c >= Constant< Container, Indexer, Value >( p ) ); }

    /// @brief 比較演算子(以上)
    ///
    /// 通常・定数反復子での比較を行うためのヘルパ関数
    ///
    /// @param p 比較対象の indexer_iterator::Normal
    /// @param c 比較対象の indexer_iterator::Constant
    /// @return p が c 以上ならば true を返す
    template< class Container, class Indexer, class Value >
    bool operator>=( const Normal< Container, Indexer, Value >& p, const Constant< Container, Indexer, Value >& c )
    { return( Constant< Container, Indexer, Value >( p ) >= c ); }

  } // namespace indexer_iterator

  /* ------------------------------------------------------------------
     Matrix : 行列クラス
     ------------------------------------------------------------------ */

  /// @cond : Doxygen で二重に表示されるのを防ぐためのコメント

  // SquareMatrix, SymmetricMatrix の前方宣言
  template< class T > class SquareMatrix;
  template< class T > class SymmetricMatrix;

  /// @endcond

  /**
     @brief 行列(二次元配列)クラス

     テンプレート引数の T は要素の型を表す。
     Matrix クラスは要素へのアクセスやサイズ変更・取得などの関数を持つが、
     演算用のメンバ関数はないため四則演算を持たないような型も要素にすることができる。

     Matrix には専用の反復子が indexer_iterator::Normal と indexer_iterator::Constant を利用して用意されている。
     Matrix の行や列に対して反復子を使ってループ処理を行うことができる。
     また、添字を使って直接特定の要素へアクセスすることも可能である。
  **/
  template< class T >
  class Matrix
  {
  public:

    /// @brief 要素の型
    using value_type = T;
    /// @brief 反復子の型
    using iterator = indexer_iterator::Normal< std::valarray< T >, indexer::Slice >;
    /// @brief 定数反復子の型
    using const_iterator = indexer_iterator::Constant< std::valarray< T >, indexer::Slice >;
    /// @brief 添字の型
    using size_type = size_t;

  private:

    std::valarray< T > container_; // 値を格納する配列
    size_type rows_;  // 行数
    size_type cols_;  // 列数

    // 行・列番号 = ( r, c ) に対する添字を返す(範囲チェックなし)
    size_type getIndexUnsafe( size_type r, size_type c ) const
    { return( c * rows() + r ); }
    // 行・列番号 = ( r, c ) に対する添字を返す
    size_type getIndex( size_type r, size_type c ) const;
    // maxSize のサイズを stride ずつアクセスする場合の回数を返す
    static size_type getSize( size_type maxSize, size_type stride );
    // 反復子上の要素を全て入れ替える
    static void swap( iterator i1, iterator i2 );

  public:

    /// @brief デフォルト・コンストラクタ
    ///
    /// 行列数はゼロで初期化する。
    Matrix() : rows_( size_type() ), cols_( size_type() ) {}

    /// @brief 行・列数を指定して構築
    ///
    /// 要素は初期値 val で初期化される。val のデフォルトは value_type のデフォルト値になる。
    ///
    /// @param rows 行数
    /// @param cols 列数
    /// @param val 初期値
    Matrix( size_type rows, size_type cols, value_type val = value_type() )
      : container_( val, rows * cols ), rows_( rows ), cols_( cols ) {}

    /// @brief データ列を指定して構築
    ///
    /// inRows が true なら sz は行のサイズを表し、全てのデータが行列に代入される最小のサイズが列数となる。
    /// inRows が false なら sz は列のサイズを表し、全てのデータが行列に代入される最小のサイズが行数となる。
    /// データが不足した場合、残りの要素は value_type のデフォルト値となる。
    ///
    /// テンプレート引数の In はデータ列 [ s, e ) に対する反復子の型を表す。
    ///
    /// - データ列 [ s, e ) の指定が不正だった場合、ExceptionNegativeLength< std::iterator_traits< In >::difference_type > 例外を投げる。
    ///
    /// @param s データ列の開始位置
    /// @param e データ列の末尾の次の位置
    /// @param sz 行(または列)のサイズ
    /// @param inRows true なら行方向(一列を並べたら次の列へ進む)に、false なら列方向(一行を並べたら次の行へ進む)にデータ列を並べる
    template< class In >
    Matrix( In s, In e, size_type sz, bool inRows )
    { assign( s, e, sz, inRows ); }

    /// @brief 行列の要素をコピーして構築(コピー・コンストラクタ)
    ///
    /// @param mat コピー元の行列
    Matrix( const Matrix& mat )
      : container_( mat.container_ ), rows_( mat.rows() ), cols_( mat.cols() ) {}

    /// @brief 任意の型の行列の要素をコピーして構築
    ///
    /// @note コピー元と要素の型が異なるため型の変換が発生することに注意。
    ///
    /// @param mat コピー元の行列
    template< class U >
    explicit Matrix( const Matrix< U >& mat )
    { assign( mat.begin(), mat.end(), mat.rows(), true ); }

    /// @brief 正方行列の要素をコピーして構築
    ///
    /// @note コピー元と要素の型が異なる場合は型の変換が発生することに注意。
    ///
    /// @param mat コピー元の正方行列
    template< class U >
    explicit Matrix( const SquareMatrix< U >& mat )
    { assign( mat.begin(), mat.end(), mat.rows(), true ); }

    /// @brief 対称行列の要素をコピーして構築
    ///
    /// @note コピー元と要素の型が異なる場合は型の変換が発生することに注意。
    ///
    /// @param mat コピー元の対称行列
    template< class U >
    explicit Matrix( const SymmetricMatrix< U >& mat );

    /// @brief 代入演算子の多重定義(型が等しい行列の代入)
    ///
    /// @param mat 代入元の行列
    Matrix& operator=( const Matrix& mat );

    /// @brief 代入演算子の多重定義(任意の型の行列の代入)
    ///
    /// @note 代入元と要素の型が異なるため型の変換が発生することに注意。
    ///
    /// @param mat 代入元の行列
    template< class U >
    Matrix& operator=( const Matrix< U >& mat );

    /// @brief 代入演算子の多重定義(正方行列の代入)
    ///
    /// @note 代入元と要素の型が異なる場合は型の変換が発生することに注意。
    ///
    /// @param mat 代入元の正方行列
    template< class U >
    Matrix& operator=( const SquareMatrix< U >& mat );

    /// @brief 代入演算子の多重定義(対称行列の代入)
    ///
    /// @note 代入元と要素の型が異なる場合は型の変換が発生することに注意。
    ///
    /// @param mat 代入元の対称行列
    template< class U >
    Matrix& operator=( const SymmetricMatrix< U >& mat );

    /// @brief 要素数 ( 行数 × 列数 ) を変更する
    ///
    /// 行・列それぞれの方向に対し、サイズが小さくなる場合は末尾側の要素は削除され、
    /// 大きくなる場合は増分が初期値 val で埋められる。
    /// 元の要素は(削除されなければ)残る。その必要がない場合は、
    /// assign( size_type rows, size_type cols, value_type val ) を使ったほうが効率がよい。
    ///
    /// @param rows 変更後の行数
    /// @param cols 変更後の列数
    /// @param val 増分に対する初期値
    /// @return なし
    void resize( size_type rows, size_type cols, value_type val = value_type() );

    /// @brief 要素数 ( 行数 × 列数 ) を変更し、値を初期化する
    ///
    /// @param rows 行数
    /// @param cols 列数
    /// @param val 初期化する値
    /// @return なし
    void assign( size_type rows, size_type cols, value_type val = value_type() )
    {
      container_.resize( rows * cols, val );
      rows_ = rows;
      cols_ = cols;
    }

    /// @brief 行列をデータ列で初期化する
    ///
    /// inRows が true なら sz は行のサイズを表し、全てのデータが行列に代入される最小のサイズが列数となる。
    /// inRows が false なら sz は列のサイズを表し、全てのデータが行列に代入される最小のサイズが行数となる。
    /// データが不足した場合、残りの要素は value_type のデフォルト値となる。
    ///
    /// テンプレート引数の In はデータ列 [ s, e ) に対する反復子の型を表す。
    ///
    /// - データ列 [ s, e ) の指定が不正だった場合、ExceptionNegativeLength< std::iterator_traits< In >::difference_type > 例外を投げる。
    ///
    /// @note データ列 [ s, e ) の長さがゼロだった場合、指定したサイズ sz が 0 より大きければ、
    ///       必ず 1 行 sz 列 ( または sz 行 1 列 ) の行列が生成される。
    ///
    /// @param s データ列の開始位置
    /// @param e データ列の末尾の次の位置
    /// @param sz 行(または列)のサイズ
    /// @param inRows true なら行方向(一列を並べたら次の列へ進む)に、false なら列方向(一行を並べたら次の行へ進む)にデータ列を並べる
    template< class In >
    void assign( In s, In e, size_type sz, bool inRows );

    /// @brief 行数を返す
    ///
    /// @return 行数
    size_type rows() const { return( rows_ ); }

    /// @brief 列数を返す
    ///
    /// @return 列数
    size_type cols() const { return( cols_ ); }

    /// @brief 指定した行番号の反復子を返す(添字演算子)
    ///
    /// @note 添字に対するチェックは行わない。
    ///
    /// @param r 行番号
    /// @return 行の反復子
    iterator operator[]( size_type r )
    { return( iterator( container_, indexer::Slice( getIndexUnsafe( r, 0 ), cols(), rows() ) ) ); }

    /// @brief 指定した行番号の定数反復子を返す(添字演算子)
    ///
    /// @note 添字に対するチェックは行わない。
    ///
    /// @param r 行番号
    /// @return 行の定数反復子
    const_iterator operator[]( size_type r ) const
    { return( const_iterator( container_, indexer::Slice( getIndexUnsafe( r, 0 ), cols(), rows() ) ) ); }

    /// @brief 指定した行番号の反復子を返す
    ///
    /// - 範囲外の行番号を指定した場合は out_of_range 例外を投げる。
    /// - 列方向の飛び幅がゼロの場合、ExceptionNotPositiveNumber< Matrix< T >::size_type > 例外を投げる。
    ///
    /// @param r 行番号
    /// @param start 列番号の開始位置
    /// @param stride 列方向の飛び幅
    /// @return 行の反復子
    iterator row( size_type r, size_type start = 0, size_type stride = 1 );

    /// @brief 指定した行番号の定数反復子を返す
    ///
    /// - 範囲外の行番号を指定した場合は out_of_range 例外を投げる。
    /// - 列方向の飛び幅がゼロの場合、ExceptionNotPositiveNumber< Matrix< T >::size_type > 例外を投げる。
    ///
    /// @param r 行番号
    /// @param start 列番号の開始位置
    /// @param stride 列方向の飛び幅
    /// @return 行の定数反復子
    const_iterator row( size_type r, size_type start = 0, size_type stride = 1 ) const;

    /// @brief 指定した列番号の反復子を返す
    ///
    /// @note 添字に対するチェックは行わない。
    ///
    /// @param c 列番号
    /// @return 列の反復子
    iterator cbegin( size_type c )
    { return( iterator( container_, indexer::Slice( getIndexUnsafe( 0, c ), rows(), 1 ) ) ); }

    /// @brief 指定した列番号の定数反復子を返す
    ///
    /// @note 添字に対するチェックは行わない。
    ///
    /// @param c 列番号
    /// @return 列の定数反復子
    const_iterator cbegin( size_type c ) const
    { return( const_iterator( container_, indexer::Slice( getIndexUnsafe( 0, c ), rows(), 1 ) ) ); }

    /// @brief 指定した列番号の反復子を返す
    ///
    /// - 範囲外の列番号を指定した場合は out_of_range 例外を投げる。
    /// - 行方向の飛び幅がゼロの場合、ExceptionNotPositiveNumber< Matrix< T >::size_type > 例外を投げる。
    ///
    /// @param c 列番号
    /// @param start 行番号の開始位置
    /// @param stride 行方向の飛び幅
    /// @return 列の反復子
    iterator column( size_type c, size_type start = 0, size_type stride = 1 );

    /// @brief 指定した列番号の定数反復子を返す
    ///
    /// - 範囲外の列番号を指定した場合は out_of_range 例外を投げる。
    /// - 行方向の飛び幅がゼロの場合、ExceptionNotPositiveNumber< Matrix< T >::size_type > 例外を投げる。
    ///
    /// @param c 列番号
    /// @param start 行番号の開始位置
    /// @param stride 行方向の飛び幅
    /// @return 列の定数反復子
    const_iterator column( size_type c, size_type start = 0, size_type stride = 1 ) const;

    /// @brief 先頭位置から始まる全要素へアクセス可能な反復子を返す
    ///
    /// この反復子は先頭から同列の次の行へとたどり、末尾に達したら次の列の先頭に移る。
    ///
    /// @return 対象の反復子
    iterator begin();

    /// @brief 先頭位置から始まる全要素へアクセス可能な定数反復子を返す
    ///
    /// この反復子は先頭から同列の次の行へとたどり、末尾に達したら次の列の先頭に移る。
    ///
    /// @return 対象の定数反復子
    const_iterator begin() const;

    /// @brief 要素の末尾を表す反復子を返す
    ///
    /// このメンバ関数で返される反復子は、begin() で得た先頭位置から始まる全要素へアクセス可能な反復子の末尾を表す。
    /// 行・列に対する反復子の末尾は、indexer_iterator::end や const_indexer_iterator::end で得ることに注意。
    /// 例えば以下は 1 列目の列ベクトルのノルムを計算するためのプログラムで、末尾の指定は a->end() ではなく col.end() となっている。
    ///
    /// @code
    /// Matrix< double >::const_iterator col = a->column( 0 ); // 行列 a の 1 列目の先頭
    /// double norm = std::inner_product( col, col.end(), col, 0.0 );
    /// @endcode
    ///
    /// @return 末尾の反復子
    iterator end();

    /// @brief 要素の末尾を表す定数反復子を返す
    ///
    /// このメンバ関数で返される反復子は、begin() で得た先頭位置から始まる全要素へアクセス可能な反復子の末尾を表す。
    /// 行・列に対する反復子の末尾は、indexer_iterator::end や const_indexer_iterator::end で得ることに注意。
    /// 例えば以下は 1 列目の列ベクトルのノルムを計算するためのプログラムで、末尾の指定は a->end() ではなく col.end() となっている。
    ///
    /// @code
    /// Matrix< double >::const_iterator col = a->column( 0 ); // 行列 a の 1 列目の先頭
    /// double norm = std::inner_product( col, col.end(), col, 0.0 );
    /// @endcode
    ///
    /// @return 末尾の定数反復子
    const_iterator end() const;

    /// @brief 要素の転置を行う
    ///
    /// @return なし
    void transpose();
  };

  /**
     @brief 正方行列クラス

     テンプレート引数の T は要素の型を表す。
     SquareMatrix クラスは要素へのアクセスやサイズ変更・取得などの関数を持つ。
     演算用のメンバ関数として operator*= を持つが、それを利用しなければ四則演算を持たないような型も要素にすることができる。

     SquareMatrix には専用の反復子が indexer_iterator::Normal と indexer_iterator::Constant を利用して用意されている。
     SquareMatrix の行や列に対して反復子を使ってループ処理を行うことができる。
     また、添字を使って直接特定の要素へアクセスすることも可能である。
  **/
  template< typename T >
  struct SquareMatrix : private Matrix< T >
  {
    /// @brief 要素の型
    using value_type = T;
    /// @brief 反復子の型
    using iterator = typename Matrix< T >::iterator;
    /// @brief 定数反復子の型
    using const_iterator = typename Matrix< T >::const_iterator;
    /// @brief 添字の型
    using size_type = typename Matrix< T >::size_type;

    /// @brief デフォルト・コンストラクタ
    ///
    /// 行列数はゼロで初期化する。
    SquareMatrix() : Matrix< T >() {}

    /// @brief 行・列数を指定して構築
    ///
    /// 要素は初期値 val で初期化される。val のデフォルトは value_type のデフォルト値になる。
    ///
    /// @param size 行・列数
    /// @param val 初期値
    explicit SquareMatrix( size_type size, value_type val = value_type() )
      : Matrix< T >( size, size, val ) {}

    /// @brief データ列を指定して構築
    ///
    /// データが不足した場合、残りの要素は value_type のデフォルト値となる。
    ///
    /// @param s データ列の開始位置
    /// @param e データ列の末尾の次の位置
    /// @param sz 行列のサイズ
    /// @param inRows true なら行方向(一列を並べたら次の列へ進む)に、false なら列方向(一行を並べたら次の行へ進む)にデータ列を並べる
    template< class In >
    SquareMatrix( In s, In e, size_type sz, bool inRows = true )
    { assign( s, e, sz, inRows ); }

    /// @brief 正方行列の要素をコピーして構築(コピー・コンストラクタ)
    ///
    /// @param mat コピー元の正方行列
    SquareMatrix( const SquareMatrix& mat )
      : Matrix< T >( mat ) {}

    /// @brief 型の異なる正方行列の要素をコピーして構築
    ///
    /// @note コピー元と要素の型が異なるため型の変換が発生することに注意。
    ///
    /// @param mat コピー元の正方行列
    template< class U >
    explicit SquareMatrix( const SquareMatrix< U >& mat )
      : Matrix< T >( mat ) {}

    /// @brief 対称行列の要素をコピーして構築
    ///
    /// @note コピー元と要素の型が異なる場合は型の変換が発生することに注意。
    ///
    /// @param mat コピー元の対称行列
    template< class U >
    explicit SquareMatrix( const SymmetricMatrix< U >& mat )
      : Matrix< T >( mat ) {}

    /// @brief 代入演算子の多重定義(型の等しい正方行列を代入する)
    ///
    /// @param mat 代入元の正方行列
    SquareMatrix& operator=( const SquareMatrix& mat )
    {
      if ( this != &mat )
	Matrix< T >::operator=( mat );

      return( *this );
    }

    /// @brief 代入演算子の多重定義(任意の型の正方行列を代入する)
    ///
    /// @note 代入元と要素の型が異なるため型の変換が発生することに注意。
    ///
    /// @param mat 代入元の正方行列
    template< class U >
    SquareMatrix& operator=( const SquareMatrix< U >& mat )
    {
      Matrix< T >::operator=( mat );
      return( *this );
    }

    /// @brief 代入演算子の多重定義(任意の型の対称行列を代入する)
    ///
    /// @note 代入元と要素の型が異なる場合は型の変換が発生することに注意。
    ///
    /// @param mat 代入元の対称行列
    template< class U >
    SquareMatrix& operator=( const SymmetricMatrix< U >& mat )
    {
      Matrix< T >::operator=( mat );
      return( *this );
    }

    /// @brief 行列のサイズを返す
    ///
    /// @return 行列のサイズ
    size_type size() const
    { return( Matrix< T >::rows() ); }

    /// @brief 行数を返す
    ///
    /// Matrix との互換性のためにあるが、SquareMatrix< T >::size() と内容は同じである。
    ///
    /// @return 行数
    size_type rows() const
    { return( Matrix< T >::rows() ); }

    /// @brief 列数を返す
    ///
    /// Matrix との互換性のためにあるが、SquareMatrix< T >::size() と内容は同じである。
    ///
    /// @return 列数
    size_type cols() const
    { return( Matrix< T >::cols() ); }

    /// @brief 要素数 ( 行数 × 列数 ) を変更する
    ///
    /// 行・列それぞれの方向に対し、サイズが小さくなる場合は末尾側の要素は削除され、
    /// 大きくなる場合は増分が初期値 val で埋められる。
    /// 元の要素は(削除されなければ)残る。その必要がない場合は、
    /// assign( size_type size, value_type val ) を使ったほうが効率がよい。
    ///
    /// @param size 行・列数
    /// @param val 増分に対する初期値
    /// @return なし
    void resize( size_type size, value_type val = value_type() )
    { Matrix< T >::resize( size, size, val ); }

    /// @brief 要素数 ( 行数 × 列数 ) を変更し、値を初期化する
    ///
    /// @param size 行・列数
    /// @param val 初期化する値
    /// @return なし
    void assign( size_type size, value_type val = value_type() )
    { Matrix< T >::assign( size, size, val ); }

    /// @brief データ列を指定して行列を初期化する
    ///
    /// データが不足した場合、残りの要素は value_type のデフォルト値となる。
    ///
    /// テンプレート引数の In はデータ列 [ s, e ) に対する反復子の型を表す。
    ///
    /// - データ列 [ s, e ) の指定が不正だった場合、ExceptionNegativeLength< std::iterator_traits< In >::difference_type > 例外を投げる。
    ///
    /// @param s データ列の開始位置
    /// @param e データ列の末尾の次の位置
    /// @param sz 行(または列)のサイズ
    /// @param inRows true なら行方向(一列を並べたら次の列へ進む)に、false なら列方向(一行を並べたら次の行へ進む)にデータ列を並べる
    template< class In >
    void assign( In s, In e, size_type sz, bool inRows );

    /// @brief 指定した行番号の反復子を返す(添字演算子)
    ///
    /// @note 添字に対するチェックは行わない。
    ///
    /// @param r 行番号
    /// @return 行の反復子
    iterator operator[]( size_type r )
    { return( Matrix< T >::operator[]( r ) ); }

    /// @brief 指定した行番号の定数反復子を返す(添字演算子)
    ///
    /// @note 添字に対するチェックは行わない。
    ///
    /// @param r 行番号
    /// @return 行の定数反復子
    const_iterator operator[]( size_type r ) const
    { return( Matrix< T >::operator[]( r ) ); }

    /// @brief 指定した行番号の反復子を返す
    ///
    /// - 範囲外の行番号を指定した場合は out_of_range 例外を投げる。
    /// - 列方向の飛び幅がゼロの場合、ExceptionNotPositiveNumber< SquareMatrix< T >::size_type > 例外を投げる。
    ///
    /// @param r 行番号
    /// @param start 列番号の開始位置
    /// @param stride 列方向の飛び幅
    /// @return 行の反復子
    iterator row( size_type r, size_type start = 0, size_type stride = 1 )
    { return( Matrix< T >::row( r, start, stride ) ); }

    /// @brief 指定した行番号の定数反復子を返す
    ///
    /// - 範囲外の行番号を指定した場合は out_of_range 例外を投げる。
    /// - 列方向の飛び幅がゼロの場合、ExceptionNotPositiveNumber< SquareMatrix< T >::size_type > 例外を投げる。
    ///
    /// @param r 行番号
    /// @param start 列番号の開始位置
    /// @param stride 列方向の飛び幅
    /// @return 行の定数反復子
    const_iterator row( size_type r, size_type start = 0, size_type stride = 1 ) const
    { return( Matrix< T >::row( r, start, stride ) ); }

    /// @brief 指定した列番号の反復子を返す
    ///
    /// @note 添字に対するチェックは行わない。
    ///
    /// @param c 列番号
    /// @return 列の反復子
    iterator cbegin( size_type c )
    { return( Matrix< T >::cbegin( c ) ); }

    /// @brief 指定した列番号の定数反復子を返す
    ///
    /// @note 添字に対するチェックは行わない。
    ///
    /// @param c 列番号
    /// @return 列の定数反復子
    const_iterator cbegin( size_type c ) const
    { return( Matrix< T >::cbegin( c ) ); }

    /// @brief 指定した列番号の反復子を返す
    ///
    /// - 範囲外の列番号を指定した場合は out_of_range 例外を投げる。
    /// - 行方向の飛び幅がゼロの場合、ExceptionNotPositiveNumber< SquareMatrix< T >::size_type > 例外を投げる。
    ///
    /// @param c 列番号
    /// @param start 行番号の開始位置
    /// @param stride 行方向の飛び幅
    /// @return 列の反復子
    iterator column( size_type c, size_type start = 0, size_type stride = 1 )
    { return( Matrix< T >::column( c, start, stride ) ); }

    /// @brief 指定した列番号の定数反復子を返す
    ///
    /// - 範囲外の列番号を指定した場合は out_of_range 例外を投げる。
    /// - 行方向の飛び幅がゼロの場合、ExceptionNotPositiveNumber< SquareMatrix< T >::size_type > 例外を投げる。
    ///
    /// @param c 列番号
    /// @param start 行番号の開始位置
    /// @param stride 行方向の飛び幅
    /// @return 列の定数反復子
    const_iterator column( size_type c, size_type start = 0, size_type stride = 1 ) const
    { return( Matrix< T >::column( c, start, stride ) ); }

    /// @brief 先頭位置から始まる全要素へアクセス可能な反復子を返す
    ///
    /// この反復子は先頭から同列の次の行へとたどり、末尾に達したら次の列の先頭に移る。
    ///
    /// @return 対象の反復子
    iterator begin()
    { return( Matrix< T >::begin() ); }

    /// @brief 先頭位置から始まる全要素へアクセス可能な定数反復子を返す
    ///
    /// この反復子は先頭から同列の次の行へとたどり、末尾に達したら次の列の先頭に移る。
    ///
    /// @return 対象の定数反復子
    const_iterator begin() const
    { return( Matrix< T >::begin() ); }

    /// @brief 要素の末尾を表す反復子を返す
    ///
    /// このメンバ関数で返される反復子は、begin() で得た先頭位置から始まる全要素へアクセス可能な反復子の末尾を表す。
    /// 行・列に対する反復子の末尾は、indexer_iterator::end や const_indexer_iterator::end で得ることに注意。
    /// 例えば以下は 1 列目の列ベクトルのノルムを計算するためのプログラムで、末尾の指定は a->end() ではなく col.end() となっている。
    ///
    /// @code
    /// SquareMatrix< double >::const_iterator col = a->column( 0 ); // 行列 a の 1 列目の先頭
    /// double norm = std::inner_product( col, col.end(), col, 0.0 );
    /// @endcode
    ///
    /// @return 末尾の反復子
    iterator end()
    { return( Matrix< T >::end() ); }

    /// @brief 要素の末尾を表す定数反復子を返す
    ///
    /// このメンバ関数で返される反復子は、begin() で得た先頭位置から始まる全要素へアクセス可能な反復子の末尾を表す。
    /// 行・列に対する反復子の末尾は、indexer_iterator::end や const_indexer_iterator::end で得ることに注意。
    /// 例えば以下は 1 列目の列ベクトルのノルムを計算するためのプログラムで、末尾の指定は a->end() ではなく col.end() となっている。
    ///
    /// @code
    /// SquareMatrix< double >::const_iterator col = a->column( 0 ); // 行列 a の 1 列目の先頭
    /// double norm = std::inner_product( col, col.end(), col, 0.0 );
    /// @endcode
    ///
    /// @return 末尾の定数反復子
    const_iterator end() const
    { return( Matrix< T >::end() ); }

    /// @brief 要素の転置を行う
    ///
    /// @return なし
    void transpose();

    /// @brief 行列と掛け合わせる
    ///
    /// - サイズが、掛け合わせる行列のサイズと等しくないときは、ExceptionNotEqualLength< SquareMatrix< T >::size_type > を投げる。
    ///
    /// @todo MatrixLib::mult( const MATRIX1& mat1, const MATRIX2& mat2, SquareMatrix< Res >* res ) が
    ///       あるので不要。将来は廃止予定。
    ///
    /// @param mat 対象行列への参照
    /// @return 自分自身への参照
    template< typename U >
    SquareMatrix< T >& operator*=( const SquareMatrix< U >& mat );
  };

  /**
     @brief 対称行列クラス

     テンプレート引数の T は要素の型を表す。
     SymmetricMatrix クラスは要素へのアクセスやサイズ変更・取得などの関数を持つが、
     演算用のメンバ関数はないため四則演算を持たないような型も要素にすることができる。

     SymmetricMatrix には専用の反復子が indexer_iterator::Normal と indexer_iterator::Constant を利用して用意されている。
     SymmetricMatrix の行や列に対して反復子を使ってループ処理を行うことができる。
     また、添字を使って直接特定の要素へアクセスすることも可能である。

     SymmetricMatrix は、対角を軸に対称な要素が等しいことが保証される。
     そのため、非対角成分を変更した場合、対称な要素も同時に変更されることに注意。
     もし、それを回避する必要があるときは Matrix や SquareMatrix へコピーして利用する。
  **/
  template< typename T >
  class SymmetricMatrix
  {
  public:

    /// @brief 要素の型
    using value_type = T;
    /// @brief 反復子の型
    using iterator = indexer_iterator::Normal< std::valarray< T >, indexer::SymMat >;
    /// @brief 定数反復子の型
    using const_iterator = indexer_iterator::Constant< std::valarray< T >, indexer::SymMat >;
    /// @brief 添字の型
    using size_type = size_t;

  private:

    std::valarray< T > container_; // 値を格納する配列
    size_type sz_;  // 行列数

    // 行・列番号 = ( r, c ) に対する添字を返す(範囲チェックなし)
    size_type getIndexUnsafe( size_type r, size_type c ) const;

    // 行・列番号 = ( r, c ) に対する添字を返す
    size_type getIndex( size_type r, size_type c ) const;

    // 行 r、開始位置 start に対する indexer::SymMat の cols を返す
    size_type getCols( size_type r, size_type start ) const;

    // maxSize のサイズを stride ずつアクセスする場合の回数を返す
    static size_type getSize( size_type maxSize, size_type stride );

  public:

    /// @brief デフォルト・コンストラクタ
    ///
    /// 行列数はゼロで初期化する。
    SymmetricMatrix() : sz_( 0 ) {}

    /// @brief 行・列数を指定して構築
    ///
    /// 要素は初期値 val で初期化される。val のデフォルトは value_type のデフォルト値になる。
    ///
    /// @param size 行・列数
    /// @param val 初期値
    explicit SymmetricMatrix( size_type size, value_type val = value_type() )
      : container_( val, size * ( size + 1 ) / 2 ), sz_( size ) {}

    /// @brief データ列を指定して構築
    ///
    /// 内部では assign( In s, In e, size_type sz ) を呼び出している。
    /// 処理の詳細については assign( In s, In e, size_type sz ) を参照。
    ///
    /// @param s データ列の開始位置
    /// @param e データ列の末尾の次の位置
    /// @param sz 行(または列)のサイズ
    template< class In >
    SymmetricMatrix( In s, In e, size_type sz )
    { assign( s, e, sz ); }

    /// @brief 対称行列の要素をコピーして構築(コピー・コンストラクタ)
    ///
    /// @param mat コピー元の行列
    SymmetricMatrix( const SymmetricMatrix& mat );

    /// @brief 型の異なる対称行列の要素をコピーして構築
    ///
    /// @note 代入元と要素の型が異なるため型の変換が発生することに注意。
    ///
    /// @param mat コピー元の行列
    template< class U >
    SymmetricMatrix( const SymmetricMatrix< U >& mat );

    /// @brief 代入演算子の多重定義
    ///
    /// @param mat 代入元の行列
    SymmetricMatrix& operator=( const SymmetricMatrix& mat );

    /// @brief 代入演算子の多重定義
    ///
    /// @note 代入元と要素の型が異なるため型の変換が発生することに注意。
    ///
    /// @param mat 代入元の行列
    template< class U >
    SymmetricMatrix& operator=( const SymmetricMatrix< U >& mat );

    /// @brief 行列のサイズを返す
    ///
    /// @return 行列のサイズ
    size_type size() const
    { return( sz_ ); }

    /// @brief 行数を返す
    ///
    /// Matrix との互換性のためにあるが、SymmetricMatrix< T >::size() と内容は同じである。
    ///
    /// @return 行数
    size_type rows() const
    { return( size() ); }

    /// @brief 列数を返す
    ///
    /// Matrix との互換性のためにあるが、SymmetricMatrix< T >::size() と内容は同じである。
    ///
    /// @return 列数
    size_type cols() const
    { return( size() ); }

    /// @brief 要素数 ( 行数 × 列数 ) を変更する
    ///
    /// 行・列それぞれの方向に対し、サイズが小さくなる場合は末尾側の要素は削除され、
    /// 大きくなる場合は増分が初期値 val で埋められる。
    /// 元の要素は(削除されなければ)残る。その必要がない場合は、
    /// assign( size_type size, value_type val ) を使ったほうが効率がよい。
    ///
    /// @param size 行・列数
    /// @param val 増分に対する初期値
    /// @return なし
    void resize( size_type size, value_type val = value_type() );

    /// @brief 要素数 ( 行数 × 列数 ) を変更し、値を初期化する
    ///
    /// @param size 行・列数
    /// @param val 初期化する値
    /// @return なし
    void assign( size_type size, value_type val = value_type() )
    {
      container_.resize( size * ( size + 1 ) / 2, val );
      sz_ = size;
    }

    /// @brief データ列を指定して行列を初期化する
    ///
    /// データ列から行列への要素の代入方法が特殊であることに注意。
    /// まず、1 列目には先頭からサイズ分の要素がコピーされる。
    /// 次の行からは、対角成分からスタートしてデータをコピーしていく。
    /// 非対角成分は対角を軸に対称なので、データは片側しか代入しない。
    /// サイズ n の対称行列に過不足なくデータを代入するには n( n + 1 ) / 2 個の要素が必要となる。
    ///
    /// <pre>
    /// ( 1, 2, 3, 4, 5, 6 ) → | 1     |
    ///                         | 2 4   |
    ///                         | 3 5 6 |
    /// </pre>
    ///
    /// データが不足した場合、残りの要素は value_type のデフォルト値となる。
    ///
    /// テンプレート引数の In はデータ列 [ s, e ) に対する反復子の型を表す。
    ///
    /// - データ列 [ s, e ) の指定が不正だった場合、ExceptionNegativeLength< std::iterator_traits< In >::difference_type >  例外を投げる。
    ///
    /// @param s データ列の開始位置
    /// @param e データ列の末尾の次の位置
    /// @param sz 行(または列)のサイズ
    template< class In >
    void assign( In s, In e, size_type sz );

    /// @brief 指定した行番号の反復子を返す(添字演算子)
    ///
    /// @note 添字に対するチェックは行わない。
    ///
    /// @param r 行番号
    /// @return 行の反復子
    iterator operator[]( size_type r )
    {
      return( iterator( container_, indexer::SymMat( getIndexUnsafe( r, 0 ), r + 1, size(), size() - 1, 1 ) ) );
    }

    /// @brief 指定した行番号の定数反復子を返す(添字演算子)
    ///
    /// @note 添字に対するチェックは行わない。
    ///
    /// @param r 行番号
    /// @return 行の定数反復子
    const_iterator operator[]( size_type r ) const
    {
      return( const_iterator( container_, indexer::SymMat( getIndexUnsafe( r, 0 ), r + 1, size(), size() - 1, 1 ) ) );
    }

    /// @brief 指定した行番号の反復子を返す
    ///
    /// - 範囲外の行番号を指定した場合は out_of_range 例外を投げる。
    /// - 列方向の飛び幅がゼロの場合、ExceptionNotPositiveNumber< SymmetricMatrix< T >::size_type > 例外を投げる。
    ///
    /// @param r 行番号
    /// @param start 列番号の開始位置
    /// @param stride 列方向の飛び幅
    /// @return 行の反復子
    iterator row( size_type r, size_type start = 0, size_type stride = 1 )
    {
      return( iterator( container_,
			indexer::SymMat( getIndex( r, start ),
					 getCols( r, start ),
					 getSize( size() - start, stride ),
					 size() - start - 1, stride )
			) );
    }

    /// @brief 指定した行番号の定数反復子を返す
    ///
    /// - 範囲外の行番号を指定した場合は out_of_range 例外を投げる。
    /// - 列方向の飛び幅がゼロの場合、ExceptionNotPositiveNumber< SymmetricMatrix< T >::size_type > 例外を投げる。
    ///
    /// @param r 行番号
    /// @param start 列番号の開始位置
    /// @param stride 列方向の飛び幅
    /// @return 行の定数反復子
    const_iterator row( size_type r, size_type start = 0, size_type stride = 1 ) const
    {
      return( const_iterator( container_,
			      indexer::SymMat( getIndex( r, start ),
					       getCols( r, start ),
					       getSize( size() - start, stride ),
					       size() - start - 1, stride )
			      ) );
    }

    /// @brief 指定した列番号の反復子を返す(添字演算子)
    ///
    /// @note 添字に対するチェックは行わない。
    ///
    /// @param c 列番号
    /// @return 列の反復子
    iterator cbegin( size_type c )
    { return( (*this)[c] ); }

    /// @brief 指定した列番号の定数反復子を返す(添字演算子)
    ///
    /// @note 添字に対するチェックは行わない。
    ///
    /// @param c 列番号
    /// @return 列の定数反復子
    const_iterator cbegin( size_type c ) const
    { return( (*this)[c] ); }

    /// @brief 指定した列番号の反復子を返す
    ///
    /// - 範囲外の列番号を指定した場合は out_of_range 例外を投げる。
    /// - 行方向の飛び幅がゼロの場合、ExceptionNotPositiveNumber< SymmetricMatrix< T >::size_type > 例外を投げる。
    ///
    /// @param c 列番号
    /// @param start 行番号の開始位置
    /// @param stride 行方向の飛び幅
    /// @return 列の反復子
    iterator column( size_type c, size_type start = 0, size_type stride = 1 )
    { return( row( c, start, stride ) ); }

    /// @brief 指定した列番号の定数反復子を返す
    ///
    /// - 範囲外の列番号を指定した場合は out_of_range 例外を投げる。
    /// - 行方向の飛び幅がゼロの場合、ExceptionNotPositiveNumber< SymmetricMatrix< T >::size_type > 例外を投げる。
    ///
    /// @param c 列番号
    /// @param start 行番号の開始位置
    /// @param stride 行方向の飛び幅
    /// @return 列の定数反復子
    const_iterator column( size_type c, size_type start = 0, size_type stride = 1 ) const
    { return( row( c, start, stride ) ); }

  };

} // namespace tadah_fussy

#include "matrix_inline.h"

#endif
