/**
   @file matrix_inline.h
   @brief 行列用インライン関数

   @author tadah_fussy
   @date 2016/01/14 コメント追加
   @date 2019/03/29 コメントの見直し
   @date 2019/04/08
                    - コピー・コンストラクタ、代入演算子の見直し
                    - assign( In s, In e, size_type sz, bool inRows ) において、
                      行列数をゼロとしてした場合は空の行列を生成する(例外は投げない)ように変更
   @date 2019/04/18 mult/multMV/multVM のバグ修正(自己参照時の対応がされていなかった)
   @date 2019/04/24 CheckMultMatrix/CheckMultMV/CheckMultVM の追加
   @date 2021/03/13 名前空間の変更
   @date 2021/04/02 文字コードを UTF-8 に変更
**/

#include <vector>

namespace tadah_fussy
{
  namespace error_lib
  {
    /// @brief 行列どうしの乗算が可能かチェックする
    ///
    /// 行列どうしの乗算が可能かチェックを行い、パスしなかった場合は例外 tadah_fussy::exception::NotEqualLength を投げる。
    ///
    /// テンプレート引数の MATRIX1, MATRIX2 は行列の型を表す。
    /// tadah_fussy::exception::NotEqualLength のテンプレート引数は MATRIX1::size_type となることに注意。
    ///
    /// @param mat1 行列(被乗数)
    /// @param mat2 行列(乗数)
    /// @param info 例外に関する情報
    /// @return なし
    template< class MATRIX1, class MATRIX2 >
    void CheckMultMatrix( const MATRIX1& mat1, const MATRIX2& mat2, const std::string& info )
    {
      using size_type = typename MATRIX1::size_type;

      CheckTool
        ( std::bind( std::equal_to< size_type >(), std::placeholders::_1, mat1.cols() ),
          exception::NotEqualLength< size_type >( info, mat1.cols(), mat2.rows() ),
          mat2.rows()
          );
    }

    /// @brief 行列どうしの乗算が可能かチェックする
    ///
    /// 行列どうしの乗算が可能かチェックを行い、パスしなかった場合は例外 tadah_fussy::exception::NotEqualLength を投げる。
    ///
    /// テンプレート引数の MATRIX1, MATRIX2 は行列の型を表す。
    /// tadah_fussy::exception::NotEqualLength のテンプレート引数は MATRIX1::size_type となることに注意。
    ///
    /// @param mat1 行列(被乗数)
    /// @param mat2 行列(乗数)
    /// @return なし
    template< class MATRIX1, class MATRIX2 >
    void CheckMultMatrix( const MATRIX1& mat1, const MATRIX2& mat2 )
    {
      const std::string INFO = "tadah_fussy::exception::NotEqualLength at CheckMultMatrix";

      CheckMultMatrix( mat1, mat2, INFO );
    }

    /// @brief 行列への列ベクトルの乗算が可能かチェックする
    ///
    /// 行列への列ベクトルの乗算が可能かチェックを行う。
    /// 行列の列数とベクトルのサイズが一致しなかった場合は例外 tadah_fussy::exception::NotEqualLength を投げる。
    /// また、ベクトルの範囲指定が異常(負数)だった場合、tadah_fussy::exception::NegativeLength を投げる。
    ///
    /// テンプレート引数の MATRIX は行列の、In はベクトルへの反復子の型を表す。
    /// tadah_fussy::exception::NotEqualLength のテンプレート引数は MATRIX::size_type となることに注意。
    /// tadah_fussy::exception::NegativeLength のテンプレート引数は std::iterator_traits< In >::difference_type となる。
    ///
    /// @param mat 行列(被乗数)
    /// @param s, e 列ベクトル(乗数)への開始・末尾の次の位置
    /// @param infoNegativeLength 例外 tadah_fussy::exception::NegativeLength に関する情報
    /// @param infoNotEqualLength 例外 tadah_fussy::exception::NotEqualLength に関する情報
    /// @return なし
    template< class MATRIX, class In >
    void CheckMultMV( const MATRIX& mat, In s, In e,
                      const std::string& infoNegativeLength,
                      const std::string& infoNotEqualLength )
    {
      using size_type = typename MATRIX::size_type;

      auto vecLen = ArrayLength( s, e, infoNegativeLength );

      CheckTool
        ( std::bind( std::equal_to< size_type >(), std::placeholders::_1, mat.cols() ),
          exception::NotEqualLength< size_type >( infoNotEqualLength, mat.cols(), vecLen ),
          vecLen
          );
    }

    /// @brief 行列への列ベクトルの乗算が可能かチェックする
    ///
    /// 行列への列ベクトルの乗算が可能かチェックを行う。
    /// 行列の列数とベクトルのサイズが一致しなかった場合は例外 tadah_fussy::exception::NotEqualLength を投げる。
    /// また、ベクトルの範囲指定が異常(負数)だった場合、tadah_fussy::exception::NegativeLength を投げる。
    ///
    /// テンプレート引数の MATRIX は行列の、In はベクトルへの反復子の型を表す。
    /// tadah_fussy::exception::NotEqualLength のテンプレート引数は MATRIX::size_type となることに注意。
    /// tadah_fussy::exception::NegativeLength のテンプレート引数は std::iterator_traits< In >::difference_type となる。
    ///
    /// @param mat 行列(被乗数)
    /// @param s, e 列ベクトル(乗数)への開始・末尾の次の位置
    /// @return なし
    template< class MATRIX, class In >
    void CheckMultMV( const MATRIX& mat, In s, In e )
    {
      const std::string NEGATIVE_LENGTH_INFO = "tadah_fussy::exception::NegativeLength at CheckMultMV";
      const std::string NOT_EQUAL_LENGTH_INFO = "tadah_fussy::exception::NotEqualLength at CheckMultMV";

      CheckMultMV( mat, s, e, NEGATIVE_LENGTH_INFO, NOT_EQUAL_LENGTH_INFO );
    }

    /// @brief 行ベクトルへの行列の乗算が可能かチェックする
    ///
    /// 行ベクトルへの行列の乗算が可能かチェックを行う。
    /// ベクトルのサイズと行列の行数が一致しなかった場合は例外 tadah_fussy::exception::NotEqualLength を投げる。
    /// また、ベクトルの範囲指定が異常(負数)だった場合、tadah_fussy::exception::NegativeLength を投げる。
    ///
    /// テンプレート引数の MATRIX は行列の、In はベクトルへの反復子の型を表す。
    /// tadah_fussy::exception::NotEqualLength のテンプレート引数は std::iterator_traits< In >::difference_type となることに注意。
    /// tadah_fussy::exception::NegativeLength のテンプレート引数はstd::iterator_traits< In >::difference_type となる。
    ///
    /// @param s, e 行ベクトル(被乗数)への開始・末尾の次の位置
    /// @param mat 行列(乗数)
    /// @param infoNegativeLength 例外 tadah_fussy::exception::NegativeLength に関する情報
    /// @param infoNotEqualLength 例外 tadah_fussy::exception::NotEqualLength に関する情報
    /// @return なし
    template< class In, class MATRIX >
    void CheckMultVM( const In s, In e, MATRIX& mat,
                      const std::string& infoNegativeLength,
                      const std::string& infoNotEqualLength )
    {
      using difference_type = typename std::iterator_traits< In >::difference_type;

      auto vecLen = ArrayLength( s, e, infoNegativeLength );

      CheckTool
        ( std::bind2nd( std::equal_to< difference_type >(), vecLen ),
          exception::NotEqualLength< difference_type >( infoNotEqualLength, vecLen, mat.rows() ),
          mat.rows()
          );
    }

    /// @brief 行ベクトルへの行列の乗算が可能かチェックする
    ///
    /// 行ベクトルへの行列の乗算が可能かチェックを行う。
    /// ベクトルのサイズと行列の行数が一致しなかった場合は例外 tadah_fussy::exception::NotEqualLength を投げる。
    /// また、ベクトルの範囲指定が異常(負数)だった場合、tadah_fussy::exception::NegativeLength を投げる。
    ///
    /// テンプレート引数の MATRIX は行列の、In はベクトルへの反復子の型を表す。
    /// tadah_fussy::exception::NotEqualLength のテンプレート引数は std::iterator_traits< In >::difference_type となることに注意。
    /// tadah_fussy::exception::NegativeLength のテンプレート引数はstd::iterator_traits< In >::difference_type となる。
    ///
    /// @param s, e 行ベクトル(被乗数)への開始・末尾の次の位置
    /// @param mat 行列(乗数)
    /// @return なし
    template< class In, class MATRIX >
    void CheckMultVM( In s, In e, const MATRIX& mat )
    {
      const std::string NEGATIVE_LENGTH_INFO = "tadah_fussy::exception::NegativeLength at CheckMultVM";
      const std::string NOT_EQUAL_LENGTH_INFO = "tadah_fussy::exception::NotEqualLength at CheckMultVM";

      CheckMultVM( s, e, mat, NEGATIVE_LENGTH_INFO, NOT_EQUAL_LENGTH_INFO );
    }

  } // namespace error_lib

  namespace indexer
  {
    /* ------------------------------------------------------------------
       Slice : slice を利用した添字演算子
       ------------------------------------------------------------------ */

    /*
      Slice::operator() : 添字 i に対し、実際にアクセスする添字を返す
    */
    inline size_t Slice::operator()( size_t i ) const
    {
      assert( i < size() );

      return( s_.start() + i * s_.stride() );
    }

    /* -----------------------------------------------------------------------
       SymMat : 対称行列に対する行・列へのアクセスを想定した添字演算子
       ----------------------------------------------------------------------- */

    /*
      SymMat::operator() : 添字 i に対し、実際にアクセスする添字を返す
    */
    inline size_t SymMat::operator()( size_t i ) const
    {
      assert( i < size() );

      i *= step_;

      if ( i < cols_ )
	return( start_ + i * ( 2 * stride_ - i + 1 ) / 2 );
      else
	return( start_ + ( cols_ - 1 ) * ( 2 * stride_ - cols_ + 2 ) / 2 + i - cols_ + 1 );
    }

  } // namespace indexer

  /* ------------------------------------------------------------------
     Matrix : 行列クラス
     ------------------------------------------------------------------ */

  /*
    CopySymmetricMatrix : 対称行列の要素をコピー
  */
  template< class T, class U >
  void CopySymmetricMatrix( const SymmetricMatrix< U >& sym, Matrix< T >* mat )
  {
    mat->assign( sym.rows(), sym.cols() );

    for ( typename Matrix< T >::size_type r = 0 ; r < mat->rows() ; ++r )
      for ( typename Matrix< T >::size_type c = r ; c < mat->cols() ; ++c )
	(*mat)[r][c] = (*mat)[c][r] = sym[r][c];
  }

  /*
    Matrix< T > コンストラクタ : 対称行列の要素をコピーして構築
  */
  template< class T >
  template< class U >
  Matrix< T >::Matrix( const SymmetricMatrix< U >& mat )
  { CopySymmetricMatrix( mat, this ); }

  /*
    Matrix< T >::operator= : 代入演算子の多重定義
  */

  // Matrix< T > の代入
  template< class T >
  Matrix< T >& Matrix< T >::operator=( const Matrix& mat )
  {
    if ( this == &mat ) return( *this );

    rows_ = mat.rows();
    cols_ = mat.cols();
    container_.resize( rows_ * cols_ );
    container_ = mat.container_;

    return( *this );
  }

  // Matrix< U > の代入
  template< class T >
  template< class U >
  Matrix< T >& Matrix< T >::operator=( const Matrix< U >& mat )
  {
    assign( mat.begin(), mat.end(), mat.rows(), true );

    return( *this );
  }

  // 正方行列の代入
  template< class T >
  template< class U >
  Matrix< T >& Matrix< T >::operator=( const SquareMatrix< U >& mat )
  {
    assign( mat.begin(), mat.end(), mat.rows(), true );

    return( *this );
  }

  // 対称行列の代入
  template< class T >
  template< class U >
  Matrix< T >& Matrix< T >::operator=( const SymmetricMatrix< U >& mat )
  {
    CopySymmetricMatrix( mat, this );

    return( *this );
  }

  /*
    Matrix< T >::getIndex : 行・列番号 = ( r, c ) に対する添字を返す
  */
  template< class T >
  typename Matrix< T >::size_type
  Matrix< T >::getIndex( size_type r, size_type c ) const
  {
    if ( r >= rows() )
      throw std::out_of_range( "Matrix< T >::getIndex : Out of row range." );
    if ( c >= cols() )
      throw std::out_of_range( "Matrix< T >::getIndex : Out of column range." );

    return( c * rows() + r );
  }

  /*
    Matrix< T >::getSize : maxSize のサイズを stride ずつアクセスする場合の回数を返す
  */
  template< class T >
  typename Matrix< T >::size_type
  Matrix< T >::getSize( size_type maxSize, size_type stride )
  {
    const std::string FUNC_NAME = "Matrix< T >::getSize";

    error_lib::PNum( stride, FUNC_NAME + " : stride " + error_message::PNUM );

    if ( stride >= maxSize ) return( 1 );

    size_type sz = maxSize / stride;
    size_type mod = ( maxSize % stride != 0 ) ? 1 : 0;

    return( sz + mod );
  }

  /*
    Matrix< T >::resize : 行列のリサイズ
  */
  template< class T >
  void Matrix< T >::resize( size_type newRows, size_type newCols, value_type val )
  {
    std::valarray< T > buff( container_ ); // 値を退避
    // 行列数の小さい側を取得
    size_type minRows = std::min( newRows, rows() );
    size_type minCols = std::min( newCols, cols() );

    container_.resize( newRows * newCols, val );

    for ( size_type r = 0 ; r < minRows ; ++r )
      for ( size_type c = 0 ; c < minCols ; ++c )
	container_[ c * newRows + r ] = buff[ c * rows() + r ];

    // 行列数を新しいものに差し替え
    rows_ = newRows;
    cols_ = newCols;
  }

  /*
    Matrix< T >::assign : データ列 [ s, e ) を指定して行列を初期化する

    sz は行または列のサイズ、inRows が true なら列方向にデータ列を並べていく。
  */
  template< class T >
  template< class In >
  void Matrix< T >::assign( In s, In e, size_type sz, bool inRows )
  {
    const std::string FUNC_NAME = "Matrix< T >::assign";
    auto diff = error_lib::ArrayLength( s, e, FUNC_NAME + " : [ s, e ) " + error_message::NEGATIVE_LENGTH );

    // 行または列のサイズがゼロの場合、サイズがゼロの行列にする
    if ( sz == 0 ) {
      assign( 0, 0 );
      return;
    }

    // もう一方(行か列)のサイズ
    // diff がゼロの(データがない)場合、sz2 は 1 としておく
    size_type sz2 = diff / sz;
    if ( diff % sz > 0 || diff == 0 ) ++sz2;

    if ( inRows )
      assign( sz, sz2 ); // 行方向が先なら sz は行数
    else
      assign( sz2, sz ); // 列方向が先なら sz は列数

    // 要素をコピー
    for ( size_type i = 0 ; i < sz2 ; ++i ) {
      iterator it = ( inRows ) ? column( i ) : row( i );
      while ( it != it.end() && s != e )
	*it++ = *s++;
    }
  }

  /*
    Matrix< T >::row : 開始位置を start、飛び幅を stride とした行番号 r の反復子を返す
  */
  template< class T >
  typename Matrix< T >::iterator
  Matrix< T >::row( size_type r, size_type start, size_type stride )
  { return( iterator( container_, indexer::Slice( getIndex( r, start ), getSize( cols_ - start, stride ), rows_ * stride ) ) ); }

  /*
    Matrix< T >::column : 開始位置を start、飛び幅を stride とした列番号 c の反復子を返す
  */
  template< class T >
  typename Matrix< T >::iterator
  Matrix< T >::column( size_type c, size_type start, size_type stride )
  { return( iterator( container_, indexer::Slice( getIndex( start, c ), getSize( rows_ - start, stride ), stride ) ) ); }

  /*
    Matrix< T >::begin : 先頭位置から始まる全要素へアクセス可能な反復子を返す
  */
  template< class T >
  typename Matrix< T >::iterator
  Matrix< T >::begin()
  { return( iterator( container_, indexer::Slice( 0, rows_ * cols_, 1 ) ) ); }

  /*
    Matrix< T >::end : 要素の末尾を表す反復子を返す
  */
  template< class T >
  typename Matrix< T >::iterator
  Matrix< T >::end()
  { return( begin().end() ); }

  /*
    Matrix< T >::row : 開始位置を start、飛び幅を stride とした行番号 r の定数反復子を返す
  */
  template< class T >
  typename Matrix< T >::const_iterator
  Matrix< T >::row( size_type r, size_type start, size_type stride ) const
  { return( const_iterator( container_, indexer::Slice( getIndex( r, start ), getSize( cols_ - start, stride ), rows_ * stride ) ) ); }

  /*
    Matrix< T >::column : 開始位置を start、飛び幅を stride とした列番号 c の定数反復子を返す
  */
  template< class T >
  typename Matrix< T >::const_iterator
  Matrix< T >::column( size_type c, size_type start, size_type stride ) const
  { return( const_iterator( container_, indexer::Slice( getIndex( start, c ), getSize( rows_ - start, stride ), stride ) ) ); }

  /*
    Matrix< T >::begin : 先頭位置から始まる全要素へアクセス可能な定数反復子を返す
  */
  template< class T >
  typename Matrix< T >::const_iterator
  Matrix< T >::begin() const
  { return( const_iterator( container_, indexer::Slice( 0, rows_ * cols_, 1 ) ) ); }

  /*
    Matrix< T >::end : 要素の末尾を表す定数反復子を返す
  */
  template< class T >
  typename Matrix< T >::const_iterator
  Matrix< T >::end() const
  { return( begin().end() ); }

  /*
    Matrix< T >::transpose : 要素の転置を行う
  */
  template< class T >
  void Matrix< T >::transpose()
  {
    std::valarray< T > buff( container_ );
    std::swap( rows_, cols_ );

    for ( size_type r = 0 ; r < rows() ; ++r )
      for ( size_type c = 0 ; c < cols() ; ++c )
	container_[c * rows() + r] = buff[r * cols() + c];
  }

  /* ------------------------------------------------------------------
     SquareMatrix : 正方行列クラス
     ------------------------------------------------------------------ */

  /*
    SquareMatrix< T >::transpose : 要素の転置を行う
  */
  template< class T >
  void SquareMatrix< T >::transpose()
  {
    for ( size_type r = 0 ; r < size() ; ++r )
      for ( size_type c = r + 1 ; c < size() ; ++c )
	std::swap( (*this)[r][c], (*this)[c][r] );
  }

  /*
    SquareMatrix< T >::assign : データ列 [ s, e ) を指定して行列を初期化する

    sz は行列のサイズ、inRows が true ならデータ列は列方向へ並べていく。
  */
  template< class T >
  template< class In >
  void SquareMatrix< T >::assign( In s, In e, size_type sz, bool inRows )
  {
    const std::string FUNC_NAME = "SquareMatrix< T >::assign";

    auto diff = error_lib::ArrayLength( s, e, FUNC_NAME + " : [ s, e ) " + error_message::NEGATIVE_LENGTH );

    assign( sz );

    // 要素をコピー
    for ( size_type i = 0 ; i < sz ; ++i ) {
      iterator it = ( inRows ) ? column( i ) : row( i );
      while ( it != it.end() && s != e )
	*it++ = *s++;
    }
  }

  /*
    SquareMatrix< T >::operator*= : 行列 mat と掛け合わせる
  */
  template< class T >
  template< typename U >
  SquareMatrix< T >& SquareMatrix< T >::operator*=( const SquareMatrix< U >& mat )
  {
    const std::string FUNC_NAME = "SquareMatrix< T >::operator*=";

    error_lib::CheckMultMatrix( *this, mat, FUNC_NAME + " : this->cols() " + error_message::NOT_EQUAL_LENGTH + " mat.rows()" );

    SquareMatrix< T > buff( size() );

    for ( size_type r = 0 ; r < size() ; ++r ) {
      const_iterator row = (*this)[r];
      for ( size_type c = 0 ; c < size() ; ++c ) {
	auto col = mat.cbegin( c );
	buff[r][c] = std::inner_product( row, row.end(), col, T() );
      }
    }

    *this = buff;

    return( *this );
  }

  /* ------------------------------------------------------------------
     SymmetricMatrix : 対称行列クラス
     ------------------------------------------------------------------ */

  /*
    SymmetricMatrix< T >::getIndexUnsafe : 行・列番号 = ( r, c ) に対する添字を返す(範囲チェックなし)
  */
  template< class T >
  typename SymmetricMatrix< T >::size_type
  SymmetricMatrix< T >::getIndexUnsafe( size_type r, size_type c ) const
  {
    if ( r < c ) std::swap( r, c );

    return( size() * c - c * ( c + 1 ) / 2 + r );
  }

  /*
    SymmetricMatrix< T >::getIndex : 行・列番号 = ( r, c ) に対する添字を返す
  */
  template< class T >
  typename SymmetricMatrix< T >::size_type
  SymmetricMatrix< T >::getIndex( size_type r, size_type c ) const
  {
    if ( r >= rows() )
      throw( std::out_of_range( "SymmetricMatrix< T >::getIndex : Out of row range." ) );
    if ( c >= cols() )
      throw( std::out_of_range( "SymmetricMatrix< T >::getIndex : Out of column range." ) );

    if ( r < c ) std::swap( r, c );

    return( size() * c - c * ( c + 1 ) / 2 + r );
  }

  /*
    SymmetricMatrix< T >::getCols : 行 r、開始位置 start に対する SymMatIndexer の cols を返す
  */
  template< class T >
  typename SymmetricMatrix< T >::size_type
  SymmetricMatrix< T >::getCols( size_type r, size_type start ) const
  {
    if ( r >= rows() )
      throw( std::out_of_range( "SymmetricMatrix< T >::getCols : Out of row range." ) );
    if ( start >= size() )
      throw( std::out_of_range( "SymmetricMatrix< T >::getCols : Out of start range." ) );

    if ( r < start ) return( 1 );

    return( r - start + 1 );
  }

  /*
    SymmetricMatrix< T >::getSize : maxSize のサイズを stride ずつアクセスする場合の回数を返す
  */
  template< class T >
  typename SymmetricMatrix< T >::size_type
  SymmetricMatrix< T >::getSize( size_type maxSize, size_type stride )
  {
    const std::string FUNC_NAME = "SymmetricMatrix< T >::getSize";

    error_lib::PNum( stride, FUNC_NAME + " : stride " + error_message::PNUM );

    if ( stride >= maxSize ) return( 1 );

    size_type sz = maxSize / stride;
    size_type mod = ( maxSize % stride != 0 ) ? 1 : 0;

    return( sz + mod );
  }

  /*
    SymmetricMatrix コンストラクタ : 行列 mat の要素をコピーして構築
  */
  template< class T >
  SymmetricMatrix< T >::SymmetricMatrix( const SymmetricMatrix< T >& mat )
  {
    container_.resize( mat.container_.size() );
    container_ = mat.container_;
    sz_ = mat.sz_;
  }

  /*
    SymmetricMatrix コンストラクタ : 型の異なる対称行列の要素をコピーして構築
  */
  template< class T >
  template< class U >
  SymmetricMatrix< T >::SymmetricMatrix( const SymmetricMatrix< U >& mat )
  {
    assign( mat.size() );
    for ( SymmetricMatrix::size_type r = 0 ; r < size() ; ++r ) {
      for ( SymmetricMatrix::size_type c = r ; c < size() ; ++c ) {
	(*this)[r][c] = mat[r][c];
      }
    }
  }

  /*
    SymmetricMatrix< T >::operator= : 代入演算子の多重定義
  */
  template< class T >
  SymmetricMatrix< T >& SymmetricMatrix< T >::operator=( const SymmetricMatrix< T >& mat )
  {
    if ( &mat != this ) {
      container_.resize( mat.container_.size() );
      container_ = mat.container_;
      sz_ = mat.sz_;
    }

    return( *this );
  }

  /// @brief 代入演算子の多重定義
  ///
  /// @note 代入元と要素の型が異なるため型の変換が発生することに注意。
  ///
  /// @param mat 代入元の行列
  template< class T >
  template< class U >
  SymmetricMatrix< T >& SymmetricMatrix< T >::operator=( const SymmetricMatrix< U >& mat )
  {
    assign( mat.size() );
    for ( SymmetricMatrix::size_type r = 0 ; r < size() ; ++r ) {
      for ( SymmetricMatrix::size_type c = r ; c < size() ; ++c ) {
	(*this)[r][c] = mat[r][c];
      }
    }

    return( *this );
  }

  /*
    SymmetricMatrix< T >::resize : 行列数を newSize に変更する

    サイズを拡大した場合、初期値は val とする。
  */
  template< class T >
  void SymmetricMatrix< T >::resize( size_type newSize, value_type val )
  {
    if ( newSize == size() ) return;

    std::valarray< T > buff( container_ ); // 値を退避

    // 変更前後でのサイズの小さい側を取得
    size_type minSize = std::min( newSize, size() );

    container_.resize( newSize * ( newSize + 1 ) / 2, val );

    size_type si = 0; // 自分自身に対する添字
    size_type bi = 0; // バッファに対する添字
    for ( size_type c = 0 ; c < minSize ; ++c ) {
      for ( size_type r = 0 ; r < minSize - c ; ++r ) {
	container_[si] = buff[bi];
	++si;
	++bi;
      }
      if ( size() < newSize )
	si += newSize - size();
      else
	bi += size() - newSize;
    }

    // サイズを新しいものに差し替え
    sz_ = newSize;
  }

  /*
    SymmetricMatrix< T >::assign : データ列 [ s, e ) を指定して行列を初期化する

    初期化時の行列数は sz とする。
  */
  template< class T >
  template< class In >
  void SymmetricMatrix< T >::assign( In s, In e, size_type sz )
  {
    const std::string FUNC_NAME = "SymmetricMatrix< T >::assign";

    auto diff = error_lib::ArrayLength( s, e,
					FUNC_NAME + " : [ s, e ) " + error_message::NEGATIVE_LENGTH );

    assign( sz );

    // 要素をコピー
    size_type i = 0;
    while ( s != e ) {
      if ( i >= container_.size() ) break;
      container_[i] = *s++;
      ++i;
    }
  }

  namespace matrix_lib
  {
    /// @brief 正方行列かどうかをチェックする
    ///
    /// @note 行列のサイズがどちらもゼロなら true を返す。
    ///
    /// @param mat 対象の行列
    /// @return 正方行列なら true を返す
    template< class MATRIX >
    bool IsSquare( const MATRIX& mat )
    {
      return( mat.rows() == mat.cols() );
    }

    /// @brief 対称行列かどうかをチェックする
    ///
    /// @note 行列のサイズがどちらもゼロなら true を返す。
    ///
    /// @param mat 対象の行列
    /// @param eq 要素が等しいかを判定するための叙述関数
    /// @return 対称行列なら true を返す
    template< class MATRIX, class Eq >
    bool IsSymmetric( const MATRIX& mat, Eq eq )
    {
      using size_type = typename MATRIX::size_type;

      if ( ! IsSquare( mat ) ) return( false );
      if ( mat.rows() == 0 ) return( true );

      for ( size_type r = 0 ; r < mat.rows() - 1 ; ++r ) {
        auto upper = mat.row( r, r + 1 );
        auto lower = mat.column( r, r + 1 );
        while ( upper != upper.end() )
          if ( ! eq( *upper++, *lower++ ) ) return( false );
      }

      return( true );
    }

    /*
      mult_ : 行列 mat1 と mat2 を掛け合わせる
    */
    template< class MATRIX1, class MATRIX2, typename Res >
    void mult_( const MATRIX1& mat1, const MATRIX2& mat2, Res* res )
    {
      for ( typename MATRIX1::size_type r = 0 ; r < mat1.rows() ; ++r ) {
        typename MATRIX1::const_iterator row = mat1[r];
        for ( typename MATRIX2::size_type c = 0 ; c < mat2.cols() ; ++c ) {
          typename MATRIX2::const_iterator col = mat2.cbegin( c );
          (*res)[r][c] = std::inner_product( row, row.end(), col, typename Res::value_type() );
        }
      }
    }

    /// @brief 行列を掛け合わせる
    ///
    /// 掛け合わせる二つの行列は一般行列(Matrix)の他、正方行列(SquareMatrix)
    /// や対称行列(SymmetricMatrix)も対象とすることができるが、
    /// 結果を保持する行列へのポインタは一般行列のみとする。
    ///
    /// - mat1 の列数と mat2 の行数が異なる場合、tadah_fussy::exception::NotEqualLength< MATRIX1::size_type > 例外を投げる。
    /// - 結果を保持する行列へのポインタ res が未定義の場合は assert を実行する。
    ///
    /// @param mat1 対象行列1への参照
    /// @param mat2 対象行列2への参照
    /// @param res 結果を保持する行列へのポインタ
    template< class MATRIX1, class MATRIX2, typename Res >
    void mult( const MATRIX1& mat1, const MATRIX2& mat2, Matrix< Res >* res )
    {
      const std::string FUNC_NAME = "MatrixLib::mult";

      assert( res != 0 );

      error_lib::CheckMultMatrix( mat1, mat2, FUNC_NAME + " : mat1.cols() " + error_message::NOT_EQUAL_LENGTH + " mat2.rows()" );

      if ( &mat1 == reinterpret_cast< MATRIX1* >( res ) || &mat2 == reinterpret_cast< MATRIX2* >( res ) ) {
        Matrix< Res > buff( mat1.rows(), mat2.cols() );
        mult_( mat1, mat2, &buff );
        *res = buff;
      } else {
        res->assign( mat1.rows(), mat2.cols() );
        mult_( mat1, mat2, res );
      }
    }

    /// @brief 行列を掛け合わせる
    ///
    /// 掛け合わせる二つの行列は一般行列(Matrix)の他、正方行列(SquareMatrix)
    /// や対称行列(SymmetricMatrix)も対象とすることができるが、
    /// 結果を保持する行列へのポインタは正方行列のみとする。
    ///
    /// - mat1 の列数と mat2 の行数が異なる場合、tadah_fussy::exception::NotEqualLength< MATRIX1::size_type > 例外を投げる。
    /// - mat1 の行数と mat2 の列数が異なる場合、tadah_fussy::exception::NotEqualLength< MATRIX2::size_type > 例外を投げる。
    /// - 結果を保持する行列へのポインタ res が未定義の場合は assert を実行する。
    ///
    /// @param mat1 対象行列1への参照
    /// @param mat2 対象行列2への参照
    /// @param res 結果を保持する正方行列へのポインタ
    template< class MATRIX1, class MATRIX2, typename Res >
    void mult( const MATRIX1& mat1, const MATRIX2& mat2, SquareMatrix< Res >* res )
    {
      const std::string FUNC_NAME = "MatrixLib::mult";

      assert( res != 0 );

      error_lib::CheckMultMatrix( mat1, mat2, FUNC_NAME + " : mat1.cols() " + error_message::NOT_EQUAL_LENGTH + " mat2.rows()" );
      error_lib::CheckMultMatrix( mat2, mat1, FUNC_NAME + " : mat1.cols() " + error_message::NOT_EQUAL_LENGTH + " mat2.rows()" );

      if ( &mat1 == reinterpret_cast< MATRIX1* >( res ) || &mat2 == reinterpret_cast< MATRIX2* >( res ) ) {
        SquareMatrix< Res > buff( mat1.rows() );
        mult_( mat1, mat2, &buff );
        *res = buff;
      } else {
        res->assign( mat1.rows() );
        mult_( mat1, mat2, res );
      }
    }

    /// @brief 行列に(列)ベクトルを掛け合わせる
    ///
    /// 行列は一般行列(Matrix)の他、正方行列(SquareMatrix)や対称行列(SymmetricMatrix)も対象とすることができる。
    /// また、ベクトルは開始位置 is (os) のみで表すものとする。
    ///
    /// @note ベクトルの長さが正しいかのチェックは行っていないことに注意。
    ///
    /// @param mat 対象行列への参照
    /// @param is 掛け合わせる列ベクトルの先頭位置
    /// @param os 結果を保持するベクトルの先頭位置
    template< class MATRIX, class In, class Out >
    void multMV( const MATRIX& mat, In is, Out os )
    {
      using value_type = typename std::iterator_traits< In >::value_type;

      std::vector< value_type > buff( mat.rows() );
      for ( typename MATRIX::size_type r = 0 ; r < mat.rows() ; ++r ) {
        auto row = mat[r];
        buff[r] = std::inner_product( row, row.end(), is, value_type() );
      }

      std::copy( buff.begin(), buff.end(), os );
    }

    /// @brief (行)ベクトルに行列を掛け合わせる
    ///
    /// 行列は一般行列(Matrix)の他、正方行列(SquareMatrix)や対称行列(SymmetricMatrix)も対象とすることができる。
    /// また、ベクトルは開始位置 is (os) のみで表すものとする。
    ///
    /// @note ベクトルの長さが正しいかのチェックは行っていないことに注意。
    ///
    /// @param is 行ベクトルの先頭位置
    /// @param mat 掛け合わせる行列への参照
    /// @param os 結果を保持するベクトルの先頭位置
    template< class In, class MATRIX, class Out >
    void multVM( In is, const MATRIX& mat, Out os )
    {
      using value_type = typename std::iterator_traits< In >::value_type;

      std::vector< value_type > buff( mat.cols() );
      for ( typename MATRIX::size_type c = 0 ; c < mat.cols() ; ++c ) {
        auto col = mat.cbegin( c );
        buff[c] = std::inner_product( col, col.end(), is, value_type() );
      }

      std::copy( buff.begin(), buff.end(), os );
    }

    /// @brief 行列の行 r1, r2 を入れ替える
    ///
    /// @note 行番号のチェックは行わないことに注意
    ///
    /// @param mat 対象の行列へのポインタ
    /// @param r1, r2 入れ替え対象の行番号
    /// @return なし
    template< class T >
    void SwapRow( Matrix< T >* mat, typename Matrix< T >::size_type r1, typename Matrix< T >::size_type r2 )
    {
      auto i1 = (*mat)[r1];
      auto i2 = (*mat)[r2];

      std::swap_ranges( i1, i1.end(), i2 );
    }

    /// @brief 正方行列の行 r1, r2 を入れ替える
    ///
    /// @note 行番号のチェックは行わないことに注意
    ///
    /// @param mat 対象の正方行列へのポインタ
    /// @param r1, r2 入れ替え対象の行番号
    /// @return なし
    template< class T >
    void SwapRow( SquareMatrix< T >* mat, typename SquareMatrix< T >::size_type r1, typename SquareMatrix< T >::size_type r2 )
    {
      auto i1 = (*mat)[r1];
      auto i2 = (*mat)[r2];

      std::swap_ranges( i1, i1.end(), i2 );
    }

    /// @brief 行列の列 c1, c2 を入れ替える
    ///
    /// @note 列番号のチェックは行わないことに注意
    ///
    /// @param mat 対象の行列へのポインタ
    /// @param c1, c2 入れ替え対象の列番号
    /// @return なし
    template< class T >
    void SwapCol( Matrix< T >* mat, typename Matrix< T >::size_type c1, typename Matrix< T >::size_type c2 )
    {
      auto i1 = mat->cbegin( c1 );
      auto i2 = mat->cbegin( c2 );

      std::swap_ranges( i1, i1.end(), i2 );
    }

    /// @brief 正方行列の列 c1, c2 を入れ替える
    ///
    /// @note 列番号のチェックは行わないことに注意
    ///
    /// @param mat 対象の正方行列へのポインタ
    /// @param c1, c2 入れ替え対象の列番号
    /// @return なし
    template< class T >
    void SwapCol( SquareMatrix< T >* mat, typename SquareMatrix< T >::size_type c1, typename SquareMatrix< T >::size_type c2 )
    {
      auto i1 = mat->cbegin( c1 );
      auto i2 = mat->cbegin( c2 );

      std::swap_ranges( i1, i1.end(), i2 );
    }

    /// @brief 配列の出力
    ///
    /// @param os 出力対象のストリーム
    /// @param s 配列の開始位置
    /// @param e 配列の末尾の次の位置
    /// @param header 出力前に付けるヘッダ
    /// @param op_bracket 配列の先頭に付ける括弧
    /// @param delimiter 配列の各要素を分けるデリミタ
    /// @param cls_bracket 配列の末尾に付ける括弧
    template< class In >
    void PrintArray( std::ostream& os, In s, In e,
                     const std::string& header = "", const std::string& op_bracket = "( ",
                     const std::string& delimiter = ", ", const std::string& cls_bracket = " )" )
    {
      os << header << op_bracket;
      if ( s != e ) os << *s++;
      for ( ; s != e ; ++s )
        os << delimiter << *s;
      os << cls_bracket;
    }

    /// @brief 正方行列・対称行列の出力
    ///
    /// @param os 出力対象のストリーム
    /// @param mat 出力する行列
    /// @param header 行列の先頭に付けるヘッダ
    /// @param op_bracket 各行ベクトルの先頭に付ける括弧
    /// @param delimiter 行列の各要素を分けるデリミタ
    /// @param cls_bracket 各行ベクトルの末尾に付ける括弧
    template< class MATRIX >
    void PrintMatrix( std::ostream& os, const MATRIX& mat,
                      const std::string& header = "", const std::string& op_bracket = "| ",
                      const std::string& delimiter = ", ", const std::string& cls_bracket = " |" )
    {
      if ( mat.rows() == 0 || mat.cols() == 0 ) {
        os << header << op_bracket << cls_bracket;
        return;
      }

      // 一行目の出力
      auto it = mat.row( 0 );
      PrintArray( os, it, it.end(), header, op_bracket, delimiter, cls_bracket  );

      std::string space( header.length(), ' ' ); // 二行目以降の空白

      for ( typename MATRIX::size_type r = 1 ; r < mat.rows() ; ++r ) {
        auto it = mat.row( r );
        os << std::endl;
        PrintArray( os, it, it.end(), space, op_bracket, delimiter, cls_bracket );
      }
    }

  } // namespace matrix_lib

  /// @brief ストリーム出力(行列)
  ///
  /// @param os 対象の出力ストリームへの参照
  /// @param mat 出力対象の行列への参照
  /// @return 出力したストリーム(os)への参照
  template< class T >
  std::ostream& operator<<( std::ostream& os, const Matrix< T >& mat )
  {
    matrix_lib::PrintMatrix( os, mat );

    return( os );
  }

  /// @brief ストリーム出力(正方行列)
  ///
  /// @param os 対象の出力ストリームへの参照
  /// @param mat 出力対象の正方行列への参照
  /// @return 出力したストリーム(os)への参照
  template< class T >
  std::ostream& operator<<( std::ostream& os, const SquareMatrix< T >& mat )
  {
    matrix_lib::PrintMatrix( os, mat );

    return( os );
  }

  /// @brief ストリーム出力(対称行列)
  ///
  /// @param os 対象の出力ストリームへの参照
  /// @param mat 出力対象の対称行列への参照
  /// @return 出力したストリーム(os)への参照
  template< class T >
  std::ostream& operator<<( std::ostream& os, const SymmetricMatrix< T >& mat )
  {
    matrix_lib::PrintMatrix( os, mat );

    return( os );
  }

  /**
     @brief エラーチェック・処理ルーチン用名前空間

     エラーチェック・処理用の関数をまとめた名前空間
  **/
  namespace error_lib
  {
    /// @brief 線形モデルとして有効かをチェックする
    ///
    /// 線形モデルとして利用可能か以下のテストを行い、テストにパスしなかった場合は例外を投げる。
    ///
    /// -# 要素がゼロでないか
    /// -# 従属変数の要素の範囲が適切か(開始と終了が逆になっていないか)
    /// -# デザイン行列の行数と従属変数の要素数が一致しているか
    ///
    /// 1 のテストにパスしなかった場合は例外 tadah_fussy::exception::NotPositiveLength が、
    /// 2 のテストにパスしなかった場合は例外 tadah_fussy::exception::NegativeLength が、
    /// 3 のテストにパスしなかった場合は例外 tadah_fussy::exception::NotEqualLength が投げられる。
    ///
    /// テンプレート引数の MATRIX はデザイン行列の型を、
    /// テンプレート引数の In は従属変数のデータ列の反復子の型をそれぞれ表す。
    ///
    /// tadah_fussy::exception::NotPositiveLength のテンプレート引数は MATRIX::size_type となる。
    ///
    /// tadah_fussy::exception::NegativeLength のテンプレート引数は std::iterator_traits< In >::difference_type となる。
    ///
    /// tadah_fussy::exception::NotEqualLength のテンプレート引数は MATRIX::size_type となることに注意。
    ///
    /// @param x デザイン行列への参照
    /// @param ys 従属変数の要素の開始位置
    /// @param ye 従属変数の要素の末尾の次の位置
    /// @param infoNotPositiveLength デザイン行列 x での例外 tadah_fussy::exception::NotPositiveLength に関する情報
    /// @param infoNegativeLength 従属変数 [ ys, ye ) での例外 tadah_fussy::exception::NegativeLength に関する情報
    /// @param infoNotEqualLength 例外 tadah_fussy::exception::NotEqualLength に関する情報
    /// @return なし
    template< class MATRIX, class In >
    void CheckLinearModel( const MATRIX& x, In ys, In ye,
                           const std::string& infoNotPositiveLength,
                           const std::string& infoNegativeLength,
                           const std::string& infoNotEqualLength )
    {
      using DX = typename MATRIX::size_type;
      using DY = typename std::iterator_traits< In >::difference_type;

      CheckTool
        ( std::bind( std::greater< DX >(), std::placeholders::_1, DX() ),
          exception::NotPositiveLength< DX >( infoNotPositiveLength, x.rows() ),
          x.rows()
          );
      CheckTool
        ( std::bind( std::greater< DX >(), std::placeholders::_1, DX() ),
          exception::NotPositiveLength< DX >( infoNotPositiveLength, x.cols() ),
          x.cols()
          );

      DY diff = error_lib::ArrayLength( ys, ye, infoNegativeLength );

      CheckTool
        ( std::bind( std::equal_to< DY >(), std::placeholders::_1, diff ),
          exception::NotEqualLength< DX >( infoNotEqualLength, x.rows(), diff ),
          x.rows()
          );
    }

    /// @brief 線形モデルとして有効かをチェックする
    ///
    /// 線形モデルとして利用可能か以下のテストを行い、テストにパスしなかった場合は例外を投げる。
    ///
    /// -# 要素がゼロでないか
    /// -# 従属変数の要素の範囲が適切か(開始と終了が逆になっていないか)
    /// -# デザイン行列の行数と従属変数の要素数が一致しているか
    ///
    /// 1 のテストにパスしなかった場合は例外 tadah_fussy::exception::NotPositiveLength が、
    /// 2 のテストにパスしなかった場合は例外 tadah_fussy::exception::NegativeLength が、
    /// 3 のテストにパスしなかった場合は例外 tadah_fussy::exception::NotEqualLength が投げられる。
    ///
    /// テンプレート引数の MATRIX はデザイン行列の型を、
    /// テンプレート引数の In は従属変数のデータ列の反復子の型をそれぞれ表す。
    ///
    /// tadah_fussy::exception::NotPositiveLength のテンプレート引数は MATRIX::size_type となる。
    ///
    /// tadah_fussy::exception::NegativeLength のテンプレート引数は std::iterator_traits< In >::difference_type となる。
    ///
    /// tadah_fussy::exception::NotEqualLength のテンプレート引数は MATRIX::size_type となることに注意。
    ///
    /// @param x デザイン行列への参照
    /// @param ys 従属変数の要素の開始位置
    /// @param ye 従属変数の要素の末尾の次の位置
    /// @return なし
    template< class MATRIX, class In >
    void CheckLinearModel( const MATRIX& x, In ys, In ye )
    {
      const std::string NOT_POSITIVE_LENGTH_INFO = "tadah_fussy::exception::NotPositiveLength at error_lib::CheckLinearModel";
      const std::string NEGATIVE_LENGTH_INFO = "tadah_fussy::exception::NegativeLength at error_lib::CheckLinearModel";
      const std::string NOT_EQUAL_LENGTH_INFO = "tadah_fussy::exception::NotEqualLength at error_lib::CheckLinearModel";

      CheckLinearModel( x, ys, ye, NOT_POSITIVE_LENGTH_INFO, NEGATIVE_LENGTH_INFO, NOT_EQUAL_LENGTH_INFO );
    }

    /// @brief 線形モデルとして有効かをチェックする
    ///
    /// 線形モデルとして利用可能か以下のテストを行い、テストにパスしなかった場合は例外を投げる。
    ///
    /// -# デザイン行列の要素がゼロでないか
    /// -# 従属変数の要素がゼロでないか
    /// -# デザイン行列の行数と従属変数の行数が一致しているか
    ///
    /// 1 のテストにパスしなかった場合は例外 tadah_fussy::exception::NotPositiveLength< MATRIX1::size_type > が、
    /// 2 のテストにパスしなかった場合は例外 tadah_fussy::exception::NotPositiveLength< MATRIX2::size_type > が、
    /// 3 のテストにパスしなかった場合は例外 tadah_fussy::exception::NotEqualLength< MATRIX1::size_type > が投げられる。
    ///
    /// テンプレート引数の MATRIX1 はデザイン行列の型を、
    /// テンプレート引数の MATRIX2 は従属変数の行列の型をそれぞれ表す。
    ///
    /// @param x デザイン行列への参照
    /// @param y 従属変数の行列への参照
    /// @param infoNotPositiveLength1 デザイン行列 x での例外 tadah_fussy::exception::NotPositiveLength に関する情報
    /// @param infoNotPositiveLength2 従属変数 y での例外 tadah_fussy::exception::NotPositiveLength に関する情報
    /// @param infoNotEqualLength 例外 tadah_fussy::exception::NotEqualLength に関する情報
    /// @return なし
    template< class MATRIX1, class MATRIX2 >
    void CheckLinearModel( const MATRIX1& x, const MATRIX2& y,
                           const std::string& infoNotPositiveLength1,
                           const std::string& infoNotPositiveLength2,
                           const std::string& infoNotEqualLength )
    {
      using DX = typename MATRIX1::size_type;
      using DY = typename MATRIX2::size_type;

      CheckTool
        ( std::bind( std::greater< DX >(), std::placeholders::_1, DX() ),
          exception::NotPositiveLength< DX >( infoNotPositiveLength1, x.rows() ),
          x.rows()
          );
      CheckTool
        ( std::bind( std::greater< DX >(), std::placeholders::_1, DX() ),
          exception::NotPositiveLength< DX >( infoNotPositiveLength1, x.cols() ),
          x.cols()
          );
      CheckTool
        ( std::bind( std::greater< DY >(), std::placeholders::_1, DY() ),
          exception::NotPositiveLength< DY >( infoNotPositiveLength2, y.rows() ),
          y.rows()
          );
      CheckTool
        ( std::bind( std::greater< DY >(), std::placeholders::_1, DY() ),
          exception::NotPositiveLength< DY >( infoNotPositiveLength2, y.cols() ),
          y.cols()
          );

      CheckTool
        ( std::bind( std::equal_to< DY >(), std::placeholders::_1, y.rows() ),
          exception::NotEqualLength< DX >( infoNotEqualLength, x.rows(), y.rows() ),
          x.rows()
          );
    }

    /// @brief 線形モデルとして有効かをチェックする
    ///
    /// 線形モデルとして利用可能か以下のテストを行い、テストにパスしなかった場合は例外を投げる。
    ///
    /// -# デザイン行列の要素がゼロでないか
    /// -# 従属変数の要素がゼロでないか
    /// -# デザイン行列の行数と従属変数の行数が一致しているか
    ///
    /// 1 のテストにパスしなかった場合は例外 tadah_fussy::exception::NotPositiveLength< MATRIX1::size_type > が、
    /// 2 のテストにパスしなかった場合は例外 tadah_fussy::exception::NotPositiveLength< MATRIX2::size_type > が、
    /// 3 のテストにパスしなかった場合は例外 tadah_fussy::exception::NotEqualLength< MATRIX1::size_type > が投げられる。
    ///
    /// テンプレート引数の MATRIX1 はデザイン行列の型を、
    /// テンプレート引数の MATRIX2 は従属変数の行列の型をそれぞれ表す。
    ///
    /// @param x デザイン行列への参照
    /// @param y 従属変数の行列への参照
    /// @return なし
    template< class MATRIX1, class MATRIX2 >
    void CheckLinearModel( const MATRIX1& x, const MATRIX2& y )
    {
      const std::string NOT_POSITIVE_LENGTH_INFO = "tadah_fussy::exception::NotPositiveLength at error_lib::CheckLinearModel";
      const std::string NOT_EQUAL_LENGTH_INFO = "tadah_fussy::exception::NotEqualLength at error_lib::CheckLinearModel";

      CheckLinearModel( x, y, NOT_POSITIVE_LENGTH_INFO, NOT_POSITIVE_LENGTH_INFO, NOT_EQUAL_LENGTH_INFO );
    }

  } // namespace error_lib

  namespace error_message
  {
    /// @brief データ列の長さがゼロ以下であった場合のエラー・メッセージ
    const std::string NOT_POSITIVE_LENGTH = "must be positive length";
  }

} // namespace tadah_fussy
