/**
   @file optimization.h
   @brief 最適化問題定義用ヘッダ

   @author tadah_fussy
   @date 2017/03/18 新規作成
   @date 2021/03/17 ニュートン・ラフソン法を math_library.h から移動
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef OPTIMIZATION_20210402_H
#define OPTIMIZATION_20210402_H

#include <vector>
#include <iostream>
#include <exception>

#include <common/math_library.h>
#include <common/matrix.h>

#include <cmath>

/**
   @brief 作成者 tadah_fussy
**/
namespace tadah_fussy
{
  namespace math_lib
  {
    namespace solver
    {
      /// @brief 勾配法 ( Gradient Method )
      ///
      /// テンプレート引数 F は関数の型を、T はデータの型をそれぞれ表す。
      /// 収束判定値 e は f(x) がゼロに収束したか、ステップ幅がゼロに収束したかのいずれにも利用される。
      ///
      /// @note 進む方向は固定していない。f が初期値 x0 から減少方向にあり f(x0) < 0 の場合、
      ///       または f が初期値 x0 から増加方向にあり f(x0) > 0 の場合、h < 0 とする必要があることに注意。
      ///       h > 0 の状態から符号を変換するか判定を行う場合は、以下のようにすればよい。
      /// @code
      /// // 逆に進む場合は h の符号を反転する
      /// if ( ( f( x0 ) > f( x0 + e ) && f( x0 ) < 0 ) ||
      ///      ( f( x0 ) < f( x0 + e ) && f( x0 ) > 0 ) )
      ///   h = -h;
      /// @endcode
      ///
      /// @param f 関数 f(x)
      /// @param x0 x の初期値
      /// @param h ステップ幅
      /// @param e 収束判定値
      template< class F, class T >
      T GradientMethod( F f, T x0, T h, T e )
      {
        // 逆に進む場合は h の符号を反転する
        //if ( ( f( x0 ) > f( x0 + e ) && f( x0 ) < 0 ) ||
        //     ( f( x0 ) < f( x0 + e ) && f( x0 ) > 0 ) )
        //  h = -h;

        while ( std::isfinite( x0 ) && std::abs( f( x0 ) ) > e && std::abs( h ) > e ) {
          T x1 = x0 + h;
          // f(x0) と f(x1) が同符号なら h を二倍にしてさらに進める
          if ( f( x0 ) * f( x1 ) > T() ) {
            do {
              h *= 2;
              x0 = x1;
              x1 += h;
            } while ( std::isfinite( x0 ) && f( x0 ) * f( x1 ) > T() );
            h /= 2;
            // f(x0) と f(x1) が異符号なら h を半分にして戻す
          } else {
            do {
              h /= 2;
              x1 -= h;
            } while ( std::isfinite( x0 ) && f( x0 ) * f( x1 ) < T() );
            x0 = x1;
            h *= 2;
          }
#ifdef DEBUG
          std::cout << "f(" << x0 << ") = " << f( x0 ) << std::endl;
#endif
        }

        return( x0 );
      }

      /// @brief 二部法 ( Bisection Method )
      ///
      /// テンプレート引数 F は関数の型を、T はデータの型をそれぞれ表す。
      /// 収束判定値 e は f(x) がゼロに収束したか、x0, x1 が充分近くなったかのいずれにも利用される。
      /// x0, x1 の間に解があることを想定している。
      ///
      /// @param f 関数 f(x)
      /// @param x0, x1 解のある範囲
      /// @param e 収束判定値
      template< class F, class T >
      T BisectionMethod( F f, T x0, T x1, T e )
      {
        if ( f( x0 ) > f( x1 ) ) std::swap( x0, x1 );
        T xm = ( x0 + x1 ) / 2; // x0, x1 の中点
        //std::cout << xm << "," << f( xm ) << std::endl;

        while ( std::abs( f( xm ) ) > e && std::abs( x1 - x0 ) > e ) {
          // 中点が負の場合は解は x1 側にあるので範囲を ( xm, x1 ) にする
          if ( f( xm ) < 0 )
            x0 = xm;
          // 中点が正の場合は解は x0 側にあるので範囲を ( x0, xm ) にする
          else
            x1 = xm;
          if ( f( x0 ) > f( x1 ) ) std::swap( x0, x1 );
          xm = ( x0 + x1 ) / 2;
#ifdef DEBUG
          std::cout << "f(" << xm << ") = " << f( xm ) << std::endl;
#endif
        }

        return( xm );
      }

      /// @brief セカント法 ( Secant Method )
      ///
      /// テンプレート引数 F は関数の型を、T はデータの型をそれぞれ表す。
      /// 収束判定値 e は f(x) がゼロに収束したかに利用される。
      ///
      /// @param f 関数 f(x)
      /// @param x0 x の初期値
      /// @param x1 直線を求めるためのもう一つの点
      /// @param e 収束判定値
      template< class F, class T >
      T SecantMethod( F f, T x0, T x1, T e )
      {
        while ( std::abs( f( x1 ) ) > e ) {
          // 直線 ( x0, x1 ) と x 軸との交点を求める
          T x2 = ( f( x1 ) * x0 - f( x0 ) * x1 ) / ( f( x1 ) - f( x0 ) );
          x0 = x1;
          x1 = x2;
#ifdef DEBUG
          std::cout << "f(" << x1 << ") = " << f( x1 ) << std::endl;
#endif
        }

        return( x1 );
      }

      /// @brief はさみうち法 ( False Position Method )
      ///
      /// テンプレート引数 F は関数の型を、T はデータの型をそれぞれ表す。
      /// 収束判定値 e は f(x) がゼロに収束したか、x0, x1 が充分近くなったかのいずれにも利用される。
      /// x0, x1 の間に解があることを想定している。
      ///
      /// @param f 関数 f(x)
      /// @param x0, x1 解のある範囲
      /// @param e 収束判定値
      template< class F, class T >
      T FalsePositionMethod( F f, T x0, T x1, T e )
      {
        if ( f( x0 ) > f( x1 ) ) std::swap( x0, x1 );

        // 直線 ( x0, x1 ) と x 軸との交点
        T xm = ( f( x1 ) * x0 - f( x0 ) * x1 ) / ( f( x1 ) - f( x0 ) );
        //std::cout << xm << "," << f( xm ) << std::endl;

        while ( std::abs( f( xm ) ) > e && std::abs( x1 - x0 ) > e ) {
          // 交点が負の場合は解は x1 側にあるので範囲を ( xm, x1 ) にする
          if ( f( xm ) < T() )
            x0 = xm;
          // 交点が正の場合は解は x0 側にあるので範囲を ( x0, xm ) にする
          else
            x1 = xm;
          if ( f( x0 ) > f( x1 ) ) std::swap( x0, x1 );
          xm = ( f( x1 ) * x0 - f( x0 ) * x1 ) / ( f( x1 ) - f( x0 ) );
#ifdef DEBUG
          std::cout << "f(" << xm << ") = " << f( xm ) << std::endl;
#endif
        }

        return( xm );
      }

      /// @brief 逆二次補間法 ( Inverse Quadratic Interpolation )
      ///
      /// テンプレート引数 F は関数の型を、T はデータの型をそれぞれ表す。
      /// 収束判定値 e は f(x) がゼロに収束したかに利用される。
      ///
      /// @param f 関数 f(x)
      /// @param x0, x2 二次関数を求めるときに利用する二点
      /// @param e 収束判定値
      template< class F, class T >
      T InverseQuadraticInterpolation( F f, T x0, T x2, T e )
      {
        T x1 = ( x0 + x2 ) / 2; // 二次関数を求めるためのもう一点

        while ( std::abs( f( x2 ) ) > e ) {
          T f0 = f( x0 );
          T f1 = f( x1 );
          T f2 = f( x2 );
          T x3 = ( f0 != f1 && f1 != f2 ) ?
            -( x0 * f1 * f2 * ( f1 - f2 ) + x1 * f2 * f0 * ( f2 - f0 ) + x2 * f0 * f1 * ( f0 - f1 ) )
            / ( ( f0 - f1 ) * ( f1 - f2 ) * ( f2 - f0 ) ) :
            ( f2 * x0 - f0 * x2 ) / ( f2 - f0 );
          x0 = x1;
          x1 = x2;
          x2 = x3;
#ifdef DEBUG
          std::cout << "f(" << x2 << ") = " << f( x2 ) << std::endl;
#endif
        }

        return( x2 );
      }

      /// @brief ブレント法 ( Brent Method )
      ///
      /// テンプレート引数 F は関数の型を、T はデータの型をそれぞれ表す。
      /// 収束判定値 e は f(x) がゼロに収束したか、x0, x2 が充分近くなったかのいずれにも利用される。
      ///
      /// @param f 関数 f(x)
      /// @param x0, x2 二次関数または直線を求めるときに利用する二点
      /// @param e 収束判定値
      template< class F, class T >
      T BrentMethod( F f, T x0, T x2, T e )
      {
        if ( f( x0 ) > f( x2 ) ) std::swap( x0, x2 );
        T x1 = x0;   // 二次関数を求めるためのもう一点
        T px1 = T(); // 前回求めた x1 の値
        bool useBisection = true;
        while ( std::abs( f( x2 ) ) > e && std::abs( x2 - x0 ) > e ) {
          T f0 = f( x0 );
          T f1 = f( x1 );
          T f2 = f( x2 );

          // 二次関数または直線による近似関数と x 軸との交点を求める
          T xs = ( f0 != f1 && f1 != f2 ) ?
            -( x0 * f1 * f2 * ( f1 - f2 ) + x1 * f2 * f0 * ( f2 - f0 ) + x2 * f0 * f1 * ( f0 - f1 ) )
            / ( ( f0 - f1 ) * ( f1 - f2 ) * ( f2 - f0 ) ) :
            ( f2 * x0 - f0 * x2 ) / ( f2 - f0 );

          // 二分法を使う場合の判定条件
          if ( ( xs < ( 3 * x0 + x2 ) / 4 && xs > x2 ) ||
               ( useBisection && std::abs( xs - x2 ) >= std::abs( x2 - x1 ) / 2 ) ||
               ( ( ! useBisection ) && std::abs( xs - x2 ) >= std::abs( x1 - px1 ) / 2 ) ||
               ( useBisection && std::abs( x2 - x1 ) < e ) ||
               ( ( ! useBisection ) && std::abs( x1 - px1 ) < e )
               ) {
            xs = ( x0 + x2 ) / 2;
            useBisection = true;
          } else {
            useBisection = false;
          }
          px1 = x1;
          x1 = x2;
          if ( f( xs ) < T() )
            x0 = xs;
          else
            x2 = xs;
          if ( f( x0 ) > f( x2 ) ) std::swap( x0, x2 );
#ifdef DEBUG
          std::cout << "f(" << x2 << ") = " << f( x2 ) << std::endl;
#endif
        }

        return( x2 );
      }

      /// @brief ニュートン-ラフソン法により f(x) = y を満たす x (x0) を求める
      ///
      /// テンプレート引数の F は関数 f(x)、DF はその導関数 f'(x) の型を表す。
      /// また、D は x の値の型を表す。D は浮動小数点数の型であることを前提とする。
      ///
      /// @param y y ( = f(x0) ) の値
      /// @param f 関数 f(x)
      /// @param df f(x) の導関数 f'(x)
      /// @param xInit x の初期値
      /// @param maxCount 処理を行う最大回数
      /// @param threshold 収束を判断するしきい値。| [ f(x) - f(x0) ] / f'(x) | がこの値より小さくなれば処理を終了する。
      /// @return f(x) = yとなる x の値(計算に失敗した場合は NaN を返す)
      template< class F, class DF, class D >
      D Newton( D y, F f, DF df, D xInit, uint32_t maxCount, D threshold )
      {
        const D nan = std::numeric_limits< D >::quiet_NaN();
        D x = xInit;

        for ( uint32_t i = 0 ; i < maxCount ; ++i ) {
          D curY = f( x ) - y;
          D curDY = df( x );
          if ( std::isnan( curY ) || std::isnan( curDY ) ) return( nan );
          if ( curY == 0 ) return( x );
          if ( curDY == 0 ) return( nan );

          D diff = curY / curDY; // = [ f(x) - f(x0) ] / f'(x)
          x -= diff;
          if ( std::abs( diff ) < threshold )
            return( x );
        }

        return( nan );
      }

      /**
         @brief 山登り法 ( Hill Climbing ) 用多変数関数オブジェクト

         利用するには fx と nabla を派生クラスで実装する必要がある。
         fx は関数 f(x) の値 ( x は多変量 )、nabla は ∇f = { ∂f / ∂xi } の値を返す関数を表す。
         平面上の楕円の方程式の場合、以下のようになる ( エラー処理は省略 )。

         @code
         double a_, b_;   // 扁平率
         double x0_, y0_; // 中心の座標

         :

         // fx : f(x) = a( x - x0 )^2 + b( y - ya )^2 を求める
         //
         // ( x, y ) = ( x[0], x[1] ) として計算する
         virtual double fx( const std::vector< double >& x ) const
         {
           return( a_ * pow( x[0] - x0_, 2 ) + b_ * pow( x[1] - y0_, 2 ) );
         }

         // ∇f : ( ∂f/∂x, ∂f/∂y ) を求める
         //
         // ( x, y ) = ( x[0], x[1] ) として計算する
         // i = 0 で ∂f/∂x、i = 1 で ∂f/∂y を返す
         virtual double nabla( const std::vector< double >& x, size_t i ) const
         {
         if ( i == 0 )
         return( 2 * a_ * ( x[0] - x0_ ) );
         else
         return( 2 * b_ * ( x[1] - y0_ ) );
         }
         @endcode
      **/
      template< class T >
      class HC_Operator
      {
        std::vector< T > x0_; // 初期値

        /// @brief f(x) の値を求める
        ///
        /// @param x x の値(多変量)
        /// @return f(x) の値
        virtual T fx( const std::vector< T >& x ) const = 0;

      public:

        typedef typename std::vector< T >::size_type size_type;

        /// @brief 仮想デストラクタ(何もしない)
        virtual ~HC_Operator() {}

        /// @brief ∇f の値を求める
        ///
        /// @param x x の値(多変量)
        /// @param i x の何番目のパラメータに対する偏微分を返すかを指定する添字
        /// @return ∇f の i 番目の値
        virtual T nabla( const std::vector< T >& x, size_type i ) const = 0;

        /// @brief 初期値を代入する
        ///
        /// @param x0 代入する初期値
        /// @return なし
        void assign( const std::vector< T >& x0 )
        { x0_.assign( x0.begin(), x0.end() ); }

        /// @brief F(t) = f(x0 + t∇f(x0)) を返す
        ///
        /// @param t t の値
        /// @return F(t) の値
        T ft( T t ) const
        {
          std::vector< T > x( x0_ );
          for ( size_type i = 0 ; i < x0_.size() ; ++i )
            x[i] += t * nabla( x0_, i );

          return( fx( x ) );
        }

        /// @brief dF(t)/dt = ( ∇f(x0 + t∇f0), ∇f(x0) ) を返す
        ///
        /// @param t t の値
        /// @return dF/dt の値
        T operator()( T t ) const
        {
          std::vector< T > x( x0_ );
          for ( size_type i = 0 ; i < x0_.size() ; ++i )
            x[i] += t * nabla( x0_, i );

          T ans = 0;
          for ( size_type i = 0 ; i < x0_.size() ; ++i )
            ans += nabla( x0_, i ) * nabla( x, i );

          return( ans );
        }
      };

      /// @brief 山登り法 ( Hill Climbing ) による多変量関数の極値計算
      ///
      /// テンプレート引数の Op は多変量関数オブジェクト HC_Operator の派生型であることを想定
      ///
      /// @param op 関数 f(x)
      /// @param x0 初期値(結果をそのまま返す)
      /// @param h ステップ幅
      /// @param e 収束判定値
      /// @return なし
      template< class Op, class T >
      void HillClimbing( Op op, std::vector< T >* x0, T h, T e )
      {
        std::vector< T > dx( x0->size() );

        for ( ; ; ) {
          op.assign( *x0 );
          h = ( ( op( 0.0 ) * op.ft( 0.0 ) < 0 ) ? 1 : -1 ) * std::abs( h );
          T t = GradientMethod( op, 0.0, h, e );

          for ( typename std::vector< T >::size_type i = 0 ; i < x0->size() ; ++i )
            dx[i] = t * op.nabla( *x0, i );
#ifdef DEBUG
          matrix_lib::PrintArray( std::cout, x0->begin(), x0->end(), "x0 = " ); std::cout << std::endl;
#endif
          bool isConvergence = true;
          for ( typename std::vector< T >::size_type i = 0 ; i < x0->size() ; ++i ) {
            (*x0)[i] += dx[i];
            if ( dx[i] > e ) isConvergence = false;
          }

          if ( isConvergence ) break;
        }
      }

      /**
         @brief 共役勾配法 ( Conjugate Gradient Method ) 用多変数関数オブジェクト

         利用するには fx、nabla、hessian を派生クラスで実装する必要がある。
         fx は関数 f(x) の値 ( x は多変量 )、nabla は ∇f = { ∂f / ∂xi } の値、
         hessian はヘッセ行列 H = { ∂^2f / ∂xi∂xj } を返す関数を表す。
         平面上の楕円の方程式の場合、以下のようになる ( エラー処理は省略 )。

         @code
         double a_, b_;   // 扁平率
         double x0_, y0_; // 中心の座標

         :

         // fx : f(x) = a( x - x0 )^2 + b( y - ya )^2 を求める
         //
         // ( x, y ) = ( x[0], x[1] ) として計算する
         virtual double fx( const std::vector< double >& x ) const
         {
           return( a_ * pow( x[0] - x0_, 2 ) + b_ * pow( x[1] - y0_, 2 ) );
         }

         // ∇f : ( ∂f/∂x, ∂f/∂y ) を求める
         //
         // ( x, y ) = ( x[0], x[1] ) として計算する
         // i = 0 で ∂f/∂x、i = 1 で ∂f/∂y を返す
         virtual double nabla( const std::vector< double >& x, size_t i ) const
         {
           if ( i == 0 )
             return( 2 * a_ * ( x[0] - x0_ ) );
           else
             return( 2 * b_ * ( x[1] - y0_ ) );
         }

         // H : | ∂^2f/∂x^2,  ∂^2f/∂x∂y | を求める
         //     | ∂^2f/∂x∂y, ∂^2f/∂y^2  |
         //
         // ( x, y ) = ( x[0], x[1] ) として計算する
         // ( i, j ) = ( 0, 0 ) で ∂^2f/∂x^2、( i, j ) = ( 1, 1 ) で ∂^2f/∂y^2
         // ( i, j ) = ( 1, 0 ), ( 0, 1 ) で ∂^2f/∂x∂y を返す
         virtual double hessian( const std::vector< double >& x, size_t i, size_t j ) const
         {
           if ( i == 0 && j == 0 )
             return( 2 * a_ );
           else if ( i == 1 && j == 1 )
             return( 2 * b_ );
           else
             return( 0 );
         }
         @endcode
      **/
      template< class T >
      class CGM_Operator
      {
        std::vector< T > x0_; // 初期値
        std::vector< T > m_;  // 探索方向ベクトル

        /// @brief f(x) の値を求める
        ///
        /// @param x x の値(多変量)
        /// @return f(x) の値
        virtual T fx( const std::vector< T >& x ) const = 0;

      public:

        typedef typename std::vector< T >::size_type size_type;

        /// @brief 仮想デストラクタ(何もしない)
        virtual ~CGM_Operator() {}

        /// @brief ∇f の値を求める
        ///
        /// @param x x の値(多変量)
        /// @param i x の何番目のパラメータに対する偏微分を返すかを指定する添字
        /// @return ∇f の i 番目の値
        virtual T nabla( const std::vector< T >& x, size_type i ) const = 0;

        /// @brief ヘッセ行列 H の値を求める
        ///
        /// @param x x の値(多変量)
        /// @param i,j H の何行・何列目の二階偏微分を返すかを指定する添字
        /// @return H の i 行・j 列目の値
        virtual T hessian( const std::vector< T >& x, size_type i, size_type j ) const = 0;

        /// @brief 初期値を代入する
        ///
        /// @param x0 代入する初期値
        /// @param m 探索方向ベクトル
        /// @return なし
        void assign( const std::vector< T >& x0, const std::vector< T >& m )
        {
          x0_.assign( x0.begin(), x0.end() );
          m_.assign( m.begin(), m.end() );
        }

        /// @brief F(t) = f(x0 + t∇f(x0)) を返す
        ///
        /// @param t t の値
        /// @return F(t) の値
        T ft( T t ) const
        {
          std::vector< T > x( x0_ );
          for ( size_type i = 0 ; i < x0_.size() ; ++i )
            x[i] += t * m_[i];

          return( fx( x ) );
        }

        /// @brief dF(t)/dt = ( ∇f(x0 + t∇f0), ∇f(x0) ) を返す
        ///
        /// @param t t の値
        /// @return dF/dt の値
        T operator()( T t ) const
        {
          std::vector< T > x( x0_ );
          for ( size_type i = 0 ; i < x0_.size() ; ++i )
            x[i] += t * m_[i];

          T ans = 0;
          for ( size_type i = 0 ; i < x0_.size() ; ++i )
            ans += m_[i] * nabla( x, i );

          return( ans );
        }
      };

      /*
        CGM_CalcAlpha : 共役勾配法による連立方程式計算 ( αを求める )

        α(k) = -( m(k-1), h(k)・nabla(k) ) / ( m(k-1), h(k)・m(k-1) )
      */
      template< class T >
      T Optimization_CGM_CalcAlpha( const std::vector< T >& m, const SquareMatrix< T >& h, const std::vector< T >& nabla )
      {
        // 分子の計算
        std::vector< T > hNabla( m.size() );
        matrix_lib::multMV( h, nabla.begin(), hNabla.begin() );
        T n = std::inner_product( m.begin(), m.end(), hNabla.begin(), T() );

        // 分母の計算
        std::vector< T > hm( m.size() );
        matrix_lib::multMV( h, m.begin(), hm.begin() );
        T d = std::inner_product( m.begin(), m.end(), hm.begin(), T() );

        return( ( d == 0 ) ? 0 : -n / d );
      }

      /*
        CGM_CalcM : 共役勾配法による連立方程式計算 ( mを求める )

        m(k) = nabla(k) + alpha(k)・m(k-1)
      */
      template< class T >
      void Optimization_CGM_CalcM( const std::vector< T >& nabla, T alpha, std::vector< T >* m )
      {
        // alpha・m(k-1) の計算
        std::transform( m->begin(), m->end(), m->begin(), std::bind2nd( std::multiplies< T >(), alpha ) );

        // 両者の和を求める
        std::transform( nabla.begin(), nabla.end(), m->begin(), m->begin(), std::plus< T >() );
      }

      /// @brief 共役勾配法による多変量関数の極値計算
      ///
      /// テンプレート引数の Op は多変量関数オブジェクト CGM_Operator の派生型であることを想定
      ///
      /// @param op 関数 f(x)
      /// @param x0 初期値(結果をそのまま返す)
      /// @param h ステップ幅
      /// @param e 収束判定値
      /// @return なし
      template< class Op, class T >
      void ConjugateGradientMethod( Op op, std::vector< T >* x0, T h, T e )
      {
        T alpha = 0;                             // α
        SquareMatrix< T > hessian( x0->size() ); // H
        std::vector< T > nabla( x0->size() );    // ∇f
        std::vector< T > m( x0->size() );        // m

        for ( ; ; ) {
          // ∇f(k) の計算
          for ( typename std::vector< T >::size_type i = 0 ; i < x0->size() ; ++i )
            nabla[i] = op.nabla( *x0, i );
          // H(k) の計算
          for ( typename SquareMatrix< T >::size_type r = 0 ; r < hessian.rows() ; ++r )
            for ( typename SquareMatrix< T >::size_type c = 0 ; c < hessian.cols() ; ++c )
              hessian[r][c] = op.hessian( *x0, r, c );

          alpha = Optimization_CGM_CalcAlpha( m, hessian, nabla ); // α(k) = -( m(k-1), H(k)・∇f(k) ) / ( m(k-1), H(k)・m(k-1) )
          Optimization_CGM_CalcM( nabla, alpha, &m );              // m(k) = ∇f(k) + α(k)m(k-1)

          op.assign( *x0, m );
          h = ( ( op( 0.0 ) * op.ft( 0.0 ) < 0 ) ? 1 : -1 ) * std::abs( h );
          T t = GradientMethod( op, 0.0, h, e );
#ifdef DEBUG
          matrix_lib::PrintArray( std::cout, x0->begin(), x0->end(), "x0 = " ); std::cout << std::endl;
#endif
          if ( CGM_CalcX( t, m, x0->begin(), e ) ) break;         // x(k+1) = x(k) + t(k)・m(k)
        }
      }

    } // namespace solver

  } // namespace math_lib

} // namespace tadah_fussy

#endif
