/**
   @file serialize.h
   @brief データのシリアライズ・デシリアライズ化

   @author tadah_fussy
   @date 2019/04/09 新規作成
   @date 2019/04/16 error_library.h を利用したエラーチェックの追加
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef SERIALIZE_20210402_H
#define SERIALIZE_20210402_H

#include <cstddef>

#include <vector>
#include <utility>
#include <limits>
#include <iostream>

#include <common/error_library.h>

/**
   @brief 作成者 tadah_fussy
**/
namespace tadah_fussy
{
  namespace gadget
  {
    /**
       @brief データをビット列に変換する

       Data 型のデータをビット列として Code 型のデータに変換する。
       クラス・テンプレート引数の Code は変換後のデータ型、Out は出力先反復子をそれぞれ表す。
       メンバ関数 Serialization< Code, Out >::operator() がビット列への変換処理を行う。
       Serialization< Code, Out >::operator() のテンプレート引数 Data は変換前のデータの型を表す。

       Out は出力イテレータなので、データ型は未定義(void)である。したがって、Out の他に出力するビット列の型 Code も指定しなければならない。
       back_insert_iterator を利用した場合のサンプル・プログラムは以下の通り。

       @code
       std::vector< unsigned char > data;
       :
       std::vector< uint32_t > code;
       Serialization< uint32_t, std::back_insert_iterator< std::vector< uint32_t > > > serial( std::back_inserter( code ) );
       for ( std::vector< unsigned char >::const_iterator i = data.begin() ; i != data.end() ; ++i ) {
       serial( *i, sizeof( unsigned char ) * CHAR_BIT );
       }
       @endcode

       型の名前が冗長だが、std::for_each と Create_BackInsertSerialization を利用することで多少はすっきりと書ける。

       @code
       using namespace std::placeholders;

       std::vector< unsigned char > data;
       :
       std::vector< uint32_t > code;
       std::for_each( data.begin(), data.end(), std::bind( Create_BackInsertSerialization( code ), _1, sizeof( unsigned char ) * CHAR_BIT ) );
       @endcode

       サンプル・プログラムにおいて、Code とコンテナの要素の型が一致しなかった場合、結果は不定となる
       (余分な部分にゼロが埋められるか、足りない領域が削られてしまう)。

       Data 型及び Code 型は符号なし整数であることを前提としている。ともにビットシフト演算 ( >>, << ) を利用し、
       両者の間で論理演算 ( &, | ) も行われる。したがって、これらの演算が定義され、互いに型変換できる必要がある。
       基本的には組み込み変数での利用を想定している。
    **/
    template< class Code, class Out >
    class Serialization
    {
      Out out_;         // 出力先反復子
      Code cacheCode_;  // キャッシュデータ
      size_t cacheLen_; // キャッシュデータの現在の長さ

      // 最初のデータ書き込み
      // キャッシュにデータを先頭から書き込んで、一杯になったら出力する
      template< class Data >
      void pushOnce( const Data& data, size_t* length );

      // コンテナの要素として書き込む
      // ビット列に分解せずに一度に書き込める部分はこれで処理する
      template< class Data >
      void pushLoop( const Data& data, size_t* length );

      // 残りの部分を書き込む
      template< class Data >
      void setLastCode( const Data& data, size_t length );

      //Serialization( const Serialization& s );

      //Serialization& operator=( const Serialization& s );

    public:

      /// @brief 出力先反復子を指定して構築
      ///
      /// @param out 出力先反復子
      explicit Serialization( Out out )
	: out_( out ), cacheCode_( Code() ), cacheLen_( 0 ) {}

      /// @brief デストラクタ
      ///
      /// キャッシュに残ったデータは解体時に書き込まれる
      ~Serialization() { flush(); }

      /// @brief データの書き込みを行う
      ///
      /// @param data 書き込むデータ
      /// @param length データ長(ビット列の長さ)
      /// @return なし
      template< class Data >
      void operator()( Data data, size_t length );

      /// @brief キャッシュ内のデータを出力する
      ///
      /// @return なし
      void flush();

#ifdef DEBUG

      Code code() const { return( cacheCode_ ); }

      size_t len() const { return( cacheLen_ ); }

#endif
    };

    // Serialization< Code, Out >::pushOnce : cacheCode に data を埋め込む
    //
    // cacheCode が満杯になった場合は out へキャッシュを登録する
    //
    // data 登録するのデータへの参照
    // length データのビット長へのポインタ
    template< class Code, class Out >
    template< class Data >
    void Serialization< Code, Out >::pushOnce
    ( const Data& data, size_t* length )
    {
      const size_t CODE_BIT = sizeof( Code ) * std::numeric_limits< unsigned char >::digits / sizeof( unsigned char );
      size_t mod = CODE_BIT - cacheLen_; // キャッシュに登録可能なビット数

      if ( *length >= mod ) {
	// キャッシュが満杯になった場合はキャッシュを出力
	if ( mod == CODE_BIT )
	  *out_++ = data >> ( *length - mod );
	else
	  *out_++ = ( cacheCode_ << mod ) | ( data >> ( *length - mod ) );
	cacheLen_ = 0;  // 保持しているビット列はゼロに
	*length -= mod; // 登録するビット列の残りを計算
      } else {
	// キャッシュがまだ満杯にならない場合キャッシュとそのビット長を更新
	cacheLen_ += *length;
	cacheCode_ = ( cacheCode_ << *length ) | data;
	*length = 0; // ビット列の残りはゼロに
      }
    }

    // Serialization< Code, Out >::pushLoop : data をそのままデータ列 out へ登録する
    //
    // length がキャッシュのビット長よりも長い場合、直接プッシュしてしまう
    //
    // data 登録するのデータへの参照
    // length 現在の残りビット長へのポインタ(pushOnce 処理後の残りビット列の長さを保持するとする)
    template< class Code, class Out >
    template< class Data >
    void Serialization< Code, Out >::pushLoop
    ( const Data& data, size_t* length )
    {
      const size_t CODE_BIT = sizeof( Code ) * std::numeric_limits< unsigned char >::digits / sizeof( unsigned char );

      while ( *length >= CODE_BIT ) {
	*out_++ = ( data >> ( *length - CODE_BIT ) ) &
	  ~( Code() );
	*length -= CODE_BIT;
      }
    }

    // Serialization< Code, Out >::setLastCode : data の残りビット列を cacheCode へ登録する
    //
    // data 登録するのデータへの参照
    // length 残りビット列の長さ
    template< class Code, class Out >
    template< class Data >
    void Serialization< Code, Out >::setLastCode
    ( const Data& data, size_t length )
    {
      if ( length > 0 ) {
	cacheCode_ = data & ~( ~( Code() ) << length ); // 残ったビット列をキャッシュに保持
	cacheLen_ = length;
      }
    }

    // Serialization< Code, Out >::operator() : 長さ length のデータ data を out に書き込む
    //
    // @param data 登録するデータへの参照
    // @param length データのビット長
    template< class Code, class Out >
    template< class Data >
    void Serialization< Code, Out >::operator()
      ( Data data, size_t length )
    {
      // data の length 以上の上位ビットをクリアする
      const size_t DATA_BIT = sizeof( Data ) * std::numeric_limits< unsigned char >::digits / sizeof( unsigned char );
      if ( length < DATA_BIT )
	data &= ~( ~( Data() ) << length );

      pushOnce( data, &length );
      pushLoop( data, &length );
      setLastCode( data, length );
    }

    // Serialization< Code, Out >::flush : キャッシュの出力とクリア
    //
    // @return なし
    template< class Code, class Out >
    void Serialization< Code, Out >::flush()
    {
      const size_t CODE_BIT = sizeof( Code ) * std::numeric_limits< unsigned char >::digits / sizeof( unsigned char );

      if ( cacheLen_ > 0 )
	*out_++ = cacheCode_ << ( CODE_BIT - cacheLen_ );

      cacheCode_ = Code();
      cacheLen_ = 0;
    }

    /***********************************************************************************************/

    /*
      GetLoop : データを一度に取得する関数用オブジェクト

      データに対する型の長さが符号のそれより短い場合、この処理は不要となる。
      コンパイル時のワーニングを避けるため、長さの比較結果によってテンプレート引数 EXEC で
      処理を分けている。データに対する型の長さが符号のそれより短い場合は EXEC = false とな
      るようにして、EXEC = false の部分特別バージョンのテンプレート関数は何もしないように
      してある。
    */
    template< class In, bool EXEC >
    struct GetLoop
    {
      // 符号の型
      typedef typename std::iterator_traits< In >::value_type Code;

      template< class Data >
      bool operator()( Data* data, size_t* length, Code* cacheCode, In* is, In ie );
    };

    // GetLoop< In, EXEC >::operator() : ビット列に分解せずに data へ取り込む
    //
    // data 読み込んだ値を保持する変数へのポインタ
    // length 現在のキャッシュのビット長へのポインタ
    // cacheCode 現在のキャッシュへのポインタ
    // is 読み込み先の符号列への反復子へのポインタ
    // ie 読み込み先の符号列の末尾
    template< class In, bool EXEC >
    template< class Data >
    bool GetLoop< In, EXEC >::operator()
      ( Data* data, size_t* length, Code* cacheCode, In* is, In ie )
    {
      const size_t CODE_BIT = sizeof( Code ) * std::numeric_limits< unsigned char >::digits / sizeof( unsigned char );

      while ( *length >= CODE_BIT ) {
	*data = ( *data << CODE_BIT ) | *cacheCode;
	*length -= CODE_BIT;
	if ( *length > 0 && *is == ie ) return( false );
	if ( *is != ie ) *cacheCode = *( ( *is )++ );
      }

      return( true );
    }

    /*
      GetLoop : データを一度に取得する関数用オブジェクト(何もしないバージョン)

      データに対する型の長さが符号のそれより短い場合、この処理は不要となる。
      コンパイル時のワーニングを避けるため、長さの比較結果によってテンプレート引数 EXEC で
      処理を分けている。データに対する型の長さが符号のそれより短い場合は EXEC = false とな
      るようにして、EXEC = false の部分特別バージョンのテンプレート関数は何もしないように
      してある。
    */
    template< class In >
    struct GetLoop< In, false >
    {
      // 符号の型
      typedef typename std::iterator_traits< In >::value_type Code;

      template< class Data >
      bool operator()( Data* data, size_t* length, Code* cacheCode, In* is, In ie )
      { return( true ); }
    };

    /**
       @brief ビット列をデータに変換する

       Code 型のビット列を Data 型のデータに変換する。
       クラス・テンプレート引数の In は入力先反復子を表す。
       メンバ関数 Deserialization< In >::operator() がビット列からの変換処理を行う。
       Deserialization< In >::operator() のテンプレート引数 Data は変換後のデータの型を表す。

       サンプル・プログラムを以下に示す。

       @code
       std::vector< uint32_t > code;
       :
       std::vector< unsigned char > obj( szOfVec );
       Deserialization< std::vector< uint32_t >::const_iterator > deserial( code.begin(), code.end() );
       for ( std::vector< unsigned char >::iterator i = obj.begin() ; i != obj.end() ; ++i ) {
       deserial( i, sizeof( unsigned char ) * CHAR_BIT );
       }
       @endcode

       Data 型及び Code 型は符号なし整数であることを前提としている。ともにビットシフト演算 ( >>, << ) を利用し、
       両者の間で論理演算 ( &, | ) も行われる。したがって、これらの演算が定義され、互いに型変換できる必要がある。
       基本的には組み込み変数での利用を想定している。
    **/
    template< class In >
    class Deserialization
    {
      // 符号(ビット列)の要素の型
      typedef typename std::iterator_traits< In >::value_type Code;

      In is_;           // 入力する符号列の先頭
      In ie_;           // 入力する符号列の末尾
      Code cacheCode_;  // キャッシュ
      size_t cacheLen_; // キャッシュの現在の長さ

      // キャッシュからの最初の読み込み
      // キャッシュからデータを読み込んで、足りなくなったら入力する
      template< class Data >
      bool getFromCache( Data* data, size_t* length );

      //template< bool EXEC, class Data >
      //void getLoop( Data* data, size_t* length );

      // キャッシュからの残りの読み込み
      template< class Data >
      void getLastCode( Data* data, size_t length );

      //Deserialization( const Deserialization& s );

      //Deserialization& operator=( const Deserialization& s );

    public:

      /// @brief 入力先を指定して構築
      ///
      /// - 符号列の先頭と末尾の次の位置が逆転していた場合、例外 ExceptionNegativeLength< std::iterator_traits< In >::difference_type > を投げる。
      ///
      /// @param is 入力する符号列の先頭
      /// @param ie 入力する符号列の末尾の次
      explicit Deserialization( In is, In ie )
	: is_( is ), ie_( ie ), cacheCode_( Code() ), cacheLen_( 0 )
      {
	const std::string FUNC_NAME = "Deserialization constructor";
	error_lib::CheckArrayLength( is_, ie_, FUNC_NAME + " : [ is, ie ) " + error_message::NEGATIVE_LENGTH );
      }

      /// @brief 符号列からデータを読み込む
      ///
      /// - 符号列から読み込むことができない(末尾に達した)場合は std::out_of_range を投げる。
      ///
      /// @param data 読み込んだデータを保持する変数へのポインタ
      /// @param length 読み込むデータの長さ(ビット長)
      /// @return なし
      template< class Data >
      void operator()( Data* data, size_t length );

#ifdef DEBUG

      Code code() const { return( cacheCode_ ); }

      size_t len() const { return( cacheLen_ ); }

#endif
    };

    // キャッシュから読み込める分だけ data へ取り込む
    //
    // キャッシュが空になったら入力先から新たに読み込む
    // data は下詰めで登録される
    //
    // data 読み込んだ値を保持する変数へのポインタ
    // length 現在のキャッシュのビット長へのポインタ
    template< class In >
    template< class Data >
    bool Deserialization< In >::getFromCache
    ( Data* data, size_t* length )
    {
      const size_t CODE_BIT = sizeof( Code ) * std::numeric_limits< unsigned char >::digits / sizeof( unsigned char );

      // キャッシュが足りない場合
      if ( cacheLen_ < *length ) {
	// キャッシュの残りを書き出す
	*data = cacheCode_ & ~( ~( static_cast< Code >( Code() ) ) << cacheLen_ );
	// 新たなデータをキャッシュに取り込む
	if ( is_ == ie_ ) return( false );
	cacheCode_ = *is_++;
	*length -= cacheLen_; // data に書き込む残りのビット数
	cacheLen_ = CODE_BIT; // キャッシュは満杯
	// キャッシュが足りる場合
      } else {
	cacheLen_ -= *length; // キャッシュの残りを更新
	if ( *length == CODE_BIT )
	  *data = cacheCode_;
	else
	  *data = ( cacheCode_ >> cacheLen_ ) & ~( ( ~( static_cast< Code >( Code() ) ) ) << *length );
	*length = 0;
      }

      return( true );
    }

    // data の未登録部分をキャッシュから取得して埋め込む
    //
    // data 読み込んだ値を保持する変数へのポインタ
    // length 現在のキャッシュのビット長
    template< class In >
    template< class Data >
    void Deserialization< In >::getLastCode
    ( Data* data, size_t length )
    {
      const size_t CODE_BIT = sizeof( Code ) * std::numeric_limits< unsigned char >::digits / sizeof( unsigned char );

      if ( length > 0 ) {
	*data = ( *data << length ) | ( cacheCode_ >> ( CODE_BIT - length ) );
	cacheLen_ -= length;
      }
    }

    // 長さ length のデータ data を符号列から取り込む
    //
    // data 読み込んだ値を保持する変数へのポインタ
    // length 現在のキャッシュのビット長
    template< class In >
    template< class Data >
    void Deserialization< In >::operator()
      ( Data* data, size_t length )
    {
      const size_t CODE_BIT = sizeof( Code ) * std::numeric_limits< unsigned char >::digits / sizeof( unsigned char );
      const size_t DATA_BIT = sizeof( Data ) * std::numeric_limits< unsigned char >::digits / sizeof( unsigned char );
      const bool EXEC = CODE_BIT < DATA_BIT;

      if ( ! getFromCache( data, &length ) )
	throw std::out_of_range( "Deserialization< In >::operator() : out of range" );
      GetLoop< In, EXEC > getLoop;
      if ( ! getLoop( data, &length, &cacheCode_, &is_, ie_ ) )
	throw std::out_of_range( "Deserialization< In >::operator() : out of range" );
      getLastCode( data, length );
    }

    /***********************************************************************************************/

    /// @brief std::insert_iterator を利用した Serialization オブジェクトを生成するヘルパ関数
    ///
    /// @param c 対象のコンテナ
    /// @param out 出力先のイテレータ
    /// @return std::insert_iterator を利用した Serialization オブジェクト
    template< class Container, class Out >
    Serialization<
      typename Container::value_type,
      std::insert_iterator< Container > >
    Create_InsertSerialization( Container& c, Out out )
    {
      return( Serialization<
	      typename Container::value_type,
	      std::insert_iterator< Container > >
	      ( std::inserter( c, out ) )
	      );
    }

    /// @brief std::back_insert_iterator を利用した Serialization オブジェクトを生成するヘルパ関数
    ///
    /// @param c 対象のコンテナ
    /// @return std::back_insert_iterator を利用した Serialization オブジェクト
    template< class Container >
    Serialization<
      typename Container::value_type,
      std::back_insert_iterator< Container > >
    Create_BackInsertSerialization( Container& c )
    {
      return( Serialization<
	      typename Container::value_type,
	      std::back_insert_iterator< Container > >
	      ( std::back_inserter( c ) )
	      );
    }

    /// @brief std::front_insert_iterator を利用した Serialization オブジェクトを生成するヘルパ関数
    ///
    /// @param c 対象のコンテナ
    /// @return std::front_insert_iterator を利用した Serialization オブジェクト
    template< class Container >
    Serialization<
      typename Container::value_type,
      std::front_insert_iterator< Container > >
    Create_FrontInsertSerialization( Container& c )
    {
      return( Serialization<
	      typename Container::value_type,
	      std::front_insert_iterator< Container > >
	      ( std::front_inserter( c ) )
	      );
    }

  } // namespace gadget

} // namespace tadah_fussy

#endif
